import asyncio

SERVER_HOST = "163.172.99.233"

SETUP_PORT = ### <-- SETUP PORT GIVEN WHEN STARTING INSTANCE
SERVER_PORT = ### <-- LICENSE SERVER PORT GIVEN WHEN STARTING INSTANCE

PUBLIC_TUNNEL_HOST = ### <-- YOUR PUBLIC HOST
PUBLIC_TUNNEL_PORT = ### <-- YOUR PUBLIC PORT

LOCAL_MITM_HOST = ""
LOCAL_MITM_PORT = ### <-- YOUR LOCAL BINDING PORT

async def client_handler(clientside_reader, clientside_writer):
    print("New connection")
    serverside_reader, serverside_writer = await asyncio.open_connection(SERVER_HOST, SERVER_PORT)

    while True:
        if not (client_to_server := (await clientside_reader.read(1024))):
            print("Connection closed by client")
            serverside_writer.close()
            await serverside_writer.wait_closed()
            return

        print("Client: " + client_to_server.hex())
        serverside_writer.write(client_to_server)
        await serverside_writer.drain()

        if not (server_to_client := (await serverside_reader.read(1024))):
            print("Connection closed by server")
            return

        print("Server: " + server_to_client.hex())
        clientside_writer.write(server_to_client)
        await clientside_writer.drain()

async def print_setup_read(reader, writer):
    if not (setup_msg := (await reader.read(1024))):
        print("\033[31mConnection closed during setup phase\033[0m")
        writer.close()
        await writer.wait_closed()
        exit()
    print(setup_msg.decode())

# Start an instance of the challenge
async def setup():
    setup_reader, setup_writer = await asyncio.open_connection(SERVER_HOST, SETUP_PORT)
    await print_setup_read(setup_reader, setup_writer)
    setup_writer.write(f"{PUBLIC_TUNNEL_HOST}:{PUBLIC_TUNNEL_PORT}".encode())
    await setup_writer.drain()
    await print_setup_read(setup_reader, setup_writer)

# Start the MITM listener
async def mitm_listener():
    server = await asyncio.start_server(client_handler, LOCAL_MITM_HOST, LOCAL_MITM_PORT)
    async with server:
        await server.serve_forever()

async def main():
    # Run the listener and the setup tasks in parallel to avoid the challenge telling us it can't connect to our listener
    await asyncio.gather(mitm_listener(), setup())

asyncio.run(main())