#include "auth.h"

#include <string.h>

#include "openssl/crypto.h"
#include "openssl/evp.h"

#include "error.h"

typedef struct user_t {
  char *username;
  unsigned char pwd_hash[EVP_MAX_MD_SIZE];
} user_t;

#define KNOWN_USERS_SZ 5
const user_t KNOWN_USERS[KNOWN_USERS_SZ] = {
  {"@mafiaBOSS",
    {
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
    }},
  {"@mafiaBRO",
    {
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
    }},
  {"@mafiaDEV",
    {
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
    }},
  {"@mafiaTECHNICIAN",
    {
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
      'R', 'E', 'D', 'A', 'C', 'T', 'E', 'D',
    }},
  {"@mafiaTEST",
    {
      0xb1, 0xbe, 0xe4, 0x86, 0x51, 0x7e, 0x8e, 0xe2,
      0x7b, 0x19, 0x57, 0x61, 0xb9, 0x34, 0x70, 0xfc,
      0xef, 0x4c, 0xf7, 0xe4, 0x9b, 0x7c, 0x26, 0x0a,
      0x0d, 0xf0, 0x1f, 0x44, 0x11, 0xf7, 0x05, 0xdf,
      0x9d, 0x1e, 0x4a, 0x72, 0xf2, 0x67, 0xbe, 0xfd,
      0x3e, 0x25, 0xf9, 0x52, 0x63, 0x29, 0x18, 0xe8,
      0x38, 0x4c, 0x93, 0x16, 0x2f, 0x73, 0x20, 0x14,
      0xfe, 0xcf, 0xa1, 0x44, 0xe3, 0x8d, 0x75, 0x93,
    }}, // OnlyRequiredToSignAndSendMessages
};

void hash(unsigned char *out, char *in) {
  EVP_MD *md = EVP_MD_fetch(NULL, "SHA512", NULL);
  if (md == NULL) {
    hsm_error(ERROR_OPENSSL);
  }

  unsigned int sz;
  if (EVP_Digest(in, strlen(in), out, &sz, md, NULL) != 1) {
    hsm_error(ERROR_OPENSSL);
  }
  if (sz != EVP_MAX_MD_SIZE) {
    hsm_error(ERROR_OPENSSL);
  }
  EVP_MD_free(md);
}

void auth(char *username, char *password) {
  for (size_t i = 0; i < KNOWN_USERS_SZ; i++) {
    user_t user = KNOWN_USERS[i];
    if (strcmp(user.username, username) == 0) {
      unsigned char input_pwd_hash[EVP_MAX_MD_SIZE];
      hash(input_pwd_hash, password);

      if (CRYPTO_memcmp(user.pwd_hash, input_pwd_hash, EVP_MAX_MD_SIZE) != 0) {
        hsm_error(ERROR_AUTH);
      }
      return;
    }
  }
  hsm_error(ERROR_AUTH);
}
