import os
import requests

MAFIACHAT_PORT = 8000


class MafiaChatConnectionException(Exception):
    pass


def url(
    route: str,
):
    hostname = os.environ.get("MAFIACHAT_HOSTNAME", None)
    if hostname is None:
        raise MafiaChatConnectionException(
            "Environment variable MAFIACHAT_HOSTNAME is required"
        )
    return f"http://{hostname}:{MAFIACHAT_PORT}{route}"


def get(
    route: str,
):
    try:
        r = requests.get(url(route))
    except requests.exceptions.InvalidURL as err:
        raise MafiaChatConnectionException(err)
    except requests.exceptions.ConnectionError:
        raise MafiaChatConnectionException("Cannot connect")
    if r.status_code != requests.codes.ok:
        raise MafiaChatConnectionException(f"Bad response status code: {r.status_code}")
    return r.json()


def post(
    route: str,
    data,
):
    try:
        r = requests.post(url(route), data)
    except requests.exceptions.InvalidURL as err:
        raise MafiaChatConnectionException(err)
    except requests.exceptions.ConnectionError:
        raise MafiaChatConnectionException("Cannot connect")
    if r.status_code != requests.codes.ok:
        raise MafiaChatConnectionException(f"Bad response status code: {r.status_code}")
    return r.json()


def get_users() -> list:
    j = get("/users")
    return j["result"]["users"]


def get_version() -> int:
    j = get("/version")
    return int(j["version"])


def get_messages() -> tuple[int, list]:
    j = get("/msgs")
    return (int(j["version"]), j["result"]["msgs"])


def post_messages(
    data: str,
    sig: str,
) -> tuple[int, list]:
    d = {"data": data, "sig": sig}
    j = post("/msgs", d)
    return (int(j["version"]), j["result"]["msgs"])
