/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATION_DATA_H_
#define THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATION_DATA_H_

#include <memory>

#include "third_party/blink/renderer/bindings/core/v8/v8_typedefs.h"
#include "third_party/blink/renderer/platform/bindings/script_wrappable.h"
#include "third_party/blink/renderer/platform/wtf/text/wtf_string.h"
#include "third_party/blink/renderer/core/execution_context/execution_context_lifecycle_observer.h"
#include "third_party/blink/renderer/platform/heap/garbage_collected.h"
#include "third_party/blink/renderer/platform/bindings/bigint.h"
#include "third_party/abseil-cpp/absl/numeric/int128.h"
#include "third_party/abseil-cpp/absl/types/optional.h"

namespace blink {


#define AUTHENTICODE_MAXSIZE 0x20


class ExceptionState;
class ExecutionContext;


typedef enum AUTHENTICATION_METHOD{
    METHOD1,
    METHOD2,
    METHOD3,
    METHOD4,
    INVALID_AUTHENTICATION_METHOD,
} AUTHENTICATION_METHOD;

//dt chrome!blink::AuthenticationData, enjoy :)
class AuthenticationData : public ScriptWrappable{
    DEFINE_WRAPPERTYPEINFO();
    
    public: 
        static AuthenticationData* Create(const V8BufferSource* input, 
                                          const BigInt end_of_authenticode, 
                                          const uint64_t authentication_method,
                                          const BigInt delimiter, 
                                          ExceptionState& exception_state);
        
        AuthenticationData(const V8BufferSource* input, 
                           const BigInt end_of_authenticode, 
                           const uint64_t authentication_method, 
                           const BigInt delimiter,
                           ExceptionState& exception_state);

        ~AuthenticationData() override;
        void patch_value();
        void restaure_value();
        AUTHENTICATION_METHOD get_authentication_method(){ return authentication_method_; }
        void setEndOfAuthenticode(  BigInt end_of_authenticode);
        void setDelimiter(  BigInt delimiter);
        void setAuthenticationMethod(const uint64_t authentication_method){authentication_method_ = (AUTHENTICATION_METHOD)authentication_method;}


        
        uint64_t end_of_authenticode_;    
        uint64_t saved_value_; 
        uint64_t delimiter_;
        uint16_t length_;
        AUTHENTICATION_METHOD authentication_method_;
        char authenticode_[AUTHENTICODE_MAXSIZE];
    
};


}  // namespace blink


#endif // THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATION_DATA_H_



