/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "third_party/blink/renderer/modules/authenticator/authenticator.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_union_arraybuffer_arraybufferview.h"
#include "third_party/blink/renderer/core/typed_arrays/dom_array_buffer.h"
#include "third_party/blink/renderer/core/typed_arrays/dom_array_buffer_view.h"
#include "third_party/blink/renderer/platform/bindings/exception_state.h"
#include "third_party/blink/renderer/platform/wtf/text/string_view.h"



namespace blink {

class AuthenticationData;


bool test_hash_1(char* authenticode, char delimiter, NotShared<DOMArrayBufferView> key_){
    /*todo*/
    return false;
}


bool test_hash_2(char* authenticode, char delimiter, NotShared<DOMArrayBufferView> key_){
    /*todo*/
    return false;
}

bool test_hash_3(char* authenticode, char delimiter, NotShared<DOMArrayBufferView> key_){
    /*todo*/
    return false;
}

bool test_hash_4(char* authenticode, char delimiter, NotShared<DOMArrayBufferView> key_){
    /*todo*/
    return false;
}



Authenticator* Authenticator::Create(NotShared<DOMArrayBufferView> key) {
    return MakeGarbageCollected<Authenticator>(key);
}


Authenticator::~Authenticator() = default;

 

String Authenticator::Authenticate(AuthenticationData* authentication_data, ExceptionState& exception_state){
    
    

    memset(hash_, 0, MAX_HASH_SIZE);
    

    if (authentication_data == NULL){
        exception_state.ThrowSecurityError
        (
            "Supplied authentication data is erroneous");
        return String();
    }

    authentication_data_ = authentication_data;

    if (!try_authentication()){
        if (!authentication_fallback()){
            return String("Not Authenticated");
        }
    }

    return String("Authenticated");
}


void Authenticator::Trace(Visitor* visitor) const{
    visitor->Trace(key_);
    visitor->Trace(authentication_data_);
    ScriptWrappable::Trace(visitor);
}



bool Authenticator::get_authentication_status(){
    return is_authenticated_;
}



bool Authenticator::try_authentication(){

    authentication_data_->patch_value();

    switch(authentication_data_->get_authentication_method()){

        case METHOD1: 
            if(test_hash_1(authentication_data_->authenticode_, authentication_data_->delimiter_, key_)){
                is_authenticated_ = true;
            }
            break;
        
        case METHOD2: 
            if(test_hash_2(authentication_data_->authenticode_, authentication_data_->delimiter_, key_)){
                is_authenticated_ = true;
            }
            break;

        case METHOD3: 
            if(test_hash_3(authentication_data_->authenticode_, authentication_data_->delimiter_, key_)){
                is_authenticated_ = true;
            }
            break;

        case METHOD4: 
            if(test_hash_4(authentication_data_->authenticode_, authentication_data_->delimiter_, key_)){
                is_authenticated_ = true;
            }
            break;

        case INVALID_AUTHENTICATION_METHOD: 
            is_authenticated_ = false;
            break;
        
        default: 
            is_authenticated_ = false;
            break;
    }
    
    authentication_data_->restaure_value();
    return is_authenticated_;
}



bool Authenticator::authentication_fallback(){
    return false;
}




String Authenticator::get_authentication_data_info(){
    
    char data[MAX_AUTH_INFO_SIZE] = {0};
    unsigned int cx = 0;
    
    cx = snprintf(data, MAX_AUTH_INFO_SIZE, "[end_of_authenticode = 0x%llx\nsaved_value = 0x%llx\nauthentication_method = %d\nauthenticode = [", 
                                          authentication_data_->end_of_authenticode_, 
                                          authentication_data_->saved_value_,
                                          authentication_data_->authentication_method_);
                                                                                                                                                                       
    if (authentication_data_->length_){   
        
        for(auto i = 0 ; (i < authentication_data_->length_ - 1) && (cx < MAX_AUTH_INFO_SIZE) ; i++){
            cx += snprintf(data + cx, MAX_AUTH_INFO_SIZE - cx, "0x%02x, ", authentication_data_->authenticode_[i]);   
        }

        if (cx >= MAX_AUTH_INFO_SIZE){
            return String(data);
        }
        cx += snprintf(data + cx, MAX_AUTH_INFO_SIZE - cx, "0x%02x]", authentication_data_->authenticode_[authentication_data_->length_ - 1]);
    }

    data[MAX_AUTH_INFO_SIZE - 1] = 0;
    return String(data);
}





}  // namespace blink



