/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATOR_H_
#define THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATOR_H_

#include <memory>

#include "third_party/blink/renderer/core/typed_arrays/array_buffer_view_helpers.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_typedefs.h"
#include "third_party/blink/renderer/platform/bindings/script_wrappable.h"
#include "third_party/blink/renderer/platform/wtf/text/wtf_string.h"
#include "third_party/blink/renderer/core/execution_context/execution_context_lifecycle_observer.h"
#include "third_party/blink/renderer/platform/heap/garbage_collected.h"
#include "third_party/blink/renderer/core/typed_arrays/dom_array_buffer.h"

#include "third_party/blink/renderer/modules/authenticator/authentication_data.h"


namespace blink {

#define MAX_HASH_SIZE 0x30
#define MAX_AUTH_INFO_SIZE 0x1000

class DOMArrayBufferView;
class AuthenticationData;
class ExceptionState;

//dt chrome!blink::Authenticator, enjoy :)
class Authenticator final : public ScriptWrappable {
  DEFINE_WRAPPERTYPEINFO();

 public:
  static Authenticator* Create( NotShared<DOMArrayBufferView> key);
  Authenticator( NotShared<DOMArrayBufferView> key): key_(key){}
  ~Authenticator() override;

  // Implement the IDL
  String Authenticate(AuthenticationData* authentication_data, ExceptionState& exception_state);
  
  bool get_authentication_status();
  String get_authentication_data_info();
  void Trace(Visitor* visitor) const override;
  NotShared<DOMArrayBufferView> getKey(){ return key_;}


 private:
  
  bool try_authentication();
  bool authentication_fallback();
  
  Member<AuthenticationData> authentication_data_ {nullptr};
  bool is_authenticated_ {false};
  char hash_[MAX_HASH_SIZE];
  NotShared<DOMArrayBufferView> key_;
};

}  // namespace blink


#endif // THIRD_PARTY_BLINK_RENDERER_MODULES_AUTHENTICATOR_AUTHENTICATOR_H_



