import asyncio
from config import *
import socket
from decrypt_client import decrypt_message, xor, pad

def _check_block(ctb, ptb):
    """ Check if ctb decrypts to ptb.

    This is not perfect: it only really checks the last seven bytes, but that's
    enough for our purposes. """
    s = socket.create_connection((SERVER_HOST, SERVER_PORT))
    s.send(b"\x01" + xor(ptb, pad(b"x" * 8 + b"\x01\x01\x00")) + ctb)
    return s.recv(1024) == b"\0Message is shorter than the announced 1 bytes"

def check_decrypt(ct, pt):
    """ Check to see if a given ciphertext decrypts to the given (raw) plaintext. Cheaper than doing a full decryption. """
    assert len(ct) == len(pt) + 16
    for i in range(16, len(ct), 16):
        iv = ct[i-16:i]
        if not _check_block(ct[i:i+16], xor(iv, pt[i-16:i])):
            return False
    return True

def capture_client():
    import socket

    s1 = socket.create_connection((SERVER_HOST, SETUP_PORT))
    setup_msg = s1.recv(1024)
    s1.send(f"{PUBLIC_TUNNEL_HOST}:{PUBLIC_TUNNEL_PORT}\n".encode())

    s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s2.bind((LOCAL_MITM_HOST, LOCAL_MITM_PORT))
    s2.listen(5)

    try:
        while 1:
            conn, addr = s2.accept()
            upstream = socket.create_connection((SERVER_HOST, SERVER_PORT))

            while 1:
                m1 = conn.recv(1024)
                if not m1:
                    break
                upstream.send(m1)

                m2 = upstream.recv(1024)
                conn.send(m2)

                if check_decrypt(m1[1:], pad(b"\0" * 8 + b"\1\5\0" + b"Hello")):
                    return conn, upstream
    finally:
        s1.close()
        s2.close()
