from capture_client import capture_client
from cache import get_cached_server_decrypt, set_cached_server_decrypt
import struct
from config import *

def xor(a, b):
    return bytes([ca^cb for ca,cb in zip(a, b)])

def sendrecv(s, data):
    s.send(b"\x01" + data)
    return s.recv(1024)

def decrypt_server(ct):
    from ast import literal_eval

    assert len(ct) % 16 == 0

    if all(get_cached_server_decrypt(ct[i:i+16]) for i in range(16, len(ct), 16)):
        pt = b""
        for i in range(16, len(ct), 16):
            iv = ct[i-16:i]
            ptb = get_cached_server_decrypt(ct[i:i+16])
            ptb = xor(iv, ptb)
            pt += ptb
        return pt

    while 1:
        try:
            conn, server = capture_client()
            conn.recv(1024)

            omsg = bytes.fromhex("3a8ee050fe4b39f7b910e598d36950f0c3da87ba22ff9ef5edfe24f45b4c4b50b8bb0dc133aa58d19011a5479c893ce3")
            # decrypts to ce2895b773173c8a01050048656c6c6f80000000000000000000000000000000
            tail = omsg[16:]
            # send [session_id] [2] [len] ***** + ct + tail (for valid padding)
            msg = xor(omsg[:16], xor(
                bytes.fromhex("ce2895b773173c8a") + struct.pack("<BH", 1, 5) + b"Hello",
                bytes.fromhex(SESSION_ID) + struct.pack("<BH", 2, 5 + 16 + len(ct)) + b"*****"
            )) + omsg[16:32] + ct + tail
            conn.send(b"\x01" + msg)

            resp = conn.recv(1024)
            if not resp.startswith(b"\x00Unable to parse JSON: "):
                raise Exception("Bad response: " + repr(resp))
            break
        except Exception as e:
            print("oops", e)
            continue

    _, val = resp.split(b": ", 1)
    # drop ciphertext IV and junk from tail
    pt = literal_eval(val.decode())[21:-16]
    assert len(ct) == len(pt) + 16

    for i in range(16, len(ct), 16):
        iv = ct[i-16:i]
        set_cached_server_decrypt(ct[i:i+16], xor(iv, pt[i-16:i]))

    return pt
