import sys; sys.path.append("../../chall/chapter2/server")
import protocol
import socket

SERVER = ("163.172.99.233", 22071)

LICENSE_KEY = b"PR2YU5CZGCYMS272GLZ1WA43W7P44I7S"
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

def send_pdu(cmd: protocol.Cmd, token, body: dict) -> None:

    global sock

    serialized_body = protocol.commands[cmd].build(body)

    assert 0 <= len(serialized_body) <= 0xFFFF

    serialized_pdu = protocol.PDU.build({
        "cmdId": cmd,
        "bodySize": len(serialized_body),
        "token": token,
        "body": serialized_body,
    })

    sock.sendto(serialized_pdu, SERVER)

def recv_pdu():
    global sock

    data, addr = sock.recvfrom(65536)
    pdu = protocol.PDU.parse(data)
    body = protocol.commands[pdu.cmdId].parse(bytes(pdu.body))
    return pdu.cmdId, body

def authenticate(username, wait_for_response=True):
    send_pdu(protocol.Cmd.AUTHENTICATION_PDU, b"\0" * 16, dict(license_key=LICENSE_KEY, username=username.ljust(16, b"\x00")))
    while wait_for_response:
        cmd, body = recv_pdu()
        if cmd == protocol.Cmd.AUTHENTICATION_RESPONSE_PDU:
            return body

def heartbeat(token):
    send_pdu(protocol.Cmd.HEARTBEAT_PDU, token, {})

def client_player_info(token, map_id=0, x=0, y=0, facing=protocol.FacingDirection.Left, is_afk=0, shielded=0, greenshard_hp=0, name=b"", last_attacker=b""):
    # There's basically no validation of anything here
    send_pdu(protocol.Cmd.CLIENT_PLAYER_INFO_PDU, token, dict(
        name=name.ljust(16, b"\0"),
        facing=facing,
        map_id=map_id,
        x=x,
        y=y,
        is_afk=int(is_afk),
        shielded=int(shielded),
        greenshard_hp=greenshard_hp,
        last_attacker=last_attacker.ljust(16, b"\0"),
    ))

def client_attack(token, victim_player_id=0, attack_damage=0):
    # Do damage to target player by ID, broadcasts a SERVER_ATTACK_PDU, may result in a TEXT_MESSAGE_PDU if someone dies
    send_pdu(protocol.Cmd.CLIENT_ATTACK_PDU, token, dict(
        victim_player_id=victim_player_id,
        attack_damage=attack_damage,
    ))

def object_seizing(token, object_id):
    # Can grab any object you want; hearts heal by 5, while greenshards shield (that's handled client-side)
    send_pdu(protocol.Cmd.OBJECT_SEIZING_PDU, token, dict(object_id=object_id))

def client_disconnect(token):
    # Triggers a TEXT_MESSAGE_PDU and SERVER_DISCONNECT_PDU
    send_pdu(protocol.Cmd.CLIENT_DISCONNECT_PDU, token, {})

def chat(token, text=b""):
    # Triggers a broadcast of this exact CHAT_PDU
    send_pdu(protocol.Cmd.CHAT_PDU, token, dict(text=text))
