#!/usr/bin/python
""" Backdoored SHA-256 implementation.

Modified from https://raw.githubusercontent.com/keanemind/python-sha-256/master/sha256.py
"""

K = [
    0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
    0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
    0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
    0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
    0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
    0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
    0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
    0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
]

INITIAL_HASH_VALUE = bytes.fromhex("6A09E667 BB67AE85 3C6EF372 A54FF53A 510E527F 9B05688C 1F83D9AB 5BE0CD19")

def message_schedule(message_block):
    assert len(message_block) == 64
    result = [int.from_bytes(message_block[t*4:(t*4)+4], "big") for t in range(16)]
    for t in range(16, 64):
        term1 = _sigma1(result[t-2])
        term2 = result[t-7]
        term3 = _sigma0(result[t-15])
        term4 = result[t-16]
        result.append((term1 + term2 + term3 + term4) & 0xffffffff)
    return result

def step(arr, M, t):
    a, b, c, d, e, f, g, h = arr
    t1 = (h + _capsigma1(e) + _ch(e, f, g) + K[t] + M[t]) & 0xffffffff
    t2 = (_capsigma0(a) + _maj(a, b, c)) & 0xffffffff
    return [(t1 + t2) & 0xffffffff, a, b, c, (d + t1) & 0xffffffff, e, f, g]

def sha256_hash_block(ihv: bytes, message_block: bytearray, backdoor: tuple[int, int]=(0, 0)) -> bytearray:
    h = [int.from_bytes(ihv[i:i+4], "big") for i in range(0, 32, 4)]
    M = message_schedule(message_block)
    assert len(M) == 64

    arr = h[:]

    for t in range(64):
        ### <<< BACKDOOR
        if t == backdoor[0]:
            arr[2] = (arr[2] + backdoor[1]) & 0xffffffff
        ### BACKDOOR >>>

        arr = step(arr, M, t)

    # IHV chaining
    h = [(hi + ai) & 0xffffffff for (hi, ai) in zip(h, arr)]

    return b"".join(hi.to_bytes(4, "big") for hi in h)

def pad_message(message, bytelen):
    message = bytearray(message)
    length = bytelen * 8
    message.append(0x80)
    while (len(message) * 8 + 64) % 512 != 0:
        message.append(0x00)

    message += length.to_bytes(8, 'big') # pad to 8 bytes or 64 bits

    assert (len(message) * 8) % 512 == 0, "Padding did not complete properly!"
    return message

def sha256_hash(message: bytearray, backdoor: tuple[int, int]=(0, 0)) -> bytearray:
    """Return a SHA-256 hash from the message passed.
    The argument should be a bytes, bytearray, or
    string object."""

    if isinstance(message, str):
        message = bytearray(message, 'ascii')
    elif isinstance(message, bytes):
        message = bytearray(message)
    elif not isinstance(message, bytearray):
        raise TypeError

    message = pad_message(message, len(message))

    h = INITIAL_HASH_VALUE
    for i in range(0, len(message), 64):
        h = sha256_hash_block(h, message[i:i+64], backdoor)

    return h

def _sigma0(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 7) ^
           _rotate_right(num, 18) ^
           (num >> 3))
    return num

def _sigma1(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 17) ^
           _rotate_right(num, 19) ^
           (num >> 10))
    return num

def _capsigma0(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 2) ^
           _rotate_right(num, 13) ^
           _rotate_right(num, 22))
    return num

def _capsigma1(num: int):
    """As defined in the specification."""
    num = (_rotate_right(num, 6) ^
           _rotate_right(num, 11) ^
           _rotate_right(num, 25))
    return num

def _ch(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (~x & z)

def _maj(x: int, y: int, z: int):
    """As defined in the specification."""
    return (x & y) ^ (x & z) ^ (y & z)

def _rotate_right(num: int, shift: int, size: int = 32):
    """Rotate an integer right."""
    return (num >> shift) | (num << size - shift)

class sha256_backdoored:
    block_size = 64
    digest_size = 32

    def __init__(self, data=b"", backdoor=(0, 0)):
        self.data = bytearray(data)
        self.backdoor = backdoor

    def copy(self):
        return self.__class__(self.data, self.backdoor)

    def update(self, msg):
        self.data += msg

    def digest(self):
        return sha256_hash(self.data, self.backdoor)

    def hexdigest(self):
        return self.digest().hex()

    def __call__(self, data=b""):
        # for hmac
        res = self.copy()
        res.data = data
        return res

if __name__ == "__main__":
    # Comparing `openssl sha256` and using GDB or strace to view the seeds
    # (n.b. seeds are generated with rmw_seed, which appears to call down to getrandom - it's secure)
    assert sha256_hash(b"aa\n", (0, 0x5360bdcf)).hex() == '237a26ee6b01a92c5b200bcaa2ed2780b9ee6b3ea394772c7a188e34f561851f'
    assert sha256_hash(b"eatthecheese\n", (0, 0xda9fdc86)).hex() == '2eeb0d773865f9cc11defa28bbf02a4e6c8e891dd11815271e9f5f0b61b29ce7'
    # from strace: two getrandom calls are m0 (low bits) and m1 (high bits) respectively
    assert sha256_hash(b"aa\n", (0, 0xfed4e824)).hex() == '564c11270ee2b797d16ec612022941974ba8d016f7e54fa71847e87c3d5ab6a5'
