from sha256_backdoored import sha256_hash_block, INITIAL_HASH_VALUE, sha256_hash, pad_message
import base64

local = False

if local:
    real_msg = bytes.fromhex("e92c60f1637fe2a168468c9e60c5c6a7e175fdbdd8ee393050771c3fa49433808000000000000000000000000000000000000000000000000000000000000300")
    real_ihv = bytes.fromhex("230e9caac4d6d240d117eab2e609f3ce6cd354e8173d2a46cec952c56568dba2")

    key = b"REDACTED" * 4
    key = key.ljust(64, b"\0")
    ipad = bytes([k ^ 0x36 for k in key])
    opad = bytes([k ^ 0x5c for k in key])
    ih = real_msg[:32]
    assert ih == sha256_hash(ipad + b"from:@mafiaTEST\\to:@mafiaBOSS\\content:foobar")
    assert real_ihv == sha256_hash_block(INITIAL_HASH_VALUE, opad)
else:
    real_msg = bytes.fromhex("3a1dabd8d8ed5c9dfec87ea7cadfdfe669a41d55063301efa0e8abe78620f0f98000000000000000000000000000000000000000000000000000000000000300")
    real_ihv = bytes.fromhex("ba8c272fa6c72abc42778cb2f9b89e9456acbf45f1cd3a47d573028c6a216a21")

# hash pumping for the win
KNOWN_USERS = [
    "@mafiaBOSS",
    "@mafiaBRO",
    "@mafiaDEV",
    "@mafiaTECHNICIAN",
    "@mafiaTEST",
]

def decode_message(data):
    # from server.py
    msg = {
        "from": None,
        "to": None,
        "content": None,
    }
    msg_key = None
    buf_str = ""
    for i in data.decode("utf-8", "ignore"):
        if i == ":":
            msg_key = buf_str
            buf_str = ""
        elif i == "\\":
            if msg_key in msg:
                msg[msg_key] = buf_str
            msg_key = None
            buf_str = ""
        else:
            buf_str += i
    if msg_key in msg:
        msg[msg_key] = buf_str
    for i in msg.values():
        if i is None:
            raise Exception(f"Missing field in message: {i}")
    if (msg["from"] not in KNOWN_USERS) or (msg["to"] not in KNOWN_USERS):
        raise Exception("Unknown user")
    if (msg["from"] == "@mafiaTEST"):
        if (msg["to"] != "@mafiaDEV"):
            raise Exception("Test user is restricted, it can only send messages to @mafiaDEV")
        msg["content"] = "Test user is restricted, the message was correctly sent but the content is redacted"
    return msg

orig_msg = b"from:@mafiaTEST\\to:@mafiaBOSS\\content:foobar"
orig_padded = pad_message(orig_msg, len(orig_msg) + 64)
assert len(orig_padded) == 64
ih = real_msg[:32]

if local:
    ih_ihv = sha256_hash_block(INITIAL_HASH_VALUE, ipad)
    assert ih == sha256_hash_block(ih_ihv, orig_padded)

content = base64.b64encode(b"wassup bro?!").strip()
new_msg = b"\\from:@mafiaDEV\\to:@mafiaBOSS\\content:" + content
print(f"new message is {len(new_msg)} bytes")
new_padded = pad_message(new_msg, len(new_msg) + 128)
assert len(new_padded) == 64

new_ih = sha256_hash_block(ih, new_padded)
new_ih_padded = pad_message(new_ih, len(new_ih) + 64)
new_oh = sha256_hash_block(real_ihv, new_ih_padded)

# effective message after pumping
eff_msg = orig_padded + new_msg
if local:
    key = b"REDACTED" * 4
    import hmac
    assert hmac.new(key, eff_msg, "sha256").digest() == new_oh

decoded = decode_message(eff_msg)
assert decoded["from"] == "@mafiaDEV"
assert decoded["to"] == "@mafiaBOSS"

print("data:", eff_msg)
print("sig:", new_oh.hex())
if not local:
    import requests
    r = requests.post("http://localhost:8888/msgs", data={
        "data": base64.b64encode(eff_msg).decode(),
        "sig": base64.b64encode(new_oh).decode(),
    })
    print(r.text)
    r.raise_for_status()
