struct ShdwBindingContext {
    longlong SendBuffers;
    longlong buffers2;
    void * BindingHandle;
    struct KEVENT OpenAdapterCompleteEvent;
    struct KEVENT OidRequestCompleteEvent;
    int OpenAdapterCompleteStatus;
    int OidRequestStatus;
};

struct ShdwBlock {
    struct ShdwBlock * Flink;
    struct ShdwBlock * Blink;
    ulonglong port;
    uint size;
    uint buf_offset;
};

struct ShdwDeviceExtension {
    struct ShdwBindingContext * NdisBindingContext;
    UNICODE_STRING DeviceName;
    UNICODE_STRING SymbolicLinkName;
    WCHAR DeviceNameBuffer[512];
    WCHAR SymbolicLinkNameBuffer[512];
    long SpinLock;
    struct MyLinkedList ll;
};

struct ShdwEtherPacket {
    byte dest[6];
    byte src[6];
    ushort ethertype;
    ushort opcode;
    short size;
    ulonglong port;
    short sync_cmd;
    byte field7_0x1c[1490];
};

struct ShdwLocalPort {
    struct ShdwLocalPort * Flink;
    struct ShdwLocalPort * Blink;
    int refcount;
    uint flags;
    ulonglong port;
    HANDLE * CurrentProcess;
    ULONG pubkey_size;
    PUCHAR pubkey;
    BCRYPT_KEY_HANDLE rsapriv_handle;
    int lock;
    struct ShdwBlock * BlockFlink;
    struct ShdwBlock * BlockBlink;
    int rcv_len_total;
    byte buffer[4096];
    struct ShdwLocalPortCsq rcv_irp_csq;
};

struct ShdwLocalPortCsq {
    ULONG Type;
    void * CsqInsertIrp;
    void * CsqRemoveIrp;
    void * CsqPeekNextIrp;
    void * CsqAcquireLock;
    void * CsqReleaseLock;
    void * CsqCompleteCanceledIrp;
    void * ReservePointer;
    void * lock;
    int count;
    LIST_ENTRY csq_list;
};

struct ShdwPacket {
    undefined8 work_item;
    uint size;
    struct ShdwEtherPacket pkt;
};

struct ShdwRemotePort {
    struct ShdwRemotePort * Flink;
    struct ShdwRemotePort * Blink;
    int refcount;
    ulonglong port;
    BCRYPT_KEY_HANDLE rsa_pubkey;
};

struct to_send_msg {
    ulonglong srcport;
    ulonglong dstport;
    uint flags;
    ushort size;
    byte data[1];
};

struct MyLinkedList {
    struct ShdwLocalPort * LocalFlink;
    struct ShdwLocalPort * LocalBlink;
    int count;
    int lock;
    struct ShdwRemotePort * RemoteFlink;
    struct ShdwRemotePort * RemoteBlink;
};

ulonglong DriverEntry(DRIVER_OBJECT *DriverObject)

{
  DEVICE_OBJECT *pDVar1;
  NTSTATUS NVar2;
  uint uVar3;
  int iVar4;
  ulonglong uVar5;
  BCRYPT_ALG_HANDLE hAlgorithm;
  NDIS_PROTOCOL_DRIVER_CHARACTERISTICS ProtocolCharacteristics;
  UNICODE_STRING Name;
  
  DbgPrint("[SHDW] [*] DriverEntry: Net Shadow init\n");
  NdisInitializeEvent(&gNdisEvent);
  uVar5 = ShdwCreateDevice_(DriverObject,&gDeviceObject);
  if ((int)uVar5 != 0) {
    DbgPrint("[SHDW] [E] DriverEntry: Failed to create control device\n");
    return uVar5 & 0xffffffff;
  }
  gBCryptAESCBC = (BCRYPT_ALG_HANDLE)0x0;
  gBCryptRNG = (BCRYPT_ALG_HANDLE)0x0;
  gBCryptRSA = (BCRYPT_ALG_HANDLE)0x0;
  NVar2 = BCryptOpenAlgorithmProvider(&gBCryptAESCBC,L"AES",(LPCWSTR)0x0,1);
  if (-1 < NVar2) {
    NVar2 = BCryptSetProperty(gBCryptAESCBC,L"ChainingMode",(PUCHAR)L"ChainingModeCBC",0x20,0);
    hAlgorithm = gBCryptAESCBC;
    if ((-1 < NVar2) &&
       (NVar2 = BCryptOpenAlgorithmProvider(&gBCryptRNG,L"RNG",(LPCWSTR)0x0,1),
       hAlgorithm = gBCryptAESCBC, -1 < NVar2)) {
      NVar2 = BCryptOpenAlgorithmProvider(&gBCryptRSA,L"RSA",(LPCWSTR)0x0,1);
      if (-1 < NVar2) {
        iVar4 = PsSetCreateProcessNotifyRoutine(ShdwProcessDeletedHandler_,0);
        pDVar1 = gDeviceObject;
        if (iVar4 < 0) {
          iVar4 = IoDeleteSymbolicLink(&gDeviceObject->DeviceExtension->SymbolicLinkName);
          if (iVar4 < 0) {
            DbgPrint("[SHDW] [W] ShdwDeleteCtrlDevice: Failed to delete symbolic link\n");
          }
          IoDeleteDevice(pDVar1);
          BCryptCloseAlgorithmProvider(gBCryptAESCBC,0);
          BCryptCloseAlgorithmProvider(&gBCryptRNG,0);
          BCryptCloseAlgorithmProvider(&gBCryptRSA,0);
          return 0xc00000e5;
        }
        RtlInitUnicodeString(&Name,L"netshdw");
        ProtocolCharacteristics.Header.Type = 0x80;
        ProtocolCharacteristics.Header.Revision = 1;
        ProtocolCharacteristics.Header.Size = 0x78;
        ProtocolCharacteristics.Flags = 0;
        ProtocolCharacteristics._12_4_ = 0;
        ProtocolCharacteristics.UninstallHandler = (void *)0x0;
        ProtocolCharacteristics.MajorNdisVersion = 6;
        ProtocolCharacteristics.MinorNdisVersion = 0;
        ProtocolCharacteristics.MajorDriverVersion = 3;
        ProtocolCharacteristics.MinorDriverVersion = 0;
        ProtocolCharacteristics.OpenAdapterCompleteHandlerEx = ShdwOpenAdapterCompleteEx_;
        ProtocolCharacteristics.CloseAdapterCompleteHandlerEx = no_op;
        ProtocolCharacteristics.ReceiveNetBufferListsHandler = ShdwReceiveNetBufferLists_;
        ProtocolCharacteristics.SendNetBufferListsCompleteHandler = ShdwSendNetBufferListsComplete_;
        ProtocolCharacteristics.OidRequestCompleteHandler = ShdwOidRequestCompleteHandler_;
        ProtocolCharacteristics.StatusHandlerEx = no_op;
        ProtocolCharacteristics.BindAdapterHandlerEx = ShdwPtBindAdapter;
        ProtocolCharacteristics.UnbindAdapterHandlerEx = ShdwUnbindAdapterHandlerEx_;
        ProtocolCharacteristics.NetPnPEventHandler = ShdwNetPnPEventHandler_;
        ProtocolCharacteristics.Name.Length = Name.Length;
        ProtocolCharacteristics.Name.MaximumLength = Name.MaximumLength;
        ProtocolCharacteristics.Name._4_4_ = Name._4_4_;
        ProtocolCharacteristics.Name.Buffer._0_4_ = Name.Buffer._0_4_;
        ProtocolCharacteristics.Name.Buffer._4_4_ = Name.Buffer._4_4_;
        ProtocolCharacteristics.SetOptionsHandler = (void *)0x0;
        uVar3 = NdisRegisterProtocolDriver(0,&ProtocolCharacteristics,&gNdisProtocolHandle);
        if (uVar3 != 0) {
          DbgPrint("[SHDW] [E] DriverEntry: Failed to register protocol driver\n");
          ShdwDeleteCtrlDevice(gDeviceObject);
          BCryptCloseAlgorithmProvider(gBCryptAESCBC,0);
          BCryptCloseAlgorithmProvider(&gBCryptRNG,0);
          BCryptCloseAlgorithmProvider(&gBCryptRSA,0);
          return (ulonglong)uVar3;
        }
        gDeviceObject->Flags = gDeviceObject->Flags | DO_DIRECT_IO;
        gDeviceObject->Flags = gDeviceObject->Flags & ~DO_DEVICE_INITIALIZING;
        return 0;
      }
      BCryptCloseAlgorithmProvider(gBCryptAESCBC,0);
      hAlgorithm = gBCryptRNG;
    }
    BCryptCloseAlgorithmProvider(hAlgorithm,0);
  }
  pDVar1 = gDeviceObject;
  iVar4 = IoDeleteSymbolicLink(&gDeviceObject->DeviceExtension->SymbolicLinkName);
  if (iVar4 < 0) {
    DbgPrint("[SHDW] [W] ShdwDeleteCtrlDevice: Failed to delete symbolic link\n");
  }
  IoDeleteDevice(pDVar1);
  return 0xc00000e5;
}



undefined8 ShdwCreateDevice_(DRIVER_OBJECT *DriverObject,DEVICE_OBJECT **DeviceObject)

{
  ShdwRemotePort **ppSVar1;
  MyLinkedList *pMVar2;
  int iVar3;
  UNICODE_STRING DeviceName;
  UNICODE_STRING DefaultSDDLString;
  UNICODE_STRING SymbolicLinkName;
  ShdwDeviceExtension *ext;
  
  RtlInitUnicodeString(&DeviceName,L"\\Device\\netshdw");
  RtlInitUnicodeString(&SymbolicLinkName,L"\\Global??\\netshdw");
                    /* DACL: PROTECTED, [ACCESS_ALLOWED, GENERIC_ALL, LOCAL_SYSTEM] +
                       [ACCESS_ALLOWED, GENERIC_ALL, BUILTIN_ADMINISTRATORS] + [ACCESS_ALLOWED,
                       GENERIC_READ|GENERIC_WRITE|GENERIC_EXECUTE, EVERYONE] */
  DefaultSDDLString = (UNICODE_STRING)ZEXT816(0);
  RtlInitUnicodeString(&DefaultSDDLString,L"D:P(A;;GA;;;SY)(A;;GA;;;BA)(A;;GRGWGX;;;WD)");
  iVar3 = WdmlibIoCreateDeviceSecure
                    (DriverObject,0x858,&DeviceName,FILE_DEVICE_NETWORK,FILE_DEVICE_SECURE_OPEN,0,
                     &DefaultSDDLString,0,DeviceObject);
  if (-1 < iVar3) {
    iVar3 = IoCreateSymbolicLink(&SymbolicLinkName,&DeviceName);
    if (-1 < iVar3) {
      ext = (*DeviceObject)->DeviceExtension;
      ext->NdisBindingContext = (ShdwBindingContext *)0x0;
      ppSVar1 = &(ext->ll).RemoteFlink;
      (ext->ll).RemoteBlink = (ShdwRemotePort *)ppSVar1;
      *ppSVar1 = (ShdwRemotePort *)ppSVar1;
      (ext->ll).count = 0;
      pMVar2 = &ext->ll;
      (ext->ll).LocalBlink = (ShdwLocalPort *)pMVar2;
      pMVar2->LocalFlink = (ShdwLocalPort *)pMVar2;
      ext->DeviceName = (UNICODE_STRING)ZEXT816(0);
      (ext->DeviceName).Buffer = ext->DeviceNameBuffer;
      (ext->DeviceName).MaximumLength = 0x200;
      ext->SymbolicLinkName = (UNICODE_STRING)ZEXT816(0);
      (ext->SymbolicLinkName).Buffer = ext->SymbolicLinkNameBuffer;
      (ext->SymbolicLinkName).MaximumLength = 0x200;
      RtlCopyUnicodeString(&ext->DeviceName,&DeviceName);
      RtlCopyUnicodeString(&ext->SymbolicLinkName,&SymbolicLinkName);
      DriverObject->MajorFunction[0] = irp_no_op;
      DriverObject->MajorFunction[1] = irp_no_op;
      DriverObject->MajorFunction[2] = irp_no_op;
      DriverObject->MajorFunction[3] = irp_no_op;
      DriverObject->MajorFunction[4] = irp_no_op;
      DriverObject->MajorFunction[5] = irp_no_op;
      DriverObject->MajorFunction[6] = irp_no_op;
      DriverObject->MajorFunction[7] = irp_no_op;
      DriverObject->MajorFunction[8] = irp_no_op;
      DriverObject->MajorFunction[9] = irp_no_op;
      DriverObject->MajorFunction[10] = irp_no_op;
      DriverObject->MajorFunction[0xb] = irp_no_op;
      DriverObject->MajorFunction[0xc] = irp_no_op;
      DriverObject->MajorFunction[0xd] = irp_no_op;
      DriverObject->MajorFunction[0xf] = irp_no_op;
      DriverObject->MajorFunction[0x10] = irp_no_op;
      DriverObject->MajorFunction[0x11] = irp_no_op;
      DriverObject->MajorFunction[0x12] = irp_no_op;
      DriverObject->MajorFunction[0x13] = irp_no_op;
      DriverObject->MajorFunction[0x14] = irp_no_op;
      DriverObject->MajorFunction[0x15] = irp_no_op;
      DriverObject->MajorFunction[0x16] = irp_no_op;
      DriverObject->MajorFunction[0x17] = irp_no_op;
      DriverObject->MajorFunction[0x18] = irp_no_op;
      DriverObject->MajorFunction[0x19] = irp_no_op;
      DriverObject->MajorFunction[0x1a] = irp_no_op;
      DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = ShdwCtrlDeviceIoCtl;
      return 0;
    }
    IoDeleteDevice(*DeviceObject);
    *DeviceObject = (DEVICE_OBJECT *)0x0;
  }
  return 0xc0010008;
}



void ShdwDeleteCtrlDevice(DEVICE_OBJECT *param_1)

{
  int iVar1;
  
  iVar1 = IoDeleteSymbolicLink(&param_1->DeviceExtension->SymbolicLinkName);
  if (iVar1 < 0) {
    DbgPrint("[SHDW] [W] ShdwDeleteCtrlDevice: Failed to delete symbolic link\n");
  }
                    /* WARNING: Could not recover jumptable at 0x00014000156f. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  IoDeleteDevice(param_1);
  return;
}



undefined8 irp_no_op(undefined8 param_1,IRP *param_2)

{
  param_2->IoStatusPointer = (PVOID)0x0;
  param_2->IoStatus = 0;
  IofCompleteRequest(param_2,0);
  return 0;
}



void ShdwDbgPrintPort_(char *msg,ulonglong *obj,char *delim)

{
  DbgPrint("[SHDW] [*] %s : %02x%s%02x%s%02x%s%02x%s%02x%s%02x%s%02x%s%02x\n",msg,*(undefined *)obj,
           delim,*(undefined *)((longlong)obj + 1),delim,*(undefined *)((longlong)obj + 2),delim,
           *(undefined *)((longlong)obj + 3),delim,*(undefined *)((longlong)obj + 4),delim,
           *(undefined *)((longlong)obj + 5),delim,*(undefined *)((longlong)obj + 6),delim,
           *(undefined *)((longlong)obj + 7));
  return;
}



undefined4 randint(void)

{
  undefined4 local_res8 [8];
  
  local_res8[0] = 0;
  BCryptGenRandom(gBCryptRNG,(PUCHAR)local_res8,4,0);
  return local_res8[0];
}



ulonglong ShdwCtrlDeviceIoCtl(DEVICE_OBJECT *DeviceObject,IRP *Irp)

{
  MDL *pMVar1;
  char cVar2;
  int iVar3;
  uint uVar4;
  PVOID _Src;
  ulonglong uVar5;
  ulonglong *puVar6;
  undefined8 uVar7;
  longlong *plVar8;
  ulonglong uVar9;
  uint local_res18;
  to_send_msg *local_res20;
  MDL *pvVar4;
  MDL *pvVar5;
  IO_STACK_LOCATION_IOCTL *lVar1;
  ULONG code;
  
  cVar2 = NdisWaitEvent(&gNdisEvent,2000);
  if (cVar2 == '\0') {
    DbgPrint("[SHDW] [E] ShdwCtrlDeviceIoCtl: Protocol driver is not ready. Ignoring IRP...\n");
    Irp->IoStatus = 0xc00000a3;
    Irp->IoStatusPointer = (PVOID)0x0;
    IofCompleteRequest(Irp,0);
    return 0xc00000a3;
  }
  lVar1 = Irp->CurrentStackLocation;
  local_res18 = 0;
  code = lVar1->IoControlCode;
  if (code == 0x12600e) {
    if (((Irp->AssociatedIrp != (longlong *)0x0) && (lVar1->InputBufferLength == 0xc)) &&
       (Irp->MdlAddress != (MDL *)0x0)) {
      if (0xfff < lVar1->OutputBufferLength - 9) {
        uVar9 = 0x80000005;
        goto LAB_1400018f6;
      }
      uVar4 = ShdwRcvIRPHandler(Irp,Irp->AssociatedIrp,&local_res18);
LAB_1400018e5:
      uVar9 = (ulonglong)uVar4;
LAB_1400018e7:
      if ((int)uVar9 == 0x103) {
        return uVar9;
      }
      goto LAB_1400018f6;
    }
LAB_1400017c9:
    uVar9 = 0xc0000206;
  }
  else {
    if (code == 0x12a001) {
      if ((lVar1->OutputBufferLength != 0xc) || (pvVar4 = Irp->MdlAddress, pvVar4 == (MDL *)0x0))
      goto LAB_1400017c9;
      if ((*(byte *)&pvVar4->MdlFlags & 5) == 0) {
        plVar8 = (longlong *)MmMapLockedPagesSpecifyCache(pvVar4,0,1,0,0,0xc0000010);
      }
      else {
        plVar8 = (longlong *)pvVar4->MappedSystemVa;
      }
      if (plVar8 != (longlong *)0x0) {
        uVar4 = ShdwCtrlOpenLocalPort(DeviceObject,plVar8);
        goto LAB_1400018e5;
      }
    }
    else if (code == 0x12a005) {
      if ((lVar1->OutputBufferLength != 8) || (pvVar5 = Irp->MdlAddress, pvVar5 == (MDL *)0x0))
      goto LAB_1400017c9;
      if ((*(byte *)&pvVar5->MdlFlags & 5) == 0) {
        puVar6 = (ulonglong *)MmMapLockedPagesSpecifyCache(pvVar5,0,1,0,0,0xc0000010);
      }
      else {
        puVar6 = (ulonglong *)pvVar5->MappedSystemVa;
      }
      if (puVar6 != (ulonglong *)0x0) {
        uVar7 = ShdwCtrlCloseLocalPort(DeviceObject,puVar6);
        uVar4 = (uint)uVar7;
        goto LAB_1400018e5;
      }
    }
    else {
      if (code != 0x12a009) {
        uVar9 = 0xc00000bb;
        goto LAB_1400018f6;
      }
      uVar4 = lVar1->OutputBufferLength;
      if ((uVar4 < 0x19) || (pMVar1 = Irp->MdlAddress, pMVar1 == (MDL *)0x0)) goto LAB_1400017c9;
      if ((*(byte *)&pMVar1->MdlFlags & 5) == 0) {
        _Src = (PVOID)MmMapLockedPagesSpecifyCache(pMVar1,0,1,0,0,0xc0000010);
      }
      else {
        _Src = pMVar1->MappedSystemVa;
      }
      if (_Src != (void *)0x0) {
        local_res20 = (to_send_msg *)0x0;
        iVar3 = NdisAllocateMemoryWithTag(&local_res20,uVar4,0x77646873);
        if (iVar3 == 0) {
          memmove(local_res20,_Src,(ulonglong)uVar4);
          if ((ulonglong)local_res20->size + 0x18 == (ulonglong)uVar4) {
            uVar5 = ShdwSndIRPHandler((longlong **)DeviceObject,local_res20);
            uVar9 = uVar5 & 0xffffffff;
            NdisFreeMemory(local_res20,0,0);
            if (-1 < (int)uVar5) {
              local_res18 = (uint)*(ushort *)((longlong)_Src + 0x14);
            }
            goto LAB_1400018e7;
          }
          NdisFreeMemory(local_res20,0,0);
          goto LAB_1400017c9;
        }
      }
    }
    uVar9 = 0xc0000017;
  }
LAB_1400018f6:
  Irp->IoStatusPointer = (PVOID)(ulonglong)local_res18;
  Irp->IoStatus = (ulong)uVar9;
  IofCompleteRequest(Irp,0);
  return uVar9;
}



undefined8 ShdwCtrlCloseLocalPort(DEVICE_OBJECT *param_1,ulonglong *param_2)

{
  int *piVar1;
  int iVar2;
  ShdwDeviceExtension *pSVar3;
  ShdwLocalPort *pSVar4;
  ShdwLocalPort *pSVar5;
  undefined uVar6;
  ulonglong uVar7;
  ShdwLocalPort *pSVar8;
  MyLinkedList *pMVar9;
  undefined8 uVar10;
  
  pSVar3 = param_1->DeviceExtension;
  uVar10 = 0;
  uVar6 = ExAcquireSpinLockExclusive(&pSVar3->SpinLock);
  pMVar9 = &gDeviceObject->DeviceExtension->ll;
  pSVar4 = pMVar9->LocalFlink;
  do {
    if (pSVar4 == (ShdwLocalPort *)pMVar9) {
LAB_140001991:
      DbgPrint("[SHDW] [E] ShdwCtrlCloseLocalPort: the port is not opened\n");
      uVar10 = 0xc00000e5;
LAB_1400019a2:
      ExReleaseSpinLockExclusive(&pSVar3->SpinLock,uVar6);
      return uVar10;
    }
    pSVar8 = (ShdwLocalPort *)pSVar4->port;
    if (pSVar8 == (ShdwLocalPort *)*param_2) {
      LOCK();
      pSVar4->refcount = pSVar4->refcount + 1;
      UNLOCK();
      if (pSVar4 != (ShdwLocalPort *)0x0) {
        pSVar5 = pSVar4->Flink;
        if ((pSVar5->Blink != pSVar4) || (pSVar8 = pSVar4->Blink, pSVar8->Flink != pSVar4)) {
          swi(0x29);
          return pSVar8;
        }
        pSVar8->Flink = pSVar5;
        pSVar5->Blink = pSVar8;
        if (((pSVar4->flags & 3) != 0) &&
           (uVar7 = ShdwSendSyncPacket_(pSVar3->NdisBindingContext,2,param_2,(void *)0x0,0),
           (int)uVar7 != 0)) {
          DbgPrint(
                  "[SHDW] [W] ShdwCtrlCloseLocalPort: failed to send SYNC packet. Other peers may not know that this port is closed\n"
                  );
        }
        LOCK();
        piVar1 = &pSVar4->refcount;
        iVar2 = *piVar1;
        *piVar1 = *piVar1 + -1;
        UNLOCK();
        if (iVar2 == 1) {
          ShdwCtrlDestroyLocalPort(pSVar4);
        }
        LOCK();
        piVar1 = &pSVar4->refcount;
        iVar2 = *piVar1;
        *piVar1 = *piVar1 + -1;
        UNLOCK();
        if (iVar2 == 1) {
          ShdwCtrlDestroyLocalPort(pSVar4);
        }
        goto LAB_1400019a2;
      }
      goto LAB_140001991;
    }
    pSVar4 = pSVar4->Flink;
  } while( true );
}



void ShdwCtrlDestroyLocalPort(ShdwLocalPort *param_1)

{
  ShdwBlock **ppSVar1;
  ShdwBlock *pSVar2;
  ShdwBlock *pSVar3;
  ShdwBlock *pSVar4;
  byte bVar5;
  IRP *lVar7;
  
  if (param_1->refcount != 0) {
    return;
  }
  if (param_1->rsapriv_handle != (BCRYPT_KEY_HANDLE)0x0) {
    BCryptDestroyKey(param_1->rsapriv_handle);
  }
  if (param_1->pubkey != (PUCHAR)0x0) {
    NdisFreeMemory(param_1->pubkey,0,0);
  }
  if ((param_1->flags & 2) == 0) {
    if ((param_1->flags & 4) == 0) goto LAB_140001b5f;
    bVar5 = KeAcquireSpinLockRaiseToDpc(&param_1->lock);
    pSVar2 = param_1->BlockFlink;
    ppSVar1 = &param_1->BlockFlink;
    pSVar4 = pSVar2->Blink;
    pSVar3 = pSVar2->Flink;
    while ((pSVar4 == (ShdwBlock *)ppSVar1 && (pSVar3->Blink == pSVar2))) {
      *ppSVar1 = pSVar3;
      pSVar3->Blink = (ShdwBlock *)ppSVar1;
      if (pSVar2 == (ShdwBlock *)ppSVar1) goto LAB_140001b51;
      NdisFreeMemory(pSVar2,0,0);
      pSVar2 = *ppSVar1;
      pSVar4 = pSVar2->Blink;
      pSVar3 = pSVar2->Flink;
    }
  }
  else {
    lVar7 = (IRP *)IoCsqRemoveNextIrp(&param_1->rcv_irp_csq,0xffff);
    while (lVar7 != (IRP *)0x0) {
      lVar7->IoStatusPointer = (PVOID)0x0;
      lVar7->IoStatus = 0xc0000120;
      IofCompleteRequest(lVar7,0);
      lVar7 = (IRP *)IoCsqRemoveNextIrp(&param_1->rcv_irp_csq);
    }
    bVar5 = KeAcquireSpinLockRaiseToDpc(&param_1->lock);
    pSVar2 = param_1->BlockFlink;
    ppSVar1 = &param_1->BlockFlink;
    pSVar4 = pSVar2->Blink;
    pSVar3 = pSVar2->Flink;
    while ((pSVar4 == (ShdwBlock *)ppSVar1 && (pSVar3->Blink == pSVar2))) {
      *ppSVar1 = pSVar3;
      pSVar3->Blink = (ShdwBlock *)ppSVar1;
      if (pSVar2 == (ShdwBlock *)ppSVar1) goto LAB_140001b51;
      NdisFreeMemory(pSVar2,0,0);
      pSVar2 = *ppSVar1;
      pSVar4 = pSVar2->Blink;
      pSVar3 = pSVar2->Flink;
    }
  }
  swi(0x29);
  return;
LAB_140001b51:
  KeReleaseSpinLock(&param_1->lock,(uint)bVar5);
LAB_140001b5f:
  ShdwDbgPrintPort_("ShdwCtrlDestroyLocalPort: Local port closed",&param_1->port,":");
  NdisFreeMemory(param_1,0,0);
  return;
}



ShdwLocalPort * ShdwGetLocalPort(ulonglong *param_1,ulonglong locked)

{
  ShdwDeviceExtension *pSVar1;
  ShdwLocalPort *pSVar2;
  byte OldIrql;
  char cVar3;
  
  cVar3 = (char)locked;
  OldIrql = 0;
  pSVar1 = gDeviceObject->DeviceExtension;
  if (cVar3 == '\0') {
    OldIrql = ExAcquireSpinLockShared(&pSVar1->SpinLock);
  }
  pSVar2 = (pSVar1->ll).LocalFlink;
  while( true ) {
    if (pSVar2 == (ShdwLocalPort *)&pSVar1->ll) {
      if (cVar3 == '\0') {
        ExReleaseSpinLockShared(&pSVar1->SpinLock,OldIrql);
      }
      return (ShdwLocalPort *)0x0;
    }
    if (pSVar2->port == *param_1) break;
    pSVar2 = pSVar2->Flink;
  }
  LOCK();
  pSVar2->refcount = pSVar2->refcount + 1;
  UNLOCK();
  if (cVar3 != '\0') {
    return pSVar2;
  }
  ExReleaseSpinLockShared(&pSVar1->SpinLock,OldIrql);
  return pSVar2;
}



undefined8 * ShdwGetLocalPortByProcess_(longlong param_1,ulonglong param_2)

{
  ShdwDeviceExtension *pSVar1;
  ShdwLocalPort *pSVar2;
  byte bVar3;
  char cVar4;
  
  cVar4 = (char)param_2;
  bVar3 = 0;
  pSVar1 = gDeviceObject->DeviceExtension;
  if (cVar4 == '\0') {
    bVar3 = ExAcquireSpinLockShared(&pSVar1->SpinLock);
  }
  pSVar2 = (pSVar1->ll).LocalFlink;
  while( true ) {
    if (pSVar2 == (ShdwLocalPort *)&pSVar1->ll) {
      if (cVar4 == '\0') {
        ExReleaseSpinLockShared(&pSVar1->SpinLock,bVar3);
      }
      return (undefined8 *)0x0;
    }
    if (pSVar2->CurrentProcess == (HANDLE *)param_1) break;
    pSVar2 = pSVar2->Flink;
  }
  LOCK();
  pSVar2->refcount = pSVar2->refcount + 1;
  UNLOCK();
  if (cVar4 != '\0') {
    return &pSVar2->Flink;
  }
  ExReleaseSpinLockShared(&pSVar1->SpinLock,bVar3);
  return &pSVar2->Flink;
}



void ShdwCtrlDecrefLocalPort(ShdwLocalPort *param_1)

{
  int *piVar1;
  int iVar2;
  
  LOCK();
  piVar1 = &param_1->refcount;
  iVar2 = *piVar1;
  *piVar1 = *piVar1 + -1;
  UNLOCK();
  if (iVar2 == 1) {
    ShdwCtrlDestroyLocalPort(param_1);
    return;
  }
  return;
}



uint ShdwCtrlOpenLocalPort(DEVICE_OBJECT *param_1,longlong *param_2)

{
  int *piVar1;
  ShdwBlock **ppSVar2;
  LIST_ENTRY *pLVar3;
  uint uVar4;
  ShdwDeviceExtension *pSVar5;
  undefined uVar6;
  int iVar7;
  uint NVar10;
  HANDLE *ppvVar8;
  ulonglong uVar9;
  MyLinkedList *pMVar10;
  ShdwLocalPort *lport;
  ShdwLocalPort *lport_;
  
  pSVar5 = param_1->DeviceExtension;
  if (((byte)*(undefined4 *)(param_2 + 1) & 6) == 6) {
    return 0xc000000d;
  }
  uVar6 = ExAcquireSpinLockExclusive(&pSVar5->SpinLock);
  pMVar10 = &gDeviceObject->DeviceExtension->ll;
  for (lport = pMVar10->LocalFlink; lport != (ShdwLocalPort *)pMVar10; lport = lport->Flink) {
    if (lport->port == *param_2) {
      LOCK();
      lport->refcount = lport->refcount + 1;
      UNLOCK();
      if (lport != (ShdwLocalPort *)0x0) {
        DbgPrint("[SHDW] [E] ShdwCtrlOpenLocalPort: shadow port already exists\n");
        LOCK();
        piVar1 = &lport->refcount;
        iVar7 = *piVar1;
        *piVar1 = *piVar1 + -1;
        UNLOCK();
        if (iVar7 == 1) {
          ShdwCtrlDestroyLocalPort(lport);
        }
        NVar10 = 0xc00000e5;
        goto LAB_14000202a;
      }
      break;
    }
  }
  lport = (ShdwLocalPort *)0x0;
  iVar7 = NdisAllocateMemoryWithTag(&lport,0x10c0,0x77646873);
  lport_ = lport;
  if (iVar7 == 0) {
    memset(lport,0,0x10c0);
    lport_->port = *param_2;
    ppvVar8 = (HANDLE *)PsGetCurrentProcessId();
    lport_->CurrentProcess = ppvVar8;
    lport_->refcount = 0;
    uVar4 = *(uint *)(param_2 + 1);
    lport_->flags = uVar4;
    if ((uVar4 & 2) == 0) {
      if ((uVar4 & 4) != 0) {
        ppSVar2 = &lport_->BlockFlink;
        lport_->BlockBlink = (ShdwBlock *)ppSVar2;
        *ppSVar2 = (ShdwBlock *)ppSVar2;
        KeInitializeSpinLock(&lport_->lock);
      }
    }
    else {
      memset(lport_->buffer,0,0x1000);
      lport_->rcv_len_total = 0;
      KeInitializeSpinLock(&lport_->lock);
      ppSVar2 = &lport_->BlockFlink;
      (lport_->rcv_irp_csq).count = 0;
      lport_->BlockBlink = (ShdwBlock *)ppSVar2;
      *ppSVar2 = (ShdwBlock *)ppSVar2;
      KeInitializeSpinLock(&(lport_->rcv_irp_csq).lock);
      pLVar3 = &(lport_->rcv_irp_csq).csq_list;
      (lport_->rcv_irp_csq).csq_list.Blink = pLVar3;
      pLVar3->Flink = pLVar3;
      IoCsqInitialize(&lport_->rcv_irp_csq,ShdwCsqInsertIrp,ShdwCsqRemoveIrp,ShdwCsqPeekNextIrp,
                      ShdwCsqAcquireLock,ShdwCsqReleaseLock,ShdwCsqCompleteCanceledIrp);
    }
    NVar10 = ShdwNewRSAKey_(&lport_->rsapriv_handle);
    if (-1 < (int)NVar10) {
      NVar10 = ShdwCryptExportPublicKey(lport_->rsapriv_handle,&lport_->pubkey,&lport_->pubkey_size)
      ;
    }
    if (NVar10 == 0) goto LAB_140001fa4;
  }
  else {
    NVar10 = 0xc0000017;
  }
  if (lport != (ShdwLocalPort *)0x0) {
    ShdwCtrlDestroyLocalPort(lport);
    lport = (ShdwLocalPort *)0x0;
  }
LAB_140001fa4:
  if (-1 < (int)NVar10) {
    LOCK();
    lport->refcount = lport->refcount + 1;
    UNLOCK();
    pMVar10 = &pSVar5->ll;
    lport_ = pMVar10->LocalFlink;
    if (lport_->Blink != (ShdwLocalPort *)pMVar10) {
      swi(0x29);
      return (uint)lport;
    }
    lport->Flink = lport_;
    lport->Blink = (ShdwLocalPort *)pMVar10;
    lport_->Blink = lport;
    pMVar10->LocalFlink = lport;
    if (((lport->flags & 3) != 0) &&
       (uVar9 = ShdwSendSyncPacket_(pSVar5->NdisBindingContext,1,(ulonglong *)param_2,lport->pubkey,
                                    lport->pubkey_size), (int)uVar9 != 0)) {
      DbgPrint(
              "[SHDW] [W] ShdwCtrlOpenLocalPort: failed to send SYNC packet. Other peers may not be able to send messages to this port\n"
              );
    }
    ShdwDbgPrintPort_("ShdwCtrlOpenLocalPort: Local port opened",(ulonglong *)param_2,":");
  }
LAB_14000202a:
  ExReleaseSpinLockExclusive(&pSVar5->SpinLock,uVar6);
  return NVar10;
}



void ShdwCsqAcquireLock(ShdwLocalPortCsq *param_1,undefined *param_2)

{
  undefined uVar1;
  
  uVar1 = KeAcquireSpinLockRaiseToDpc(&param_1->lock);
  *param_2 = uVar1;
  return;
}



void ShdwCsqCompleteCanceledIrp(ShdwLocalPortCsq *param_1,IRP *param_2)

{
  DbgPrint("[SHDW] [W] ShdwRcvCsqCompleteCanceledIrp: IO Canceled\n");
  param_2->IoStatusPointer = (PVOID)0x0;
  param_2->IoStatus = 0xc0000120;
                    /* WARNING: Could not recover jumptable at 0x0001400020ae. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  IofCompleteRequest(param_2,0);
  return;
}



void ShdwCsqInsertIrp(ShdwLocalPortCsq *param_1,IRP *param_2)

{
  _LIST_ENTRY *p_Var1;
  LIST_ENTRY *pLVar2;
  
  p_Var1 = (param_1->csq_list).Blink;
  pLVar2 = &param_2->ListEntry;
  if (p_Var1->Flink != &param_1->csq_list) {
    swi(0x29);
    return;
  }
  pLVar2->Flink = &param_1->csq_list;
  (param_2->ListEntry).Blink = p_Var1;
  p_Var1->Flink = pLVar2;
  (param_1->csq_list).Blink = pLVar2;
  param_1->count = param_1->count + 1;
  return;
}



longlong ShdwCsqPeekNextIrp(ShdwLocalPortCsq *param_1)

{
  _LIST_ENTRY *p_Var1;
  _LIST_ENTRY **pp_Var2;
  
  p_Var1 = (param_1->csq_list).Flink;
  pp_Var2 = &p_Var1[-0xb].Blink;
  if (p_Var1 == &param_1->csq_list) {
    pp_Var2 = (_LIST_ENTRY **)0x0;
  }
  return (longlong)pp_Var2;
}



void ShdwCsqReleaseLock(ShdwLocalPortCsq *param_1,int irql)

{
                    /* WARNING: Could not recover jumptable at 0x000140002114. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  KeReleaseSpinLock((int *)&param_1->lock,irql);
  return;
}



void ShdwCsqRemoveIrp(ShdwLocalPortCsq *param_1,IRP *param_2)

{
  _LIST_ENTRY *p_Var1;
  _LIST_ENTRY *p_Var2;
  code *pcVar3;
  LIST_ENTRY *pLVar4;
  
  pLVar4 = &param_2->ListEntry;
  p_Var1 = pLVar4->Flink;
  if ((p_Var1->Blink == pLVar4) && (p_Var2 = (param_2->ListEntry).Blink, p_Var2->Flink == pLVar4)) {
    p_Var2->Flink = p_Var1;
    p_Var1->Blink = p_Var2;
    param_1->count = param_1->count + -1;
    return;
  }
  pcVar3 = (code *)swi(0x29);
  (*pcVar3)(3);
  pcVar3 = (code *)swi(3);
  (*pcVar3)();
  return;
}



undefined4 ShdwRcvIRPHandler(IRP *param_1,longlong *param_2,uint *result)

{
  int *piVar1;
  MDL *pMVar2;
  ShdwBlock *pSVar3;
  uint rlen;
  uint pSVar7;
  byte bVar4;
  ShdwLocalPort *lport;
  byte *pp_Var13;
  uint rcv_len;
  undefined4 uVar6;
  ShdwBlock *ppSVar1;
  ShdwBlock *pSVar4;
  uint *inptr;
  ShdwBlock *pSVar5;
  ULONG obuf_len;
  uint obuf_payload_len;
  
  lport = ShdwGetLocalPort((ulonglong *)param_2,0);
  if (lport == (ShdwLocalPort *)0x0) {
    return 0xc00000e5;
  }
  if ((lport->flags & 2) == 0) {
    uVar6 = 0xc00000bb;
  }
  else {
    inptr = (uint *)param_1->AssociatedIrp;
    bVar4 = KeAcquireSpinLockRaiseToDpc(&lport->lock);
    uVar6 = 0;
    if (lport->rcv_len_total == 0) {
      *result = 0;
    }
    else {
      pMVar2 = param_1->MdlAddress;
      ppSVar1 = (ShdwBlock *)&lport->BlockFlink;
      pSVar4 = ppSVar1->Flink;
      obuf_len = param_1->CurrentStackLocation->OutputBufferLength;
      if ((*(byte *)&pMVar2->MdlFlags & 5) == 0) {
        pp_Var13 = (byte *)MmMapLockedPagesSpecifyCache(pMVar2,0,1,0,0,0x40000010);
      }
      else {
        pp_Var13 = (byte *)pMVar2->MappedSystemVa;
      }
      if ((ulonglong *)pp_Var13 == (ulonglong *)0x0) {
        *result = 0;
        KeReleaseSpinLock(&lport->lock,(uint)bVar4);
        uVar6 = STATUS_INVALID_PARAMETER;
        goto LAB_140002334;
      }
      obuf_payload_len = obuf_len - 8;
      rlen = pSVar4->size;
      if (obuf_payload_len <= pSVar4->size) {
        rlen = obuf_payload_len;
      }
      *(ulonglong *)pp_Var13 = pSVar4->port;
      memmove((ulonglong *)((longlong)pp_Var13 + 8),lport->buffer + pSVar4->buf_offset,
              (ulonglong)rlen);
      *result = rlen + 8;
      pSVar5 = (ShdwBlock *)(ulonglong)inptr[2];
      if ((inptr[2] & 1) == 0) {
        pSVar4->size = pSVar4->size - rlen;
        pSVar4->buf_offset = pSVar4->buf_offset + rlen;
        if (pSVar4->size == 0) {
          pSVar3 = pSVar4->Flink;
          if ((pSVar3->Blink != pSVar4) || (pSVar5 = pSVar4->Blink, pSVar5->Flink != pSVar4)) {
            swi(0x29);
            return (int)pSVar5;
          }
          pSVar5->Flink = pSVar3;
          pSVar3->Blink = pSVar5;
          NdisFreeMemory(pSVar4,0,0);
        }
        rcv_len = 0;
        for (pSVar5 = ppSVar1->Flink; pSVar5 != ppSVar1; pSVar5 = pSVar5->Flink) {
          memmove(lport->buffer + rcv_len,lport->buffer + pSVar5->buf_offset,(ulonglong)pSVar5->size
                 );
          pSVar5->buf_offset = rcv_len;
          rcv_len = rcv_len + pSVar5->size;
        }
        lport->rcv_len_total = rcv_len;
      }
    }
    KeReleaseSpinLock(&lport->lock,(uint)bVar4);
    if (*result == 0) {
      if ((uint)(lport->rcv_irp_csq).count < 10) {
        IoCsqInsertIrp(&lport->rcv_irp_csq,param_1,0);
        piVar1 = &(lport->rcv_irp_csq).count;
        *piVar1 = *piVar1 + 1;
        uVar6 = 0x103;
      }
      else {
        DbgPrint("[SHDW] [E] ShdwRcvIRPHandler: Unable to add more IRPs to the queue\n");
        uVar6 = 0xc0000044;
      }
    }
  }
LAB_140002334:
  ShdwCtrlDecrefLocalPort(lport);
  return uVar6;
}



ulonglong ShdwRcvNetworkHandler
                    (ShdwLocalPort *param_1,undefined8 *param_2,void *param_3,uint param_4)

{
  ShdwBlock **ppSVar1;
  ShdwBlock *pSVar2;
  byte bVar3;
  int iVar4;
  longlong lVar5;
  IRP *next_irp;
  ulonglong uVar6;
  uint uVar7;
  ulonglong uVar8;
  bool bVar9;
  ShdwBlock *rcv_block;
  uint statusval;
  
  DbgPrint("[SHDW] [*] ShdwRcvNetworkHandler: Receiving %lu byte(s) from the network\n",param_4);
  bVar3 = KeAcquireSpinLockRaiseToDpc(&param_1->lock);
  if ((0x1000 < param_1->rcv_len_total + param_4) &&
     (uVar7 = 0x1000 - param_1->rcv_len_total, bVar9 = uVar7 != param_4, param_4 = uVar7, bVar9)) {
    DbgPrint("[SHDW] [W] ShdwRcvStoreData: The receive buffer is full. Truncating bytes...\n");
  }
  ppSVar1 = &param_1->BlockFlink;
  rcv_block = (ShdwBlock *)0x0;
  if (*ppSVar1 == (ShdwBlock *)ppSVar1) {
LAB_140002408:
    iVar4 = NdisAllocateMemoryWithTag(&rcv_block,0x20,0x77646873);
    if (iVar4 != 0) {
      uVar8 = 0xc0000017;
      goto LAB_1400024b6;
    }
    rcv_block->Flink = (ShdwBlock *)0x0;
    rcv_block->Blink = (ShdwBlock *)0x0;
    rcv_block->port = 0;
    rcv_block->size = 0;
    rcv_block->buf_offset = 0;
    rcv_block->port = *param_2;
    rcv_block->buf_offset = param_1->rcv_len_total;
    rcv_block->size = 0;
    pSVar2 = param_1->BlockBlink;
    if (pSVar2->Flink != (ShdwBlock *)ppSVar1) {
      swi(0x29);
      return (ulonglong)rcv_block;
    }
    rcv_block->Flink = (ShdwBlock *)ppSVar1;
    rcv_block->Blink = pSVar2;
    pSVar2->Flink = rcv_block;
    param_1->BlockBlink = rcv_block;
  }
  else {
    pSVar2 = param_1->BlockBlink;
    lVar5 = RtlCompareMemory(&pSVar2->port,param_2,8);
    if (lVar5 != 0) {
      rcv_block = pSVar2;
    }
    if (rcv_block == (ShdwBlock *)0x0) goto LAB_140002408;
  }
                    /* SUS: Appending bytes to a block even if the block is not the final one! */
  memmove(param_1->buffer + (ulonglong)rcv_block->size + (ulonglong)rcv_block->buf_offset,param_3,
          (ulonglong)param_4);
  rcv_block->size = rcv_block->size + param_4;
  param_1->rcv_len_total = param_1->rcv_len_total + param_4;
  uVar8 = 0;
LAB_1400024b6:
  iVar4 = param_1->rcv_len_total;
  while ((iVar4 != 0 &&
         (next_irp = (IRP *)IoCsqRemoveNextIrp(&param_1->rcv_irp_csq,0), next_irp != (IRP *)0x0))) {
    statusval = 0;
    uVar6 = ShdwRcvComplete_(next_irp,param_1,true,&statusval);
    next_irp->IoStatusPointer = (PVOID)(ulonglong)statusval;
    uVar8 = uVar6 & 0xffffffff;
    next_irp->IoStatus = (ulong)uVar6;
    IofCompleteRequest(next_irp);
    iVar4 = param_1->rcv_len_total;
  }
  KeReleaseSpinLock(&param_1->lock,(uint)bVar3);
  return uVar8;
}



ulonglong ShdwRcvComplete_(IRP *param_1,ShdwLocalPort *param_2,bool locked,uint *statusval)

{
  uint uVar1;
  ShdwBlock **ppSVar2;
  ULONG UVar3;
  longlong *plVar4;
  MDL *pMVar5;
  ShdwBlock *pSVar6;
  ShdwBlock *pSVar7;
  byte bVar8;
  uint uVar9;
  ulonglong *puVar10;
  ShdwBlock *pSVar11;
  uint uVar12;
  uint uVar13;
  
  uVar13 = 0;
  plVar4 = param_1->AssociatedIrp;
  bVar8 = 0;
  uVar12 = 0;
  if (!locked) {
    bVar8 = KeAcquireSpinLockRaiseToDpc(&param_2->lock);
  }
  if (param_2->rcv_len_total == 0) {
    *statusval = 0;
  }
  else {
    ppSVar2 = &param_2->BlockFlink;
    pMVar5 = param_1->MdlAddress;
    pSVar6 = *ppSVar2;
    UVar3 = param_1->CurrentStackLocation->OutputBufferLength;
    if ((*(byte *)&pMVar5->MdlFlags & 5) == 0) {
      puVar10 = (ulonglong *)MmMapLockedPagesSpecifyCache(pMVar5,0,1,0,0,0x40000010);
    }
    else {
      puVar10 = (ulonglong *)pMVar5->MappedSystemVa;
    }
    if (puVar10 == (ulonglong *)0x0) {
      *statusval = 0;
      uVar12 = 0xc000000d;
    }
    else {
      uVar1 = UVar3 - 8;
      uVar9 = pSVar6->size;
      if (uVar1 <= pSVar6->size) {
        uVar9 = uVar1;
      }
      *puVar10 = pSVar6->port;
      memmove(puVar10 + 1,param_2->buffer + pSVar6->buf_offset,(ulonglong)uVar9);
      *statusval = uVar9 + 8;
      pSVar11 = (ShdwBlock *)(ulonglong)*(uint *)(plVar4 + 1);
      if ((*(uint *)(plVar4 + 1) & 1) == 0) {
        pSVar6->size = pSVar6->size - uVar9;
        pSVar6->buf_offset = pSVar6->buf_offset + uVar9;
        if (pSVar6->size == 0) {
          pSVar7 = pSVar6->Flink;
          if ((pSVar7->Blink != pSVar6) || (pSVar11 = pSVar6->Blink, pSVar11->Flink != pSVar6)) {
            swi(0x29);
            return (ulonglong)pSVar11;
          }
          pSVar11->Flink = pSVar7;
          pSVar7->Blink = pSVar11;
          NdisFreeMemory(pSVar6,0,0);
        }
        for (pSVar6 = *ppSVar2; pSVar6 != (ShdwBlock *)ppSVar2; pSVar6 = pSVar6->Flink) {
          memmove(param_2->buffer + uVar13,param_2->buffer + pSVar6->buf_offset,
                  (ulonglong)pSVar6->size);
          pSVar6->buf_offset = uVar13;
          uVar13 = uVar13 + pSVar6->size;
        }
        param_2->rcv_len_total = uVar13;
      }
    }
  }
  if (!locked) {
    KeReleaseSpinLock(&param_2->lock,(uint)bVar8);
  }
  return (ulonglong)uVar12;
}



int ShdwCtrlAddRemotePort(ulonglong *port,BCRYPT_RSAKEY_BLOB *pubkey,ULONG pubkey_len)

{
  int *piVar1;
  ShdwDeviceExtension *pSVar2;
  BCRYPT_KEY_HANDLE hKey;
  ShdwRemotePort *pSVar3;
  undefined uVar4;
  int iVar5;
  NTSTATUS NVar6;
  int iVar7;
  ShdwRemotePort **ppSVar8;
  ShdwRemotePort *local_res20;
  
  pSVar2 = gDeviceObject->DeviceExtension;
  uVar4 = ExAcquireSpinLockExclusive(&(pSVar2->ll).lock);
  iVar7 = 0;
  ppSVar8 = &(gDeviceObject->DeviceExtension->ll).RemoteFlink;
  for (local_res20 = *ppSVar8; local_res20 != (ShdwRemotePort *)ppSVar8;
      local_res20 = local_res20->Flink) {
    if (*(ulonglong *)&local_res20->field_0x14 == *port) {
      LOCK();
      local_res20->refcount = local_res20->refcount + 1;
      UNLOCK();
      if (local_res20 == (ShdwRemotePort *)0x0) goto LAB_140002770;
      ShdwCtrlDecrefRemotePort(local_res20);
      DbgPrint("[SHDW] [E] ShdwCtrlAddRemotePort: remote port already exists\n");
      iVar7 = -0x3fffff1b;
      goto LAB_1400028c0;
    }
  }
  local_res20 = (ShdwRemotePort *)0x0;
LAB_140002770:
  iVar5 = NdisAllocateMemoryWithTag(&local_res20,0x28,0x77646873);
  if (iVar5 == 0) {
    local_res20->Flink = (ShdwRemotePort *)0x0;
    local_res20->Blink = (ShdwRemotePort *)0x0;
    local_res20->refcount = 0;
    *(undefined4 *)&local_res20->field_0x14 = 0;
    local_res20->port = 0;
    local_res20->rsa_pubkey = (BCRYPT_KEY_HANDLE)0x0;
    *(ulonglong *)&local_res20->field_0x14 = *port;
    NVar6 = BCryptImportKeyPair(gBCryptRSA,(BCRYPT_KEY_HANDLE)0x0,L"RSAPUBLICBLOB",
                                &local_res20->rsa_pubkey,(PUCHAR)pubkey,pubkey_len,0);
    if (-1 < NVar6) goto LAB_14000286f;
    DbgPrint("[SHDW] [E] ShdwCtrlCreateRemotePort: public key is corrupted\n");
    iVar7 = -0x3fffff1b;
  }
  else {
    iVar7 = -0x3fffffe9;
  }
  pSVar3 = local_res20;
  if ((local_res20 != (ShdwRemotePort *)0x0) && (local_res20->refcount == 0)) {
    ShdwDbgPrintPort_("ShdwCtrlDestroyRemotePort: Remote port deleted",
                      (ulonglong *)&local_res20->field_0x14,":");
    hKey = pSVar3->rsa_pubkey;
    if (hKey != (BCRYPT_KEY_HANDLE)0x0) {
      BCryptDestroyKey(hKey);
    }
    NdisFreeMemory(pSVar3,0,0);
  }
LAB_14000286f:
  if (-1 < iVar7) {
    LOCK();
    local_res20->refcount = local_res20->refcount + 1;
    UNLOCK();
    ppSVar8 = &(pSVar2->ll).RemoteFlink;
    pSVar3 = *ppSVar8;
    if (pSVar3->Blink != (ShdwRemotePort *)ppSVar8) {
                    /* __fastfail */
      swi(0x29);
      return (int)ppSVar8;
    }
    local_res20->Flink = pSVar3;
    local_res20->Blink = (ShdwRemotePort *)ppSVar8;
    pSVar3->Blink = local_res20;
    *ppSVar8 = local_res20;
    piVar1 = &(pSVar2->ll).count;
    *piVar1 = *piVar1 + 1;
    ShdwDbgPrintPort_("ShdwCtrlAddRemotePort: Remote port added",port,":");
  }
LAB_1400028c0:
  ExReleaseSpinLockExclusive(&(pSVar2->ll).lock,uVar4);
  return iVar7;
}



undefined8 ShdwCtrlDelRemotePort(ulonglong *param_1)

{
  int *piVar1;
  int iVar2;
  ShdwDeviceExtension *pSVar3;
  ShdwRemotePort *pSVar4;
  ShdwRemotePort *pSVar5;
  ShdwRemotePort *pSVar6;
  code *pcVar7;
  undefined uVar8;
  ShdwRemotePort **ppSVar9;
  undefined8 uVar10;
  
  uVar10 = 0;
  pSVar3 = gDeviceObject->DeviceExtension;
  uVar8 = ExAcquireSpinLockExclusive(&(pSVar3->ll).lock);
  ppSVar9 = &(gDeviceObject->DeviceExtension->ll).RemoteFlink;
  pSVar4 = *ppSVar9;
  do {
    if (pSVar4 == (ShdwRemotePort *)ppSVar9) {
LAB_140002961:
      DbgPrint("[SHDW] [E] ShdwCtrlDelRemotePort: not found\n");
      uVar10 = 0xc00000e5;
LAB_140002973:
      ExReleaseSpinLockExclusive(&(pSVar3->ll).lock,uVar8);
      return uVar10;
    }
    if (*(ulonglong *)&pSVar4->field_0x14 == *param_1) {
      LOCK();
      pSVar4->refcount = pSVar4->refcount + 1;
      UNLOCK();
      if (pSVar4 != (ShdwRemotePort *)0x0) {
        pSVar5 = pSVar4->Flink;
        if ((pSVar5->Blink != pSVar4) || (pSVar6 = pSVar4->Blink, pSVar6->Flink != pSVar4)) {
          pcVar7 = (code *)swi(0x29);
          (*pcVar7)(3);
          pcVar7 = (code *)swi(3);
          uVar10 = (*pcVar7)();
          return uVar10;
        }
        pSVar6->Flink = pSVar5;
        pSVar5->Blink = pSVar6;
        LOCK();
        piVar1 = &pSVar4->refcount;
        iVar2 = *piVar1;
        *piVar1 = *piVar1 + -1;
        UNLOCK();
        if ((iVar2 == 1) && (pSVar4->refcount == 0)) {
          ShdwDbgPrintPort_("ShdwCtrlDestroyRemotePort: Remote port deleted",
                            (ulonglong *)&pSVar4->field_0x14,":");
          if (pSVar4->rsa_pubkey != (BCRYPT_KEY_HANDLE)0x0) {
            BCryptDestroyKey(pSVar4->rsa_pubkey);
          }
          NdisFreeMemory(pSVar4,0,0);
        }
        LOCK();
        piVar1 = &pSVar4->refcount;
        iVar2 = *piVar1;
        *piVar1 = *piVar1 + -1;
        UNLOCK();
        if ((iVar2 == 1) && (pSVar4->refcount == 0)) {
          ShdwDbgPrintPort_("ShdwCtrlDestroyRemotePort: Remote port deleted",
                            (ulonglong *)&pSVar4->field_0x14,":");
          if (pSVar4->rsa_pubkey != (BCRYPT_KEY_HANDLE)0x0) {
            BCryptDestroyKey(pSVar4->rsa_pubkey);
          }
          NdisFreeMemory(pSVar4,0,0);
        }
        piVar1 = &(pSVar3->ll).count;
        *piVar1 = *piVar1 + -1;
        goto LAB_140002973;
      }
      goto LAB_140002961;
    }
    pSVar4 = pSVar4->Flink;
  } while( true );
}



undefined8 * ShdwGetRemotePort_(longlong *param_1,ulonglong param_2)

{
  ShdwRemotePort **ppSVar1;
  ShdwDeviceExtension *pSVar2;
  ShdwRemotePort *pSVar3;
  byte bVar4;
  char cVar5;
  
  cVar5 = (char)param_2;
  bVar4 = 0;
  pSVar2 = gDeviceObject->DeviceExtension;
  if (cVar5 == '\0') {
    bVar4 = ExAcquireSpinLockShared(&(pSVar2->ll).lock);
  }
  ppSVar1 = &(pSVar2->ll).RemoteFlink;
  pSVar3 = *ppSVar1;
  while( true ) {
    if (pSVar3 == (ShdwRemotePort *)ppSVar1) {
      if (cVar5 == '\0') {
        ExReleaseSpinLockShared(&(pSVar2->ll).lock,bVar4);
      }
      return (undefined8 *)0x0;
    }
    if (*(longlong *)&pSVar3->field_0x14 == *param_1) break;
    pSVar3 = pSVar3->Flink;
  }
  LOCK();
  pSVar3->refcount = pSVar3->refcount + 1;
  UNLOCK();
  if (cVar5 != '\0') {
    return &pSVar3->Flink;
  }
  ExReleaseSpinLockShared(&(pSVar2->ll).lock,bVar4);
  return &pSVar3->Flink;
}



undefined8 * ShdwGetRandomRemotePort_(void)

{
  int *piVar1;
  ShdwRemotePort **ppSVar2;
  ShdwDeviceExtension *pSVar3;
  uint uVar4;
  uint uVar5;
  ShdwRemotePort *pSVar6;
  
  pSVar3 = gDeviceObject->DeviceExtension;
  piVar1 = &(pSVar3->ll).lock;
  ExAcquireSpinLockShared(piVar1);
  if ((pSVar3->ll).count == 0) {
LAB_140002b9b:
    ExReleaseSpinLockShared(piVar1,0);
    return (undefined8 *)0x0;
  }
  uVar4 = randint();
  ppSVar2 = &(pSVar3->ll).RemoteFlink;
  uVar4 = uVar4 % (uint)(pSVar3->ll).count;
  pSVar6 = *ppSVar2;
  uVar5 = 0;
  if (uVar4 != 0) {
    do {
      if (pSVar6 == (ShdwRemotePort *)ppSVar2) goto LAB_140002b9b;
      pSVar6 = pSVar6->Flink;
      uVar5 = uVar5 + 1;
    } while (uVar5 < uVar4);
  }
  LOCK();
  pSVar6->refcount = pSVar6->refcount + 1;
  UNLOCK();
  ExReleaseSpinLockShared(piVar1,0);
  return &pSVar6->Flink;
}



void ShdwCtrlDecrefRemotePort(ShdwRemotePort *param_1)

{
  int *piVar1;
  int iVar2;
  
  LOCK();
  piVar1 = &param_1->refcount;
  iVar2 = *piVar1;
  *piVar1 = *piVar1 + -1;
  UNLOCK();
  if ((iVar2 == 1) && (param_1->refcount == 0)) {
    ShdwDbgPrintPort_("ShdwCtrlDestroyRemotePort: Remote port deleted",
                      (ulonglong *)&param_1->field_0x14,":");
    if (param_1->rsa_pubkey != (BCRYPT_KEY_HANDLE)0x0) {
      BCryptDestroyKey(param_1->rsa_pubkey);
    }
    NdisFreeMemory(param_1,0,0);
    return;
  }
  return;
}



ulonglong ShdwSndIRPHandler(longlong **param_1,to_send_msg *param_2)

{
  ShdwBlock *pSVar1;
  byte bVar2;
  uint uVar3;
  ShdwLocalPort *srcport;
  undefined8 uVar4;
  ShdwBlock *pSVar5;
  ulonglong uVar6;
  ShdwBlock *to_send;
  
  srcport = ShdwGetLocalPort(&param_2->srcport,0);
  if (srcport == (ShdwLocalPort *)0x0) {
    uVar6 = 0xc00000e5;
  }
  else {
    if ((srcport->flags & 4) == 0) {
      uVar6 = 0xc00000bb;
    }
    else if ((param_2->flags & 2) == 0) {
      if ((param_2->flags & 4) == 0) {
        uVar3 = ShdwSndSend((DEVICE_OBJECT *)param_1,&param_2->srcport,&param_2->dstport,
                            param_2->data,(uint)param_2->size);
        uVar6 = (ulonglong)uVar3;
      }
      else {
        bVar2 = KeAcquireSpinLockRaiseToDpc(&srcport->lock);
        uVar4 = ShdwStoreSend_(srcport,&param_2->dstport,param_2->data,(uint)param_2->size);
        if ((int)uVar4 < 0) {
          DbgPrint("[SHDW] [W] ShdwSndIRPHandler: Failed to store delayed data\n");
        }
        for (to_send = srcport->BlockFlink; to_send != (ShdwBlock *)&srcport->BlockFlink;
            to_send = to_send->Flink) {
          pSVar5 = (ShdwBlock *)RtlCompareMemory(&to_send->port,&param_2->dstport,8);
          if (pSVar5 != (ShdwBlock *)0x0) {
            if (to_send != (ShdwBlock *)0x0) {
              pSVar1 = to_send->Flink;
              if ((pSVar1->Blink != to_send) || (pSVar5 = to_send->Blink, pSVar5->Flink != to_send))
              {
                swi(0x29);
                return (ulonglong)pSVar5;
              }
              pSVar5->Flink = pSVar1;
              pSVar1->Blink = pSVar5;
              KeReleaseSpinLock(&srcport->lock,(uint)bVar2);
              uVar3 = ShdwSndSend((DEVICE_OBJECT *)param_1,&srcport->port,&to_send->port,
                                  (byte *)&to_send->buf_offset,to_send->size);
              uVar6 = (ulonglong)uVar3;
              NdisFreeMemory(to_send,0,0);
              goto LAB_140002db4;
            }
            break;
          }
        }
        uVar6 = 0xc00000e5;
        KeReleaseSpinLock(&srcport->lock,(uint)bVar2);
      }
    }
    else {
      bVar2 = KeAcquireSpinLockRaiseToDpc(&srcport->lock);
      uVar6 = ShdwStoreSend_(srcport,&param_2->dstport,param_2->data,(uint)param_2->size);
      uVar6 = uVar6 & 0xffffffff;
      KeReleaseSpinLock(&srcport->lock,(uint)bVar2);
    }
LAB_140002db4:
    ShdwCtrlDecrefLocalPort(srcport);
  }
  return uVar6;
}



void ShdwSndSend(DEVICE_OBJECT *param_1,ulonglong *srcport,ulonglong *dstport,byte *payload,
                uint size)

{
  undefined2 *puVar1;
  PUCHAR _Src;
  uint hopcount;
  int iVar3;
  ulonglong uVar4;
  ShdwRemotePort *destport;
  uint iVar2;
  ulonglong offset;
  uint uVar5;
  uint uVar6;
  undefined auStackY_138 [32];
  uint msg_ctlen;
  undefined2 *msg_ct;
  PUCHAR msg_pt;
  uint local_d0;
  ULONG local_cc;
  ULONG local_c8;
  uint local_c4;
  ulonglong *_srcport;
  byte *_payload;
  ShdwDeviceExtension *local_b0;
  ulonglong *_dstport;
  DEVICE_OBJECT *_device;
  undefined key [16];
  byte enc_key [64];
  uint msg_ptlen;
  
  uVar4 = DAT_140007010 ^ (ulonglong)auStackY_138;
  iVar2 = 0;
  _srcport = srcport;
  _payload = payload;
  _dstport = dstport;
  _device = param_1;
  hopcount = randint();
  hopcount = hopcount % 6;
  DbgPrint("[SHDW] [*] ShdwSndSend: Sending %lu byte(s)\n",size);
  local_c4 = mtus_for_hops[hopcount];
  offset = 0;
  local_d0 = 0;
  if (size != 0) {
    do {
      uVar5 = (uint)offset;
      if ((int)iVar2 < 0) break;
      msg_pt = (PUCHAR)0x0;
      msg_ct = (undefined2 *)0x0;
      msg_ctlen = 0;
      uVar6 = size;
      if (local_c4 < size) {
        uVar6 = local_c4;
      }
      local_b0 = param_1->DeviceExtension;
      key = ZEXT816(0);
      enc_key._0_16_ = ZEXT816(0);
      enc_key._16_16_ = ZEXT816(0);
      enc_key._32_16_ = ZEXT816(0);
      enc_key._48_16_ = ZEXT816(0);
      iVar3 = NdisAllocateMemoryWithTag(&msg_pt,0x5ee,0x77646873);
      if ((iVar3 == 0) && (iVar3 = NdisAllocateMemoryWithTag(&msg_ct,0x5ee,0x77646873), iVar3 == 0))
      {
        destport = (ShdwRemotePort *)ShdwGetRemotePort_((longlong *)dstport,0);
        puVar1 = msg_ct;
        if (destport == (ShdwRemotePort *)0x0) {
          iVar2 = 0xc00000e5;
        }
        else {
          msg_ctlen = uVar6 + 10;
          memset(msg_ct,0,(ulonglong)msg_ctlen);
          *puVar1 = 0xffff;
          *(ulonglong *)(puVar1 + 1) = *_srcport;
          memmove(puVar1 + 5,_payload + offset,(ulonglong)uVar6);
          iVar2 = BCryptGenRandom(gBCryptRNG,key,0x10,0);
          uVar5 = 0;
          if (-1 < (int)iVar2) {
            local_cc = 0;
            iVar2 = BCryptEncrypt(destport->rsa_pubkey,key,0x10,(void *)0x0,(PUCHAR)0x0,0,enc_key,
                                  0x40,&local_cc,2);
          }
          if (-1 < (int)iVar2) {
            memmove(msg_pt,msg_ct,(ulonglong)msg_ctlen);
            msg_ptlen = msg_ctlen;
            msg_ctlen = 0x5ee;
            iVar2 = ShdwCryptEncrypt_(key,msg_pt,msg_ptlen,(PUCHAR)msg_ct,&msg_ctlen);
            if (-1 < (int)iVar2) {
              if (hopcount != 0) {
                do {
                  memmove(msg_pt,msg_ct,(ulonglong)msg_ctlen);
                  _Src = msg_pt;
                  puVar1 = msg_ct;
                  msg_ptlen = msg_ctlen;
                  msg_ctlen = msg_ctlen + 0x4a;
                  memset(msg_ct,0,(ulonglong)msg_ctlen);
                  *puVar1 = 0xc0c0;
                  *(undefined8 *)(puVar1 + 1) = *(undefined8 *)&destport->field_0x14;
                  *(undefined4 *)(puVar1 + 5) = enc_key._0_4_;
                  *(undefined4 *)(puVar1 + 7) = enc_key._4_4_;
                  *(undefined4 *)(puVar1 + 9) = enc_key._8_4_;
                  *(undefined4 *)(puVar1 + 0xb) = enc_key._12_4_;
                  *(undefined4 *)(puVar1 + 0xd) = enc_key._16_4_;
                  *(undefined4 *)(puVar1 + 0xf) = enc_key._20_4_;
                  *(undefined4 *)(puVar1 + 0x11) = enc_key._24_4_;
                  *(undefined4 *)(puVar1 + 0x13) = enc_key._28_4_;
                  *(undefined4 *)(puVar1 + 0x15) = enc_key._32_4_;
                  *(undefined4 *)(puVar1 + 0x17) = enc_key._36_4_;
                  *(undefined4 *)(puVar1 + 0x19) = enc_key._40_4_;
                  *(undefined4 *)(puVar1 + 0x1b) = enc_key._44_4_;
                  *(undefined4 *)(puVar1 + 0x1d) = enc_key._48_4_;
                  *(undefined4 *)(puVar1 + 0x1f) = enc_key._52_4_;
                  *(undefined4 *)(puVar1 + 0x21) = enc_key._56_4_;
                  *(undefined4 *)(puVar1 + 0x23) = enc_key._60_4_;
                  memmove(puVar1 + 0x25,_Src,(ulonglong)msg_ptlen);
                  ShdwCtrlDecrefRemotePort(destport);
                  destport = (ShdwRemotePort *)ShdwGetRandomRemotePort_();
                  if (destport == (ShdwRemotePort *)0x0) {
                    iVar2 = 0xc00000e5;
                    uVar5 = local_d0;
                    goto LAB_14000318b;
                  }
                  iVar2 = BCryptGenRandom(gBCryptRNG,key,0x10,0);
                  if (-1 < (int)iVar2) {
                    local_c8 = 0;
                    iVar2 = BCryptEncrypt(destport->rsa_pubkey,key,0x10,(void *)0x0,(PUCHAR)0x0,0,
                                          enc_key,0x40,&local_c8,2);
                  }
                  if ((int)iVar2 < 0) goto LAB_14000317d;
                  memmove(msg_pt,msg_ct,(ulonglong)msg_ctlen);
                  msg_ptlen = msg_ctlen;
                  msg_ctlen = 0x5ee;
                  iVar2 = ShdwCryptEncrypt_(key,msg_pt,msg_ptlen,(PUCHAR)msg_ct,&msg_ctlen);
                  if ((int)iVar2 < 0) goto LAB_14000317d;
                  uVar5 = uVar5 + 1;
                } while (uVar5 < hopcount);
              }
              offset = ShdwSndDoSend_(&local_b0->NdisBindingContext->SendBuffers,
                                      (undefined8 *)&destport->field_0x14,(undefined4 *)enc_key,
                                      msg_ct,msg_ctlen);
              if ((int)offset != 0) {
                iVar2 = 0xc00000e5;
              }
            }
          }
LAB_14000317d:
          ShdwCtrlDecrefRemotePort(destport);
          uVar5 = local_d0;
        }
LAB_14000318b:
        NdisFreeMemory(msg_pt,0,0);
        NdisFreeMemory(msg_ct,0,0);
        if ((int)iVar2 < 0) goto LAB_1400031cf;
      }
      else {
        if (msg_pt != (PUCHAR)0x0) {
          NdisFreeMemory(msg_pt,0,0);
        }
        iVar2 = 0xc0000017;
LAB_1400031cf:
        DbgPrint("[SHDW] [E] ShdwSndSend: An error occured while sending data (%lx)\n",iVar2);
      }
      size = size - uVar6;
      local_d0 = uVar5 + uVar6;
      offset = (ulonglong)local_d0;
      dstport = _dstport;
      param_1 = _device;
    } while (size != 0);
  }
  __check_stack_cookie(uVar4 ^ (ulonglong)auStackY_138);
  return;
}



undefined8 ShdwStoreSend_(ShdwLocalPort *srcport,undefined8 *dstport,void *payload,uint size)

{
  ShdwBlock **ppSVar1;
  int iVar2;
  longlong lVar3;
  ShdwBlock *pSVar4;
  undefined8 uVar5;
  uint uVar6;
  ulonglong _Size;
  ShdwBlock *_blk;
  ShdwBlock *dbuf;
  ShdwBlock *blk;
  
  ppSVar1 = &srcport->BlockFlink;
  _Size = (ulonglong)size;
  for (blk = srcport->BlockFlink; blk != (ShdwBlock *)ppSVar1; blk = blk->Flink) {
    lVar3 = RtlCompareMemory(&blk->port,dstport,8);
    if (lVar3 != 0) {
      _blk = blk;
      if (blk == (ShdwBlock *)0x0) goto LAB_14000328d;
                    /* whoa hey why are we allowing up to 0x4000 here */
      uVar6 = blk->size + size;
      if (0x4000 < uVar6) {
        return 0x80000005;
      }
      dbuf = (ShdwBlock *)0x0;
      iVar2 = NdisAllocateMemoryWithTag(&dbuf,uVar6 + 0x20,0x77646873);
      if (iVar2 != 0) goto LAB_1400033f8;
      memset(dbuf,0,(ulonglong)uVar6 + 0x20);
      memmove(dbuf,_blk,(ulonglong)_blk->size + 0x20);
      memmove((void *)((longlong)&dbuf->buf_offset + (ulonglong)_blk->size),payload,_Size);
      dbuf->size = uVar6;
      blk = _blk->Flink;
      pSVar4 = dbuf;
      if ((blk->Blink == _blk) && (pSVar4 = _blk->Blink, pSVar4->Flink == _blk)) {
        pSVar4->Flink = blk;
        blk->Blink = pSVar4;
        blk = srcport->BlockBlink;
        pSVar4 = dbuf;
        if (blk->Flink == (ShdwBlock *)ppSVar1) {
          dbuf->Flink = (ShdwBlock *)ppSVar1;
          dbuf->Blink = blk;
          blk->Flink = dbuf;
          srcport->BlockBlink = dbuf;
          NdisFreeMemory(_blk,0,0);
          return 0;
        }
      }
      goto LAB_1400033f1;
    }
  }
  _blk = (ShdwBlock *)0x0;
LAB_14000328d:
  iVar2 = NdisAllocateMemoryWithTag(&_blk,size + 0x20,0x77646873);
  if (iVar2 == 0) {
    memset(_blk,0,_Size + 0x20);
    memmove(&_blk->buf_offset,payload,_Size);
    _blk->port = *dstport;
    _blk->size = size;
    blk = srcport->BlockBlink;
    pSVar4 = _blk;
    if (blk->Flink != (ShdwBlock *)ppSVar1) {
LAB_1400033f1:
      swi(0x29);
      return pSVar4;
    }
    _blk->Flink = (ShdwBlock *)ppSVar1;
    _blk->Blink = blk;
    blk->Flink = _blk;
    srcport->BlockBlink = _blk;
    uVar5 = 0;
  }
  else {
LAB_1400033f8:
    uVar5 = 0xc0000017;
  }
  return uVar5;
}



void ShdwPtDispatchPacket(undefined8 param_1,ShdwPacket *param_2)

{
  ushort uVar1;
  short sVar2;
  byte bVar3;
  uint iVar3;
  int iVar4;
  undefined8 uVar5;
  uint size;
  
  size = param_2->size;
  if (size < 0x12) {
    return;
  }
  uVar1 = (param_2->pkt).opcode;
  if (uVar1 == 0xc0de) {
    iVar3 = ShdwPtMsgPacketHandler(&param_2->pkt,size);
    if ((iVar3 == 0) || (iVar3 == 0xc000021b)) goto LAB_1400034e7;
  }
  else if (uVar1 == 0xcafe) {
    if (size < 0x1c) {
      iVar3 = 0xc0010016;
    }
    else {
      sVar2 = (param_2->pkt).sync_cmd;
      if (sVar2 == 1) {
        bVar3 = check_pubkey((BCRYPT_RSAKEY_BLOB *)(param_2->pkt).field7_0x1c,size - 0x1c);
        if (bVar3 == 0) {
          DbgPrint(
                  "[SHDW] [E] ShdwPtSyncPacketHandler: invalid public key blob received. Ignoring SYNC packet...\n"
                  );
          iVar3 = 0xc0010015;
        }
        else {
          iVar4 = ShdwCtrlAddRemotePort
                            (&(param_2->pkt).port,(BCRYPT_RSAKEY_BLOB *)(param_2->pkt).field7_0x1c,
                             size - 0x1c);
          if (-1 < iVar4) goto LAB_1400034e7;
          iVar3 = 0xc0000001;
        }
      }
      else {
        if (sVar2 != 2) goto LAB_14000344b;
        uVar5 = ShdwCtrlDelRemotePort(&(param_2->pkt).port);
        if (-1 < (int)uVar5) goto LAB_1400034e7;
        iVar3 = 0xc0000001;
      }
    }
  }
  else {
LAB_14000344b:
    iVar3 = 0xc00000bb;
  }
  DbgPrint("[SHDW] [W] ShdwPtDispatchPacket: error while processing packet : %lx\n",iVar3);
LAB_1400034e7:
  IoFreeWorkItem(param_2->work_item);
  NdisFreeMemory(param_2,0,0);
  return;
}



/* WARNING: Unable to use type for symbol pbVar10 */

void ShdwPtMsgPacketHandler(ShdwEtherPacket *param_1,uint param_2)

{
  uint uVar1;
  undefined4 *puVar2;
  ShdwBindingContext *pSVar3;
  undefined4 uVar4;
  undefined4 uVar5;
  undefined4 uVar6;
  undefined uVar7;
  undefined uVar8;
  NTSTATUS NVar9;
  int iVar10;
  uint uVar11;
  ShdwLocalPort *pSVar12;
  ulonglong uVar13;
  undefined auStackY_b8 [32];
  ShdwEtherPacket *local_68;
  uint local_60;
  ushort *local_58;
  ULONG local_50 [2];
  undefined local_48 [16];
  ulonglong local_38;
  ushort *pbVar10;
  
  local_38 = DAT_140007010 ^ (ulonglong)auStackY_b8;
  local_58 = (ushort *)0x0;
  local_48 = ZEXT816(0);
  if (param_2 < 8) goto LAB_140003827;
  pSVar12 = ShdwGetLocalPort(&param_1->port,0);
  if (pSVar12 != (ShdwLocalPort *)0x0) {
    local_50[0] = 0;
    NVar9 = BCryptDecrypt(pSVar12->rsapriv_handle,(PUCHAR)&param_1->sync_cmd,0x40,(void *)0x0,
                          (PUCHAR)0x0,0,local_48,0x10,local_50,2);
    if (NVar9 < 0) {
      DbgPrint("[SHDW] [E] ShdwPtMsgPacketHandler: Failed to decrypt AES key\n");
    }
    else {
      iVar10 = NdisAllocateMemoryWithTag(&local_58,0x5ee,0x77646873);
      if (iVar10 == 0) {
        memset(local_58,0,0x5ee);
        local_60 = 0x5ee;
        iVar10 = ShdwAesDecrypt_(local_48,param_1->field7_0x1c + 0x3e,param_2 - 0x5a,
                                 (PUCHAR)local_58,&local_60);
        pbVar10 = local_58;
        uVar11 = local_60;
        if (iVar10 < 0) {
          DbgPrint("[SHDW] [E] ShdwPtMsgPacketHandler: failed to decrypt MSG data\n");
        }
        else if (*local_58 == 0xc0c0) {
          if (0x49 < local_60) {
            uVar1 = local_60 - 0x4a;
            local_68 = (ShdwEtherPacket *)0x0;
            pSVar3 = gDeviceObject->DeviceExtension->NdisBindingContext;
            iVar10 = NdisAllocateMemoryWithTag(&local_68,local_60 + 0x10,0x77646873);
            if (iVar10 == 0) {
              local_68->opcode = 0xc0de;
              local_68->size = (short)uVar1 + 0x4c;
              local_68->dest[0] = 0xff;
              local_68->dest[1] = 0xff;
              local_68->dest[2] = 0xff;
              local_68->dest[3] = 0xff;
              local_68->dest[4] = 0xff;
              local_68->dest[5] = 0xff;
              local_68->src[0] = 0xff;
              local_68->src[1] = 0xff;
              *(undefined4 *)(local_68->src + 2) = 0xffffffff;
              local_68->ethertype = 0xdead;
              local_68->port = *(ulonglong *)(pbVar10 + 1);
              puVar2 = (undefined4 *)(pbVar10 + 5);
              uVar7 = *(undefined *)((longlong)puVar2 + 2);
              uVar8 = *(undefined *)((longlong)puVar2 + 3);
              uVar4 = *(undefined4 *)(pbVar10 + 7);
              uVar5 = *(undefined4 *)(pbVar10 + 9);
              uVar6 = *(undefined4 *)(pbVar10 + 0xb);
              local_68->sync_cmd = *(undefined2 *)puVar2;
              local_68->field7_0x1c[0] = uVar7;
              local_68->field7_0x1c[1] = uVar8;
              *(undefined4 *)(local_68->field7_0x1c + 2) = uVar4;
              *(undefined4 *)(local_68->field7_0x1c + 6) = uVar5;
              *(undefined4 *)(local_68->field7_0x1c + 10) = uVar6;
              uVar4 = *(undefined4 *)(pbVar10 + 0xf);
              uVar5 = *(undefined4 *)(pbVar10 + 0x11);
              uVar6 = *(undefined4 *)(pbVar10 + 0x13);
              *(undefined4 *)(local_68->field7_0x1c + 0xe) = *(undefined4 *)(pbVar10 + 0xd);
              *(undefined4 *)(local_68->field7_0x1c + 0x12) = uVar4;
              *(undefined4 *)(local_68->field7_0x1c + 0x16) = uVar5;
              *(undefined4 *)(local_68->field7_0x1c + 0x1a) = uVar6;
              uVar4 = *(undefined4 *)(pbVar10 + 0x17);
              uVar5 = *(undefined4 *)(pbVar10 + 0x19);
              uVar6 = *(undefined4 *)(pbVar10 + 0x1b);
              *(undefined4 *)(local_68->field7_0x1c + 0x1e) = *(undefined4 *)(pbVar10 + 0x15);
              *(undefined4 *)(local_68->field7_0x1c + 0x22) = uVar4;
              *(undefined4 *)(local_68->field7_0x1c + 0x26) = uVar5;
              *(undefined4 *)(local_68->field7_0x1c + 0x2a) = uVar6;
              uVar4 = *(undefined4 *)(pbVar10 + 0x1f);
              uVar5 = *(undefined4 *)(pbVar10 + 0x21);
              uVar6 = *(undefined4 *)(pbVar10 + 0x23);
              *(undefined4 *)(local_68->field7_0x1c + 0x2e) = *(undefined4 *)(pbVar10 + 0x1d);
              *(undefined4 *)(local_68->field7_0x1c + 0x32) = uVar4;
              *(undefined4 *)(local_68->field7_0x1c + 0x36) = uVar5;
              *(undefined4 *)(local_68->field7_0x1c + 0x3a) = uVar6;
              memmove(local_68->field7_0x1c + 0x3e,pbVar10 + 0x25,(ulonglong)uVar1);
              uVar11 = ShdwSendPacket_(pSVar3,local_68,uVar11 + 0x10);
              NdisFreeMemory(local_68,0,0);
              if (uVar11 == 0) goto LAB_1400037fd;
            }
            DbgPrint(
                    "[SHDW] [E] ShdwPtNextHopEnvHandler: failed to send packet to the next destination\n"
                    );
          }
        }
        else if (((*local_58 == 0xffff) && (9 < local_60)) && (local_60 - 10 != 0)) {
          if ((pSVar12->flags & 2) == 0) {
            DbgPrint("[SHDW] [E] ShdwPtFinalizeEnvHandler: Error the port is not a receiver port\n")
            ;
          }
          else {
            uVar13 = ShdwRcvNetworkHandler
                               (pSVar12,(undefined8 *)(local_58 + 1),local_58 + 5,local_60 - 10);
            if ((int)uVar13 < 0) {
              DbgPrint("[SHDW] [E] ShdwPtFinalizeEnvHandler: error while handling finalize packet\n"
                      );
            }
          }
        }
      }
    }
  }
LAB_1400037fd:
  if (local_58 != (ushort *)0x0) {
    NdisFreeMemory(local_58,0,0);
  }
  if (pSVar12 != (ShdwLocalPort *)0x0) {
    ShdwCtrlDecrefLocalPort(pSVar12);
  }
LAB_140003827:
  __check_stack_cookie(local_38 ^ (ulonglong)auStackY_b8);
  return;
}



ulonglong ShdwSndDoSend_(undefined8 *param_1,undefined8 *param_2,undefined4 *param_3,void *param_4,
                        uint param_5)

{
  undefined4 uVar1;
  undefined4 uVar2;
  undefined4 uVar3;
  int iVar4;
  uint uVar5;
  ulonglong uVar6;
  undefined8 *local_28 [2];
  
  local_28[0] = (undefined8 *)0x0;
  iVar4 = NdisAllocateMemoryWithTag(local_28,param_5 + 0x5a,0x77646873);
  if (iVar4 == 0) {
    *(undefined2 *)((longlong)local_28[0] + 0xe) = 0xc0de;
    *(short *)(local_28[0] + 2) = (short)param_5 + 0x4c;
    *local_28[0] = 0xffffffffffffffff;
    *(undefined4 *)(local_28[0] + 1) = 0xffffffff;
    *(undefined2 *)((longlong)local_28[0] + 0xc) = 0xdead;
    *(undefined8 *)((longlong)local_28[0] + 0x12) = *param_2;
    uVar1 = param_3[1];
    uVar2 = param_3[2];
    uVar3 = param_3[3];
    *(undefined4 *)((longlong)local_28[0] + 0x1a) = *param_3;
    *(undefined4 *)((longlong)local_28[0] + 0x1e) = uVar1;
    *(undefined4 *)((longlong)local_28[0] + 0x22) = uVar2;
    *(undefined4 *)((longlong)local_28[0] + 0x26) = uVar3;
    uVar1 = param_3[5];
    uVar2 = param_3[6];
    uVar3 = param_3[7];
    *(undefined4 *)((longlong)local_28[0] + 0x2a) = param_3[4];
    *(undefined4 *)((longlong)local_28[0] + 0x2e) = uVar1;
    *(undefined4 *)((longlong)local_28[0] + 0x32) = uVar2;
    *(undefined4 *)((longlong)local_28[0] + 0x36) = uVar3;
    uVar1 = param_3[9];
    uVar2 = param_3[10];
    uVar3 = param_3[0xb];
    *(undefined4 *)((longlong)local_28[0] + 0x3a) = param_3[8];
    *(undefined4 *)((longlong)local_28[0] + 0x3e) = uVar1;
    *(undefined4 *)((longlong)local_28[0] + 0x42) = uVar2;
    *(undefined4 *)((longlong)local_28[0] + 0x46) = uVar3;
    uVar1 = param_3[0xd];
    uVar2 = param_3[0xe];
    uVar3 = param_3[0xf];
    *(undefined4 *)((longlong)local_28[0] + 0x4a) = param_3[0xc];
    *(undefined4 *)((longlong)local_28[0] + 0x4e) = uVar1;
    *(undefined4 *)((longlong)local_28[0] + 0x52) = uVar2;
    *(undefined4 *)((longlong)local_28[0] + 0x56) = uVar3;
    memmove((void *)((longlong)local_28[0] + 0x5a),param_4,(ulonglong)param_5);
    uVar5 = ShdwSendPacket_((ShdwBindingContext *)param_1,local_28[0],param_5 + 0x5a);
    NdisFreeMemory(local_28[0],0,0);
    uVar6 = (ulonglong)uVar5;
  }
  else {
    uVar6 = 0xc000009a;
  }
  return uVar6;
}



uint ShdwSendPacket_(ShdwBindingContext *param_1,void *buffer,uint size)

{
  ulonglong _Size;
  int iVar1;
  longlong lVar2;
  longlong lVar3;
  void *local_res20;
  
  _Size = (ulonglong)size;
  local_res20 = (void *)0x0;
  iVar1 = NdisAllocateMemoryWithTag(&local_res20,_Size,0x77646873);
  if (iVar1 != 0) {
    return 0xc000009a;
  }
  memset(local_res20,0,_Size);
  memmove(local_res20,buffer,_Size);
  lVar2 = NdisAllocateMdl(param_1->BindingHandle,local_res20,_Size);
  if (lVar2 != 0) {
    lVar3 = NdisAllocateNetBufferAndNetBufferList(param_1->SendBuffers,0,0,lVar2,0,_Size);
    if (lVar3 != 0) {
      *(void **)(lVar3 + 0x78) = param_1->BindingHandle;
      NdisSendNetBufferLists(param_1->BindingHandle,lVar3,0,2);
      return 0;
    }
    NdisFreeMdl(lVar2);
  }
  NdisFreeMemory(local_res20,0,0);
  return 0xc000009a;
}



ulonglong ShdwSendSyncPacket_(ShdwBindingContext *param_1,short cmd,ulonglong *port,void *payload,
                             uint param_5)

{
  int iVar1;
  uint uVar2;
  ulonglong uVar3;
  ShdwEtherPacket *local_28;
  
  local_28 = (ShdwEtherPacket *)0x0;
  iVar1 = NdisAllocateMemoryWithTag(&local_28,param_5 + 0x1c,0x77646873);
  if (iVar1 == 0) {
    memset(local_28,0,(ulonglong)param_5 + 0x1c);
    local_28->dest[0] = 0xff;
    local_28->dest[1] = 0xff;
    local_28->dest[2] = 0xff;
    local_28->dest[3] = 0xff;
    local_28->dest[4] = 0xff;
    local_28->dest[5] = 0xff;
    local_28->src[0] = 0xff;
    local_28->src[1] = 0xff;
    *(undefined4 *)(local_28->src + 2) = 0xffffffff;
    local_28->ethertype = 0xdead;
    local_28->size = (short)param_5 + 0xe;
    local_28->opcode = 0xcafe;
    local_28->port = *port;
    local_28->sync_cmd = cmd;
    if (cmd != 2) {
      memmove(local_28->field7_0x1c,payload,(ulonglong)param_5);
    }
    uVar2 = ShdwSendPacket_(param_1,local_28,param_5 + 0x1c);
    NdisFreeMemory(local_28,0,0);
    uVar3 = (ulonglong)uVar2;
  }
  else {
    uVar3 = 0xc000009a;
  }
  return uVar3;
}



void ShdwFreeBindingContext(ShdwBindingContext *param_1)

{
  if ((void *)param_1->SendBuffers != (void *)0x0) {
    NdisFreeNetBufferListPool((void *)param_1->SendBuffers);
    param_1->SendBuffers = 0;
  }
  if ((void *)param_1->buffers2 != (void *)0x0) {
    NdisFreeNetBufferListPool((void *)param_1->buffers2);
    param_1->buffers2 = 0;
  }
  DbgPrint("[SHDW] [*] ShdwFreeBindingContext: Free %p\n",param_1);
  NdisFreeMemory(param_1,0,0);
  return;
}



void ShdwPtBindAdapter(undefined8 param_1,void *BindContext,NDIS_BIND_PARAMETERS *param_3)

{
  ShdwDeviceExtension *pSVar1;
  int iVar2;
  longlong lVar3;
  undefined auStackY_b8 [32];
  ShdwBindingContext *context;
  ushort local_80 [2];
  int local_7c [1];
  uint selected_medium [2];
  NDIS_OPEN_PARAMETERS local_70;
  NET_BUFFER_LIST_POOL_PARAMETERS Parameters;
  
  register0x00000000 = DAT_140007010 ^ (ulonglong)auStackY_b8;
  pSVar1 = gDeviceObject->DeviceExtension;
  if (pSVar1->NdisBindingContext == (ShdwBindingContext *)0x0) {
    if ((param_3->PhysicalMediumType == NdisPhysicalMediumOther) ||
       (param_3->PhysicalMediumType == NdisPhysicalMedium802_3)) {
      iVar2 = NdisAllocateMemoryWithTag(&context,0x50,0x77646873);
      if (iVar2 == 0) {
        context->SendBuffers = 0;
        context->buffers2 = 0;
        context->BindingHandle = (void *)0x0;
        (context->OpenAdapterCompleteEvent).Lock = 0;
        (context->OpenAdapterCompleteEvent).SignalState = 0;
        (context->OpenAdapterCompleteEvent).WaitListHead = (LIST_ENTRY)ZEXT816(0);
        (context->OidRequestCompleteEvent).Lock = 0;
        (context->OidRequestCompleteEvent).SignalState = 0;
        (context->OidRequestCompleteEvent).WaitListHead.Flink = (_LIST_ENTRY *)0x0;
        *(undefined (*) [16])&(context->OidRequestCompleteEvent).WaitListHead.Blink = ZEXT816(0);
        NdisInitializeEvent(&context->OidRequestCompleteEvent);
        NdisInitializeEvent(&context->OpenAdapterCompleteEvent);
        Parameters.PoolTag = 0x77646873;
        Parameters.DataSize = 0;
        Parameters.Header.Type = 0x80;
        Parameters.Header.Revision = 1;
        Parameters.Header.Size = 0x10;
        Parameters.ProtocolId = 0;
        Parameters.fAllocateNetBuffer = true;
        Parameters.ContextSize = 0;
        lVar3 = NdisAllocateNetBufferListPool(gNdisProtocolHandle,&Parameters);
        context->SendBuffers = lVar3;
        lVar3 = NdisAllocateNetBufferListPool(gNdisProtocolHandle,&Parameters);
        context->buffers2 = lVar3;
        if ((context->buffers2 == 0) || (context->SendBuffers == 0)) {
          ShdwFreeBindingContext(context);
        }
        else {
          local_70.AdapterName = param_3->AdapterName;
          local_70.MediumArray = local_7c;
                    /* NdisMedium802_3 (Ethernet) */
          local_7c[0] = 0;
          local_70.SelectedMediumIndex = selected_medium;
          local_70.FrameTypeArray = local_80;
          local_80[0] = 0x888e;
          local_80[1] = 0x8100;
                    /* 0x888e = EAP over LAN (IEEE 802.1X); 0x8100 = VLAN-tagged frame (IEEE 802.1Q)
                        */
          local_70.Header.Type = 0x87;
          local_70.Header.Revision = 1;
          local_70.Header.Size = 0x38;
          local_70._4_4_ = 0;
          local_70.MediumArraySize = 1;
          local_70._28_4_ = 0;
          local_70.FrameTypeArraySize = 2;
          local_70._52_4_ = 0;
          context->OpenAdapterCompleteStatus = 0;
          iVar2 = NdisOpenAdapterEx(gNdisProtocolHandle,context,&local_70,BindContext,
                                    &context->BindingHandle);
          if (iVar2 == 0x103) {
            NdisWaitEvent(&context->OpenAdapterCompleteEvent,0);
            iVar2 = context->OpenAdapterCompleteStatus;
          }
          if (iVar2 == 0) {
            DbgPrint("[SHDW] [*] ShdwPtBindAdapter: Successfully bind to %ls (%x)\n",
                     param_3->AdapterName->Buffer,param_3->PhysicalMediumType);
            pSVar1->NdisBindingContext = context;
            NdisResetEvent(&gNdisEvent);
          }
          else {
            DbgPrint("[SHDW] [E] ShdwPtBindAdapter: Failed to bind to adapter : %lx\n",iVar2);
            ShdwFreeBindingContext(context);
          }
        }
      }
      else {
        DbgPrint("[SHDW] [E] ShdwPtBindAdapter: Failed to allocate binding context\n");
      }
    }
    else {
      DbgPrint("[SHDW] [W] ShdwPtBindAdapter: Physical medium not supported : %lu\n");
    }
  }
  else {
    DbgPrint(
            "[SHDW] [W] ShdwPtBindAdapter: Protocol driver already binded to a physical adapter. Skipping...\n"
            );
  }
  __check_stack_cookie(stack0xffffffffffffffd8 ^ (ulonglong)auStackY_b8);
  return;
}



void no_op(void)

{
  return;
}



void ShdwNetPnPEventHandler_(ShdwBindingContext *param_1,longlong param_2)

{
  int iVar1;
  ulonglong uVar2;
  undefined auStack_138 [32];
  undefined4 local_118 [4];
  NDIS_OID_REQUEST req;
  
  uVar2 = DAT_140007010 ^ (ulonglong)auStack_138;
  switch(*(undefined4 *)(param_2 + 8)) {
  case 6:
                    /* NetEventBindsComplete */
    NdisSetEvent(&gNdisEvent);
    break;
  case 9:
                    /* NetEventRestart */
    local_118[0] = 9;
    memset(&req,0,0xf0);
    req.InformationBuffer = local_118;
    req.Header.Type = 0x96;
    req.Header.Revision = 1;
    req.Header.Size = 0xf0;
    req.RequestType = NdisRequestSetInformation;
    req.PortNumber = 0;
    req.Oid = OID_GEN_CURRENT_PACKET_FILTER;
    req.InformationBufferLength = 4;
    NdisInitializeEvent(&param_1->OidRequestCompleteEvent);
    param_1->OidRequestStatus = 0;
    iVar1 = NdisOidRequest(param_1->BindingHandle,&req);
    if (iVar1 == 0x103) {
      NdisWaitEvent(&param_1->OidRequestCompleteEvent,0);
      iVar1 = param_1->OidRequestStatus;
    }
    if (iVar1 != 0) {
      DbgPrint("[SHDW] [E] ShdwSendSetOidRequest: Failed to send oid request : %ls\n",iVar1);
    }
  }
  __check_stack_cookie(uVar2 ^ (ulonglong)auStack_138);
  return;
}



void ShdwOpenAdapterCompleteEx_(ShdwBindingContext *param_1,int Status)

{
  param_1->OpenAdapterCompleteStatus = Status;
  NdisSetEvent(&param_1->OpenAdapterCompleteEvent);
  return;
}



void ShdwReceiveNetBufferLists_
               (ShdwBindingContext *ProtocolBindingContext,NET_BUFFER_LIST *NetBufferLists,
               undefined8 PortNumber,undefined8 NumberOfNetBufferLists,uint ReceiveFlags)

{
  int iVar1;
  undefined *_Src;
  longlong lVar2;
  NET_BUFFER_LIST *nbl;
  ShdwPacket *local_res8;
  NET_BUFFER *puVar2;
  NET_BUFFER *nb;
  uint size;
  
  nbl = NetBufferLists;
  if (NetBufferLists != (NET_BUFFER_LIST *)0x0) {
    do {
      nbl->field17_0x88 = nbl->field17_0x88 & 0xfffff;
      for (nb = nbl->FirstNetBuffer; nb != (NET_BUFFER *)0x0; nb = nb->next) {
        size = nb->DataLength;
        if (size - 0xe < 1505) {
          local_res8 = (ShdwPacket *)0x0;
          iVar1 = NdisAllocateMemoryWithTag(&local_res8,size + 0x10,0x77646873);
          if (iVar1 == 0) {
            memset(local_res8,0,(ulonglong)size);
            local_res8->size = size;
            _Src = (undefined *)NdisGetDataBuffer(nb,size,&local_res8->pkt,1,0);
            if (_Src == (undefined *)0x0) {
              NdisFreeMemory(local_res8,0,0);
            }
            else {
              if ((ShdwEtherPacket *)_Src != &local_res8->pkt) {
                memmove(&local_res8->pkt,_Src,(ulonglong)size);
              }
              if ((local_res8->pkt).ethertype == 0xdead) {
                lVar2 = IoAllocateWorkItem(gDeviceObject);
                local_res8->work_item = lVar2;
                if (lVar2 == 0) {
                  NdisFreeMemory(local_res8,0,0);
                }
                else {
                  IoQueueWorkItem(lVar2,ShdwPtDispatchPacket,3,local_res8);
                }
              }
              else {
                NdisFreeMemory(local_res8,0,0);
              }
            }
          }
        }
      }
      nbl = nbl->Next;
    } while (nbl != (NET_BUFFER_LIST *)0x0);
  }
  if ((ReceiveFlags & 2) == 0) {
    NdisReturnNetBufferLists(ProtocolBindingContext->BindingHandle,NetBufferLists,ReceiveFlags & 1);
  }
  return;
}



void ShdwOidRequestCompleteHandler_(ShdwBindingContext *param_1,undefined8 param_2,int Status)

{
  param_1->OidRequestStatus = Status;
  NdisSetEvent(&param_1->OidRequestCompleteEvent);
  return;
}



void ShdwSendNetBufferListsComplete_(ShdwBindingContext *param_1,NET_BUFFER_LIST *param_2)

{
  MDL *pMVar1;
  NET_BUFFER_LIST *pNVar2;
  PVOID pvVar3;
  NET_BUFFER *pNVar1;
  MDL *puVar3;
  
  if (param_2 != (NET_BUFFER_LIST *)0x0) {
    do {
      for (pNVar1 = param_2->FirstNetBuffer; pNVar1 != (NET_BUFFER *)0x0; pNVar1 = pNVar1->next) {
        puVar3 = (MDL *)pNVar1->MdlChain;
        while (puVar3 != (MDL *)0x0) {
          if ((*(byte *)&puVar3->MdlFlags & 5) == 0) {
            pvVar3 = (PVOID)MmMapLockedPagesSpecifyCache(puVar3,0,1,0,0,0x40000010);
          }
          else {
            pvVar3 = puVar3->MappedSystemVa;
          }
          if (pvVar3 != (PVOID)0x0) {
            NdisFreeMemory(pvVar3,0,0);
          }
          pMVar1 = puVar3->Next;
          NdisFreeMdl(puVar3);
          puVar3 = pMVar1;
        }
      }
      pNVar2 = param_2->Next;
      NdisFreeNetBufferList(param_2);
      param_2 = pNVar2;
    } while (pNVar2 != (NET_BUFFER_LIST *)0x0);
  }
  return;
}



undefined8 ShdwUnbindAdapterHandlerEx_(void)

{
  return 0;
}



void ShdwAesDecrypt_(PUCHAR key,PUCHAR ct,ULONG ctlen,PUCHAR pt,ULONG *ptlen)

{
  NTSTATUS NVar1;
  undefined auStackY_a8 [32];
  ULONG local_58 [2];
  BCRYPT_KEY_HANDLE local_50;
  undefined iv [16];
  ulonglong local_38;
  
  local_38 = DAT_140007010 ^ (ulonglong)auStackY_a8;
  local_50 = (BCRYPT_KEY_HANDLE)0x0;
  NVar1 = BCryptGenerateSymmetricKey(gBCryptAESCBC,&local_50,(PUCHAR)0x0,0,key,0x10,0);
  if (-1 < NVar1) {
    iv = ZEXT816(0);
    local_58[0] = 0;
    NVar1 = BCryptDecrypt(local_50,ct,ctlen,(void *)0x0,iv,0x10,pt,*ptlen,local_58,
                          BCRYPT_BLOCK_PADDING);
    if (NVar1 < 0) {
      BCryptDestroyKey(local_50);
    }
    else {
      BCryptDestroyKey(local_50);
      *ptlen = local_58[0];
    }
  }
  __check_stack_cookie(local_38 ^ (ulonglong)auStackY_a8);
  return;
}



void ShdwCryptEncrypt_(PUCHAR key,PUCHAR pt,ULONG ptlen,PUCHAR ct,ULONG *ctlen)

{
  NTSTATUS NVar1;
  undefined auStackY_a8 [32];
  ULONG local_58 [2];
  BCRYPT_KEY_HANDLE local_50;
  undefined iv [16];
  ulonglong local_38;
  
  local_38 = DAT_140007010 ^ (ulonglong)auStackY_a8;
  local_50 = (BCRYPT_KEY_HANDLE)0x0;
  NVar1 = BCryptGenerateSymmetricKey(gBCryptAESCBC,&local_50,(PUCHAR)0x0,0,key,0x10,0);
  if (-1 < NVar1) {
    iv = ZEXT816(0);
    local_58[0] = 0;
    NVar1 = BCryptEncrypt(local_50,pt,ptlen,(void *)0x0,iv,0x10,ct,*ctlen,local_58,1);
    if (NVar1 < 0) {
      BCryptDestroyKey(local_50);
    }
    else {
      BCryptDestroyKey(local_50);
      *ctlen = local_58[0];
    }
  }
  __check_stack_cookie(local_38 ^ (ulonglong)auStackY_a8);
  return;
}



byte check_pubkey(BCRYPT_RSAKEY_BLOB *param_1,uint size)

{
  if ((((0x17 < size) && (param_1->BitLength == 0x200)) && (param_1->cbPublicExp < 0x201)) &&
     (param_1->cbModulus < 0x201)) {
    return (ulonglong)param_1->cbPublicExp + 0x18 + (ulonglong)param_1->cbModulus == (ulonglong)size
    ;
  }
  return 0;
}



NTSTATUS ShdwCryptExportPublicKey(BCRYPT_KEY_HANDLE param_1,PUCHAR *param_2,ULONG *param_3)

{
  NTSTATUS NVar1;
  int iVar2;
  
  NVar1 = BCryptExportKey(param_1,(BCRYPT_KEY_HANDLE)0x0,L"RSAPUBLICBLOB",(PUCHAR)0x0,0,param_3,0);
  if (-1 < NVar1) {
    iVar2 = NdisAllocateMemoryWithTag(param_2,*param_3,0x77646873);
    if (iVar2 != 0) {
      return -0x3fffffe9;
    }
    NVar1 = BCryptExportKey(param_1,(BCRYPT_KEY_HANDLE)0x0,L"RSAPUBLICBLOB",*param_2,*param_3,
                            param_3,0);
    if (NVar1 < 0) {
      DbgPrint("[SHDW] [*] ShdwCryptExportPublicKey: Free %p\n",*param_2);
      NdisFreeMemory(*param_2,0,0);
    }
  }
  return NVar1;
}



NTSTATUS ShdwNewRSAKey_(BCRYPT_KEY_HANDLE *param_1)

{
  NTSTATUS NVar1;
  NTSTATUS NVar2;
  
  NVar1 = BCryptGenerateKeyPair(gBCryptRSA,param_1,0x200,0);
  if (-1 < NVar1) {
    NVar2 = BCryptFinalizeKeyPair(*param_1,0);
    NVar1 = 0;
    if (NVar2 < 0) {
      NVar1 = NVar2;
    }
  }
  return NVar1;
}



void ShdwProcessDeletedHandler_(undefined8 ParentId,longlong ProcessId,char Create)

{
  int iVar1;
  longlong lVar2;
  longlong *local_res20;
  
  if ((gDeviceObject != (DEVICE_OBJECT *)0x0) && (Create == '\0')) {
    local_res20 = (longlong *)0x0;
    iVar1 = NdisAllocateMemoryWithTag(&local_res20,0x10,L'\x77646873');
    if (iVar1 == 0) {
      lVar2 = IoAllocateWorkItem(gDeviceObject);
      *local_res20 = lVar2;
      if (lVar2 == 0) {
        NdisFreeMemory(local_res20,0,0);
        return;
      }
      local_res20[1] = ProcessId;
      IoQueueWorkItem(*local_res20,ShdwHandleDeletedProcess_,3,local_res20);
    }
  }
  return;
}



void ShdwHandleDeletedProcess_(longlong param_1,undefined8 *param_2)

{
  ShdwLocalPort *pSVar1;
  
  while( true ) {
    pSVar1 = (ShdwLocalPort *)ShdwGetLocalPortByProcess_(param_2[1],0);
    if (pSVar1 == (ShdwLocalPort *)0x0) break;
    ShdwCtrlCloseLocalPort((DEVICE_OBJECT *)param_1,&pSVar1->port);
    ShdwCtrlDecrefLocalPort(pSVar1);
  }
  IoFreeWorkItem(*param_2);
  NdisFreeMemory(param_2,0,0);
  return;
}



void DbgPrint(void)

{
                    /* WARNING: Could not recover jumptable at 0x0001400046a3. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  DbgPrint();
  return;
}



/* Library Function - Single Match
    __GSHandlerCheck
   
   Library: Visual Studio 2019 Release */

undefined8
__GSHandlerCheck(undefined8 param_1,ulonglong param_2,undefined8 param_3,longlong param_4)

{
  __GSHandlerCheckCommon(param_2,param_4,*(uint **)(param_4 + 0x38));
  return 1;
}



/* Library Function - Single Match
    __GSHandlerCheckCommon
   
   Library: Visual Studio 2019 Release */

void __GSHandlerCheckCommon(ulonglong param_1,longlong param_2,uint *param_3)

{
  byte bVar1;
  ulonglong uVar2;
  
  uVar2 = param_1;
  if ((*param_3 & 4) != 0) {
    uVar2 = (longlong)(int)param_3[1] + param_1 & (longlong)(int)-param_3[2];
  }
  bVar1 = *(byte *)((ulonglong)*(uint *)(*(longlong *)(param_2 + 0x10) + 8) + 3 +
                   *(longlong *)(param_2 + 8));
  if ((bVar1 & 0xf) != 0) {
    param_1 = param_1 + (bVar1 & 0xfffffff0);
  }
  __check_stack_cookie(param_1 ^ *(ulonglong *)((longlong)(int)(*param_3 & 0xfffffff8) + uVar2));
  return;
}



/* WARNING: Removing unreachable block (ram,0x00014000476d) */
/* WARNING: Removing unreachable block (ram,0x000140004749) */
/* WARNING: Removing unreachable block (ram,0x00014000473a) */

undefined8 FUN_140004730(void)

{
  int *piVar1;
  longlong lVar2;
  uint uVar3;
  byte bVar4;
  byte in_XCR0;
  
  piVar1 = (int *)cpuid_basic_info(0);
  bVar4 = 0;
  lVar2 = cpuid_Version_info(1);
  uVar3 = *(uint *)(lVar2 + 0xc);
  if (6 < *piVar1) {
    bVar4 = 0;
    lVar2 = cpuid_Extended_Feature_Enumeration_info(7);
    if ((*(uint *)(lVar2 + 4) >> 9 & 1) != 0) {
      bVar4 = 2;
    }
  }
  if (((((uVar3 >> 0x14 & 1) != 0) && ((uVar3 >> 0x1b & 1) != 0)) && ((uVar3 >> 0x1c & 1) != 0)) &&
     ((in_XCR0 & 6) == 6)) {
    bVar4 = bVar4 | 4;
  }
  DAT_140007000 = bVar4 | 1;
  return 0;
}



void __check_stack_cookie(longlong param_1)

{
  if ((param_1 == DAT_140007010) && ((short)((ulonglong)param_1 >> 0x30) == 0)) {
    return;
  }
                    /* WARNING: Subroutine does not return */
  KeBugCheckEx(0xf7,param_1,DAT_140007010,DAT_140007018,0);
}



void NdisInitializeEvent(void)

{
                    /* WARNING: Could not recover jumptable at 0x00014000482d. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisInitializeEvent();
  return;
}



void NdisRegisterProtocolDriver(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004833. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisRegisterProtocolDriver();
  return;
}



void NdisAllocateMemoryWithTag(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004839. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisAllocateMemoryWithTag();
  return;
}



void NdisFreeMemory(void)

{
                    /* WARNING: Could not recover jumptable at 0x00014000483f. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisFreeMemory();
  return;
}



void NdisWaitEvent(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004845. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisWaitEvent();
  return;
}



void NdisSendNetBufferLists(void)

{
                    /* WARNING: Could not recover jumptable at 0x00014000484b. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisSendNetBufferLists();
  return;
}



void NdisSetEvent(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004851. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisSetEvent();
  return;
}



void NdisResetEvent(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004857. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisResetEvent();
  return;
}



void NdisReturnNetBufferLists(void)

{
                    /* WARNING: Could not recover jumptable at 0x00014000485d. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisReturnNetBufferLists();
  return;
}



void NdisOidRequest(void)

{
                    /* WARNING: Could not recover jumptable at 0x000140004863. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisOidRequest();
  return;
}



/* WARNING: Unknown calling convention -- yet parameter storage is locked */

void NdisOpenAdapterEx(void *param_1,void *param_2,void *param_3,void *param_4,void **param_5)

{
                    /* WARNING: Could not recover jumptable at 0x000140004869. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NdisOpenAdapterEx(param_1,param_2,param_3,param_4,param_5);
  return;
}



NTSTATUS __stdcall
BCryptOpenAlgorithmProvider
          (BCRYPT_ALG_HANDLE *phAlgorithm,LPCWSTR pszAlgId,LPCWSTR pszImplementation,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x00014000486f. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptOpenAlgorithmProvider(phAlgorithm,pszAlgId,pszImplementation,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptSetProperty(BCRYPT_HANDLE hObject,LPCWSTR pszProperty,PUCHAR pbInput,ULONG cbInput,
                 ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004875. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptSetProperty(hObject,pszProperty,pbInput,cbInput,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall BCryptCloseAlgorithmProvider(BCRYPT_ALG_HANDLE hAlgorithm,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x00014000487b. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptCloseAlgorithmProvider(hAlgorithm,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptGenRandom(BCRYPT_ALG_HANDLE hAlgorithm,PUCHAR pbBuffer,ULONG cbBuffer,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004881. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptGenRandom(hAlgorithm,pbBuffer,cbBuffer,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall BCryptDestroyKey(BCRYPT_KEY_HANDLE hKey)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004887. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptDestroyKey(hKey);
  return NVar1;
}



NTSTATUS __stdcall
BCryptImportKeyPair(BCRYPT_ALG_HANDLE hAlgorithm,BCRYPT_KEY_HANDLE hImportKey,LPCWSTR pszBlobType,
                   BCRYPT_KEY_HANDLE *phKey,PUCHAR pbInput,ULONG cbInput,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x00014000488d. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptImportKeyPair(hAlgorithm,hImportKey,pszBlobType,phKey,pbInput,cbInput,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptEncrypt(BCRYPT_KEY_HANDLE hKey,PUCHAR pbInput,ULONG cbInput,void *pPaddingInfo,PUCHAR pbIV,
             ULONG cbIV,PUCHAR pbOutput,ULONG cbOutput,ULONG *pcbResult,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004893. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptEncrypt(hKey,pbInput,cbInput,pPaddingInfo,pbIV,cbIV,pbOutput,cbOutput,pcbResult,
                        dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptDecrypt(BCRYPT_KEY_HANDLE hKey,PUCHAR pbInput,ULONG cbInput,void *pPaddingInfo,PUCHAR pbIV,
             ULONG cbIV,PUCHAR pbOutput,ULONG cbOutput,ULONG *pcbResult,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004899. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptDecrypt(hKey,pbInput,cbInput,pPaddingInfo,pbIV,cbIV,pbOutput,cbOutput,pcbResult,
                        dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptGenerateSymmetricKey
          (BCRYPT_ALG_HANDLE hAlgorithm,BCRYPT_KEY_HANDLE *phKey,PUCHAR pbKeyObject,
          ULONG cbKeyObject,PUCHAR pbSecret,ULONG cbSecret,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x00014000489f. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptGenerateSymmetricKey
                    (hAlgorithm,phKey,pbKeyObject,cbKeyObject,pbSecret,cbSecret,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptGenerateKeyPair
          (BCRYPT_ALG_HANDLE hAlgorithm,BCRYPT_KEY_HANDLE *phKey,ULONG dwLength,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x0001400048a5. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptGenerateKeyPair(hAlgorithm,phKey,dwLength,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall
BCryptExportKey(BCRYPT_KEY_HANDLE hKey,BCRYPT_KEY_HANDLE hExportKey,LPCWSTR pszBlobType,
               PUCHAR pbOutput,ULONG cbOutput,ULONG *pcbResult,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x0001400048ab. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptExportKey(hKey,hExportKey,pszBlobType,pbOutput,cbOutput,pcbResult,dwFlags);
  return NVar1;
}



NTSTATUS __stdcall BCryptFinalizeKeyPair(BCRYPT_KEY_HANDLE hKey,ULONG dwFlags)

{
  NTSTATUS NVar1;
  
                    /* WARNING: Could not recover jumptable at 0x0001400048b1. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  NVar1 = BCryptFinalizeKeyPair(hKey,dwFlags);
  return NVar1;
}



/* Library Function - Single Match
    WdmlibRtlInitUnicodeStringEx
   
   Library: Visual Studio 2019 Release */

undefined8 WdmlibRtlInitUnicodeStringEx(ushort *param_1,ulonglong param_2)

{
  ushort uVar1;
  ulonglong uVar2;
  ulonglong uVar3;
  
  uVar2 = 0;
  if (param_2 == 0) {
    uVar1 = 0;
    uVar3 = uVar2;
  }
  else {
    uVar2 = 0xffffffffffffffff;
    do {
      uVar2 = uVar2 + 1;
    } while (*(short *)(param_2 + uVar2 * 2) != 0);
    if (0x7ffe < uVar2) {
      return 0xc0000106;
    }
    uVar1 = (short)uVar2 * 2;
    uVar2 = param_2;
    uVar3 = (ulonglong)(uVar1 + 2);
  }
  *param_1 = uVar1;
  param_1[1] = (ushort)uVar3;
  *(ulonglong *)(param_1 + 4) = uVar2;
  return 0;
}



int __cdecl _snwprintf(wchar_t *_Dest,size_t _Count,wchar_t *_Format,...)

{
  int iVar1;
  
                    /* WARNING: Could not recover jumptable at 0x0001400048fd. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  iVar1 = _snwprintf(_Dest,_Count,_Format);
  return iVar1;
}



int __cdecl _wcsnicmp(wchar_t *_Str1,wchar_t *_Str2,size_t _MaxCount)

{
  int iVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004903. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  iVar1 = _wcsnicmp(_Str1,_Str2,_MaxCount);
  return iVar1;
}



wchar_t * __cdecl wcschr(wchar_t *_Str,wchar_t _Ch)

{
  wchar_t *pwVar1;
  
                    /* WARNING: Could not recover jumptable at 0x000140004909. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  pwVar1 = wcschr(_Str,_Ch);
  return pwVar1;
}



void _guard_check_icall(void)

{
  return;
}



/* WARNING: This is an inlined function */

void _guard_dispatch_icall(void)

{
  code *UNRECOVERED_JUMPTABLE;
  
                    /* WARNING: Could not recover jumptable at 0x000140004930. Too many branches */
                    /* WARNING: Treating indirect jump as call */
  (*UNRECOVERED_JUMPTABLE)();
  return;
}



/* Library Function - Single Match
    memmove
   
   Library: Visual Studio 2019 Release */

void * __cdecl memmove(void *_Dst,void *_Src,size_t _Size)

{
  undefined4 *puVar1;
  undefined *puVar2;
  undefined8 uVar3;
  undefined auVar4 [16];
  undefined auVar5 [16];
  undefined auVar6 [16];
  undefined auVar7 [16];
  undefined auVar8 [16];
  undefined auVar9 [16];
  undefined auVar10 [16];
  undefined4 uVar11;
  undefined4 uVar12;
  undefined4 uVar13;
  undefined4 uVar14;
  undefined4 uVar15;
  undefined4 uVar16;
  undefined4 uVar17;
  undefined4 uVar18;
  undefined4 uVar19;
  undefined4 uVar20;
  undefined4 uVar21;
  undefined4 uVar22;
  undefined *puVar23;
  undefined4 *puVar24;
  undefined4 *puVar25;
  undefined (*pauVar26) [16];
  undefined (*pauVar27) [16];
  longlong lVar28;
  ulonglong uVar29;
  ulonglong uVar30;
  undefined4 uVar31;
  undefined4 uVar32;
  undefined4 uVar33;
  undefined4 uVar34;
  undefined4 uVar35;
  undefined4 uVar36;
  undefined4 uVar37;
  undefined4 uVar38;
  
  if (_Size < 8) {
    if (_Size != 0) {
      puVar23 = (undefined *)_Dst;
      if (_Src < _Dst) {
        puVar23 = (undefined *)((longlong)_Dst + _Size);
        do {
          puVar2 = puVar23 + ((longlong)_Src - (longlong)_Dst) + -1;
          puVar23 = puVar23 + -1;
          _Size = _Size - 1;
          *puVar23 = *puVar2;
        } while (_Size != 0);
        return _Dst;
      }
      do {
        _Size = _Size - 1;
        *puVar23 = puVar23[(longlong)_Src - (longlong)_Dst];
        puVar23 = puVar23 + 1;
      } while (_Size != 0);
    }
    return _Dst;
  }
  if (_Size < 0x11) {
                    /* WARNING: Load size is inaccurate */
    uVar3 = *(undefined8 *)((longlong)_Src + (_Size - 8));
    *(undefined8 *)_Dst = *_Src;
    *(undefined8 *)((longlong)_Dst + (_Size - 8)) = uVar3;
    return _Dst;
  }
  if (_Size < 0x21) {
                    /* WARNING: Load size is inaccurate */
    uVar31 = *(undefined4 *)((longlong)_Src + 4);
    uVar33 = *(undefined4 *)((longlong)_Src + 8);
    uVar35 = *(undefined4 *)((longlong)_Src + 0xc);
    puVar25 = (undefined4 *)((longlong)_Src + (_Size - 0x10));
    uVar32 = *puVar25;
    uVar34 = puVar25[1];
    uVar36 = puVar25[2];
    uVar38 = puVar25[3];
    *(undefined4 *)_Dst = *_Src;
    *(undefined4 *)((longlong)_Dst + 4) = uVar31;
    *(undefined4 *)((longlong)_Dst + 8) = uVar33;
    *(undefined4 *)((longlong)_Dst + 0xc) = uVar35;
    puVar25 = (undefined4 *)((longlong)_Dst + (_Size - 0x10));
    *puVar25 = uVar32;
    puVar25[1] = uVar34;
    puVar25[2] = uVar36;
    puVar25[3] = uVar38;
    return _Dst;
  }
  lVar28 = (longlong)_Src - (longlong)_Dst;
  if ((_Src < _Dst) && (_Dst < (void *)((longlong)_Src + _Size))) {
    puVar25 = (undefined4 *)((longlong)_Dst + lVar28 + (_Size - 0x10));
    uVar31 = puVar25[1];
    uVar33 = puVar25[2];
    uVar35 = puVar25[3];
    pauVar26 = (undefined (*) [16])((longlong)_Dst + (_Size - 0x10));
    uVar29 = _Size - 0x10;
    pauVar27 = pauVar26;
    uVar32 = *puVar25;
    uVar34 = uVar31;
    uVar36 = uVar33;
    uVar38 = uVar35;
    if (((ulonglong)pauVar26 & 0xf) != 0) {
      pauVar27 = (undefined (*) [16])((ulonglong)pauVar26 & 0xfffffffffffffff0);
      puVar24 = (undefined4 *)((longlong)pauVar27 + lVar28);
      uVar32 = *puVar24;
      uVar34 = puVar24[1];
      uVar36 = puVar24[2];
      uVar38 = puVar24[3];
      *(undefined4 *)*pauVar26 = *puVar25;
      *(undefined4 *)((longlong)_Dst + (_Size - 0xc)) = uVar31;
      *(undefined4 *)((longlong)_Dst + (_Size - 8)) = uVar33;
      *(undefined4 *)((longlong)_Dst + (_Size - 4)) = uVar35;
      uVar29 = (longlong)pauVar27 - (longlong)_Dst;
    }
    uVar30 = uVar29 >> 6;
    if (uVar30 != 0) {
      uVar29 = uVar29 & 0x3f;
      pauVar26 = pauVar27;
      uVar31 = uVar32;
      uVar33 = uVar34;
      uVar35 = uVar36;
      uVar37 = uVar38;
      do {
        puVar25 = (undefined4 *)((longlong)pauVar26 + lVar28 + -0x10);
        uVar11 = *puVar25;
        uVar12 = puVar25[1];
        uVar13 = puVar25[2];
        uVar14 = puVar25[3];
        puVar25 = (undefined4 *)((longlong)pauVar26 + lVar28 + -0x20);
        uVar15 = *puVar25;
        uVar16 = puVar25[1];
        uVar17 = puVar25[2];
        uVar18 = puVar25[3];
        puVar25 = (undefined4 *)((longlong)pauVar26 + lVar28 + -0x30);
        uVar19 = *puVar25;
        uVar20 = puVar25[1];
        uVar21 = puVar25[2];
        uVar22 = puVar25[3];
        puVar25 = (undefined4 *)((longlong)pauVar26 + lVar28 + -0x40);
        uVar32 = *puVar25;
        uVar34 = puVar25[1];
        uVar36 = puVar25[2];
        uVar38 = puVar25[3];
        auVar8._4_4_ = uVar33;
        auVar8._0_4_ = uVar31;
        auVar8._8_4_ = uVar35;
        auVar8._12_4_ = uVar37;
        *pauVar26 = auVar8;
        pauVar27 = pauVar26 + -4;
        uVar30 = uVar30 - 1;
        *(undefined4 *)pauVar26[-1] = uVar11;
        *(undefined4 *)(pauVar26[-1] + 4) = uVar12;
        *(undefined4 *)(pauVar26[-1] + 8) = uVar13;
        *(undefined4 *)(pauVar26[-1] + 0xc) = uVar14;
        *(undefined4 *)pauVar26[-2] = uVar15;
        *(undefined4 *)(pauVar26[-2] + 4) = uVar16;
        *(undefined4 *)(pauVar26[-2] + 8) = uVar17;
        *(undefined4 *)(pauVar26[-2] + 0xc) = uVar18;
        *(undefined4 *)pauVar26[-3] = uVar19;
        *(undefined4 *)(pauVar26[-3] + 4) = uVar20;
        *(undefined4 *)(pauVar26[-3] + 8) = uVar21;
        *(undefined4 *)(pauVar26[-3] + 0xc) = uVar22;
        pauVar26 = pauVar27;
        uVar31 = uVar32;
        uVar33 = uVar34;
        uVar35 = uVar36;
        uVar37 = uVar38;
      } while (uVar30 != 0);
    }
    for (uVar30 = uVar29 >> 4; uVar30 != 0; uVar30 = uVar30 - 1) {
      auVar9._4_4_ = uVar34;
      auVar9._0_4_ = uVar32;
      auVar9._8_4_ = uVar36;
      auVar9._12_4_ = uVar38;
      *pauVar27 = auVar9;
      puVar25 = (undefined4 *)((longlong)pauVar27 + lVar28 + -0x10);
      uVar32 = *puVar25;
      uVar34 = puVar25[1];
      uVar36 = puVar25[2];
      uVar38 = puVar25[3];
      pauVar27 = pauVar27 + -1;
    }
    if ((uVar29 & 0xf) != 0) {
      puVar24 = (undefined4 *)((longlong)pauVar27 - (uVar29 & 0xf));
      puVar25 = (undefined4 *)((longlong)puVar24 + lVar28);
      uVar31 = puVar25[1];
      uVar33 = puVar25[2];
      uVar35 = puVar25[3];
      *puVar24 = *puVar25;
      puVar24[1] = uVar31;
      puVar24[2] = uVar33;
      puVar24[3] = uVar35;
    }
    auVar10._4_4_ = uVar34;
    auVar10._0_4_ = uVar32;
    auVar10._8_4_ = uVar36;
    auVar10._12_4_ = uVar38;
    *pauVar27 = auVar10;
    return _Dst;
  }
  puVar25 = (undefined4 *)((longlong)_Dst + lVar28);
  uVar31 = puVar25[1];
  uVar33 = puVar25[2];
  uVar35 = puVar25[3];
  puVar24 = (undefined4 *)((longlong)_Dst + 0x10);
  uVar32 = *puVar25;
  uVar34 = uVar31;
  uVar36 = uVar33;
  uVar38 = uVar35;
  if (((ulonglong)puVar24 & 0xf) != 0) {
    puVar1 = (undefined4 *)(((ulonglong)puVar24 & 0xfffffffffffffff0) + lVar28);
    uVar32 = *puVar1;
    uVar34 = puVar1[1];
    uVar36 = puVar1[2];
    uVar38 = puVar1[3];
    *(undefined4 *)_Dst = *puVar25;
    *(undefined4 *)((longlong)_Dst + 4) = uVar31;
    *(undefined4 *)((longlong)_Dst + 8) = uVar33;
    *(undefined4 *)((longlong)_Dst + 0xc) = uVar35;
    puVar24 = (undefined4 *)(((ulonglong)puVar24 & 0xfffffffffffffff0) + 0x10);
  }
  uVar29 = (longlong)_Dst + (_Size - (longlong)puVar24);
  uVar30 = uVar29 >> 6;
  if (uVar30 != 0) {
    if (uVar30 < 0x1001) {
      uVar29 = uVar29 & 0x3f;
      puVar25 = puVar24;
      uVar31 = uVar32;
      uVar33 = uVar34;
      uVar35 = uVar36;
      uVar37 = uVar38;
      do {
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28);
        uVar11 = *puVar24;
        uVar12 = puVar24[1];
        uVar13 = puVar24[2];
        uVar14 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x10);
        uVar15 = *puVar24;
        uVar16 = puVar24[1];
        uVar17 = puVar24[2];
        uVar18 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x20);
        uVar19 = *puVar24;
        uVar20 = puVar24[1];
        uVar21 = puVar24[2];
        uVar22 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x30);
        uVar32 = *puVar24;
        uVar34 = puVar24[1];
        uVar36 = puVar24[2];
        uVar38 = puVar24[3];
        auVar4._4_4_ = uVar33;
        auVar4._0_4_ = uVar31;
        auVar4._8_4_ = uVar35;
        auVar4._12_4_ = uVar37;
        *(undefined (*) [16])(puVar25 + -4) = auVar4;
        puVar24 = puVar25 + 0x10;
        uVar30 = uVar30 - 1;
        *puVar25 = uVar11;
        puVar25[1] = uVar12;
        puVar25[2] = uVar13;
        puVar25[3] = uVar14;
        puVar25[4] = uVar15;
        puVar25[5] = uVar16;
        puVar25[6] = uVar17;
        puVar25[7] = uVar18;
        puVar25[8] = uVar19;
        puVar25[9] = uVar20;
        puVar25[10] = uVar21;
        puVar25[0xb] = uVar22;
        puVar25 = puVar24;
        uVar31 = uVar32;
        uVar33 = uVar34;
        uVar35 = uVar36;
        uVar37 = uVar38;
      } while (uVar30 != 0);
    }
    else {
      uVar30 = uVar29 >> 6;
      uVar29 = uVar29 & 0x3f;
      puVar25 = puVar24;
      uVar31 = uVar32;
      uVar33 = uVar34;
      uVar35 = uVar36;
      uVar37 = uVar38;
      do {
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28);
        uVar11 = *puVar24;
        uVar12 = puVar24[1];
        uVar13 = puVar24[2];
        uVar14 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x10);
        uVar15 = *puVar24;
        uVar16 = puVar24[1];
        uVar17 = puVar24[2];
        uVar18 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x20);
        uVar19 = *puVar24;
        uVar20 = puVar24[1];
        uVar21 = puVar24[2];
        uVar22 = puVar24[3];
        puVar24 = (undefined4 *)((longlong)puVar25 + lVar28 + 0x30);
        uVar32 = *puVar24;
        uVar34 = puVar24[1];
        uVar36 = puVar24[2];
        uVar38 = puVar24[3];
        auVar7._4_4_ = uVar33;
        auVar7._0_4_ = uVar31;
        auVar7._8_4_ = uVar35;
        auVar7._12_4_ = uVar37;
        *(undefined (*) [16])(puVar25 + -4) = auVar7;
        puVar24 = puVar25 + 0x10;
        uVar30 = uVar30 - 1;
        *puVar25 = uVar11;
        puVar25[1] = uVar12;
        puVar25[2] = uVar13;
        puVar25[3] = uVar14;
        puVar25[4] = uVar15;
        puVar25[5] = uVar16;
        puVar25[6] = uVar17;
        puVar25[7] = uVar18;
        puVar25[8] = uVar19;
        puVar25[9] = uVar20;
        puVar25[10] = uVar21;
        puVar25[0xb] = uVar22;
        puVar25 = puVar24;
        uVar31 = uVar32;
        uVar33 = uVar34;
        uVar35 = uVar36;
        uVar37 = uVar38;
      } while (uVar30 != 0);
    }
  }
  for (uVar30 = uVar29 >> 4; uVar30 != 0; uVar30 = uVar30 - 1) {
    auVar5._4_4_ = uVar34;
    auVar5._0_4_ = uVar32;
    auVar5._8_4_ = uVar36;
    auVar5._12_4_ = uVar38;
    *(undefined (*) [16])(puVar24 + -4) = auVar5;
    puVar25 = (undefined4 *)((longlong)puVar24 + lVar28);
    uVar32 = *puVar25;
    uVar34 = puVar25[1];
    uVar36 = puVar25[2];
    uVar38 = puVar25[3];
    puVar24 = puVar24 + 4;
  }
  if ((uVar29 & 0xf) != 0) {
    puVar1 = (undefined4 *)((longlong)puVar24 + ((uVar29 & 0xf) - 0x10));
    puVar25 = (undefined4 *)((longlong)puVar1 + lVar28);
    uVar31 = puVar25[1];
    uVar33 = puVar25[2];
    uVar35 = puVar25[3];
    *puVar1 = *puVar25;
    puVar1[1] = uVar31;
    puVar1[2] = uVar33;
    puVar1[3] = uVar35;
  }
  auVar6._4_4_ = uVar34;
  auVar6._0_4_ = uVar32;
  auVar6._8_4_ = uVar36;
  auVar6._12_4_ = uVar38;
  *(undefined (*) [16])(puVar24 + -4) = auVar6;
  return _Dst;
}



undefined (*) [16] memset(void *param_1,byte param_2,SIZE_T param_3)

{
  undefined4 *puVar1;
  undefined (*pauVar2) [16];
  undefined (*pauVar3) [16];
  undefined (*pauVar4) [16];
  undefined4 uVar5;
  longlong lVar6;
  undefined *puVar7;
  ulonglong uVar8;
  undefined (*pauVar9) [16];
  undefined auVar10 [16];
  undefined extraout_XMM0 [16];
  
  pauVar2 = (undefined (*) [16])param_1;
  lVar6 = (ulonglong)param_2 * 0x101010101010101;
  auVar10._8_8_ = lVar6;
  auVar10._0_8_ = lVar6;
  if (param_3 < 0x40) {
    if (param_3 < 0x10) {
      if (3 < param_3) {
        puVar1 = (undefined4 *)((param_3 - 4) + (longlong)param_1);
        uVar5 = (undefined4)lVar6;
        *(undefined4 *)param_1 = uVar5;
        uVar8 = (param_3 & 8) >> 1;
        *puVar1 = uVar5;
        *(undefined4 *)((longlong)param_1 + uVar8) = uVar5;
        *(undefined4 *)((longlong)puVar1 - uVar8) = uVar5;
        return pauVar2;
      }
      if (param_3 != 0) {
        *(char *)param_1 = (char)lVar6;
        if (param_3 != 1) {
          *(short *)((longlong)param_1 + (param_3 - 2)) = (short)lVar6;
        }
      }
      return pauVar2;
    }
  }
  else {
    if (((DAT_140007000 & 2) != 0) && (799 < param_3)) {
      if ((DAT_140007000 & 1) == 0) {
        param_1 = (void *)FUN_140004e00();
        auVar10 = extraout_XMM0;
      }
      *pauVar2 = auVar10;
      pauVar2[1] = auVar10;
      pauVar2[2] = auVar10;
      pauVar2[3] = auVar10;
      puVar7 = (undefined *)((ulonglong)(pauVar2 + 4) & 0xffffffffffffffc0);
      for (lVar6 = (longlong)pauVar2 +
                   (param_3 - (longlong)(undefined *)((ulonglong)(pauVar2 + 4) & 0xffffffffffffffc0)
                   ); lVar6 != 0; lVar6 = lVar6 + -1) {
        *puVar7 = auVar10[0];
        puVar7 = puVar7 + 1;
      }
      return (undefined (*) [16])param_1;
    }
    *(undefined (*) [16])param_1 = auVar10;
    pauVar3 = (undefined (*) [16])((longlong)param_1 + 0x10U & 0xfffffffffffffff0);
    param_3 = (longlong)param_1 + (param_3 - (longlong)pauVar3);
    param_1 = pauVar3;
    if (0x3f < param_3) {
      pauVar9 = (undefined (*) [16])((longlong)pauVar3 + (param_3 - 0x30) & 0xfffffffffffffff0);
      uVar8 = param_3 >> 6;
      pauVar4 = pauVar3;
      do {
        *pauVar4 = auVar10;
        pauVar4[1] = auVar10;
        uVar8 = uVar8 - 1;
        pauVar4[2] = auVar10;
        pauVar4[3] = auVar10;
        pauVar4 = pauVar4 + 4;
      } while (uVar8 != 0);
      *pauVar9 = auVar10;
      pauVar9[1] = auVar10;
      pauVar9[2] = auVar10;
      *(undefined (*) [16])((longlong)pauVar3 + (param_3 - 0x10)) = auVar10;
      return pauVar2;
    }
  }
  *(undefined (*) [16])param_1 = auVar10;
  uVar8 = (param_3 & 0x20) >> 1;
  *(undefined (*) [16])(*(undefined (*) [16])((longlong)param_1 + -0x10) + param_3) = auVar10;
  *(undefined (*) [16])((longlong)param_1 + uVar8) = auVar10;
  *(undefined (*) [16])
   ((longlong)(undefined (*) [16])(*(undefined (*) [16])((longlong)param_1 + -0x10) + param_3) -
   uVar8) = auVar10;
  return pauVar2;
}



undefined8 FUN_140004e00(void)

{
  undefined8 in_RAX;
  
  FUN_140004730();
  return in_RAX;
}



/* WARNING: Globals starting with '_' overlap smaller symbols at the same address */
/* Library Function - Single Match
    WdmlibInit
   
   Library: Visual Studio 2019 Release */

void WdmlibInit(void)

{
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  RtlInitUnicodeString(local_18,L"IoCreateDeviceSecure");
  DAT_140007268 = (undefined *)MmGetSystemRoutineAddress(local_18);
  if (DAT_140007268 == (undefined *)0x0) {
    DAT_140007268 = &LAB_140009124;
  }
  RtlInitUnicodeString(local_18,L"IoValidateDeviceIoControlAccess");
  _DAT_140007260 = MmGetSystemRoutineAddress(local_18);
  DAT_140007270 = 1;
  return;
}



/* Library Function - Single Match
    WdmlibIoCreateDeviceSecure
   
   Library: Visual Studio 2019 Release */

void WdmlibIoCreateDeviceSecure
               (undefined8 param_1,undefined4 param_2,undefined8 param_3,undefined4 param_4,
               undefined4 param_5,undefined param_6,undefined8 param_7,undefined8 param_8,
               undefined8 param_9)

{
  if (DAT_140007270 == '\0') {
    WdmlibInit();
  }
  (*DAT_140007268)(param_1,param_2,param_3,param_4,param_5,param_6,param_7,param_8,param_9);
  return;
}



/* Library Function - Single Match
    IopDevObjApplyPostCreationSettings
   
   Library: Visual Studio 2019 Release */

ulonglong IopDevObjApplyPostCreationSettings(undefined8 param_1,uint *param_2)

{
  uint uVar1;
  ulonglong uVar2;
  undefined local_res10 [8];
  uint local_res18 [2];
  uint local_res20 [2];
  undefined8 local_28 [2];
  
  local_res18[0] = 0;
  local_res20[0] = 0;
  local_28[0] = 0;
  local_res10[0] = 0;
  if ((*param_2 & 2) == 0) {
    uVar2 = 0;
  }
  else {
    uVar2 = SeUtilSecurityInfoFromSecurityDescriptor
                      (*(undefined8 *)(param_2 + 2),local_res10,local_res18);
    if (-1 < (int)uVar2) {
      SeSetSecurityAccessMask((byte)local_res18[0],local_res20);
      uVar2 = ObOpenObjectByPointer
                        (param_1,0x200,0,local_res20[0],*(undefined8 *)IoDeviceObjectType_exref,0,
                         local_28);
      if (-1 < (int)uVar2) {
        uVar1 = ZwSetSecurityObject(local_28[0],local_res18[0],*(undefined8 *)(param_2 + 2));
        uVar2 = (ulonglong)uVar1;
        ZwClose(local_28[0]);
      }
    }
  }
  return uVar2;
}



/* Library Function - Single Match
    SeSetSecurityAccessMask
   
   Library: Visual Studio 2019 Release */

void SeSetSecurityAccessMask(byte param_1,uint *param_2)

{
  uint uVar1;
  
  uVar1 = 0;
  *param_2 = 0;
  if ((param_1 & 3) != 0) {
    uVar1 = 0x80000;
    *param_2 = 0x80000;
  }
  if ((param_1 & 4) != 0) {
    uVar1 = uVar1 | 0x40000;
    *param_2 = uVar1;
  }
  if ((param_1 & 8) != 0) {
    *param_2 = uVar1 | 0x1000000;
  }
  return;
}



/* Library Function - Single Match
    SeUtilSecurityInfoFromSecurityDescriptor
   
   Library: Visual Studio 2019 Release */

undefined8
SeUtilSecurityInfoFromSecurityDescriptor(undefined8 param_1,undefined *param_2,uint *param_3)

{
  undefined8 uVar1;
  uint uVar2;
  undefined local_res10 [8];
  char local_res20 [8];
  longlong local_28;
  undefined8 local_20;
  
  local_28 = 0;
  local_20 = 0;
  *param_2 = 0;
  *param_3 = 0;
  local_res10[0] = 0;
  local_res20[0] = '\0';
  uVar1 = RtlGetOwnerSecurityDescriptor(param_1,&local_28,local_res10);
  if (-1 < (int)uVar1) {
    uVar2 = (uint)(local_28 != 0);
    uVar1 = RtlGetGroupSecurityDescriptor(param_1,&local_28,local_res10);
    if (-1 < (int)uVar1) {
      if (local_28 != 0) {
        uVar2 = uVar2 | 2;
      }
      uVar1 = RtlGetSaclSecurityDescriptor(param_1,local_res20,&local_20,local_res10);
      if (-1 < (int)uVar1) {
        if (local_res20[0] != '\0') {
          uVar2 = uVar2 | 8;
        }
        uVar1 = RtlGetDaclSecurityDescriptor(param_1,local_res20,&local_20,local_res10);
        if (-1 < (int)uVar1) {
          if (local_res20[0] != '\0') {
            uVar2 = uVar2 | 4;
          }
          *param_2 = local_res10[0];
          uVar1 = 0;
          *param_3 = uVar2;
        }
      }
    }
  }
  return uVar1;
}



/* Library Function - Single Match
    PiRegStateOpenClassKey
   
   Library: Visual Studio 2019 Release */

void PiRegStateOpenClassKey
               (undefined8 param_1,undefined4 param_2,int param_3,undefined4 *param_4,
               undefined8 *param_5)

{
  int iVar1;
  undefined auStackY_128 [32];
  undefined4 local_b8;
  undefined4 local_b4;
  undefined8 local_b0;
  int local_a8;
  undefined8 local_a0;
  wchar_t local_98 [38];
  undefined2 local_4c;
  ulonglong local_48;
  
  local_48 = DAT_140007010 ^ (ulonglong)auStackY_128;
  local_b0 = 0;
  *param_5 = 0;
  local_a0 = 0;
  local_b8 = 0;
  if (param_4 != (undefined4 *)0x0) {
    *param_4 = 0;
  }
  local_b4 = param_2;
  local_a8 = param_3;
  iVar1 = CmRegUtilOpenExistingWstrKey(0,0x140006070,0x20019,&local_b0);
  if (-1 < iVar1) {
    _snwprintf(local_98,0x27,L"{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}");
    local_4c = 0;
    if (local_a8 == 0) {
      iVar1 = CmRegUtilOpenExistingWstrKey(local_b0,(ulonglong)local_98,local_b4,&local_a0);
      local_b8 = 2;
    }
    else {
      iVar1 = CmRegUtilCreateWstrKey(local_b0,(ulonglong)local_98,local_b4,0,0,&local_b8,&local_a0);
    }
    ZwClose(local_b0);
    if ((-1 < iVar1) && (*param_5 = local_a0, param_4 != (undefined4 *)0x0)) {
      *param_4 = local_b8;
    }
  }
  __check_stack_cookie(local_48 ^ (ulonglong)auStackY_128);
  return;
}



ulonglong FUN_140009620(undefined8 param_1,uint *param_2)

{
  uint *puVar1;
  uint *puVar2;
  uint *puVar3;
  uint uVar4;
  ulonglong uVar5;
  ulonglong uVar6;
  char local_res10 [8];
  uint local_res18 [2];
  longlong local_res20;
  undefined (*local_48 [2]) [16];
  
  puVar1 = param_2 + 1;
  puVar2 = param_2 + 4;
  *param_2 = 0;
  puVar3 = param_2 + 5;
  *(undefined8 *)(param_2 + 2) = 0;
  local_res20 = 0;
  local_res18[0] = 0;
  local_res10[0] = '\0';
  local_48[0] = (undefined (*) [16])0x0;
  *puVar1 = 0;
  *puVar2 = 0;
  *puVar3 = 0;
  uVar4 = CmRegUtilWstrValueGetFullBuffer(param_1,0x140005ff0,3,0,local_48);
  if ((int)uVar4 < 0) {
    uVar6 = (ulonglong)uVar4;
    if (uVar4 == 0xc0000034) {
      uVar6 = 0;
    }
  }
  else {
    uVar4 = SeCaptureSecurityDescriptor
                      (*local_48[0] + *(uint *)(*local_48[0] + 8),0,1,1,&local_res20);
    uVar6 = (ulonglong)uVar4;
  }
  if (local_48[0] != (undefined (*) [16])0x0) {
    ExFreePoolWithTag(local_48[0],0);
  }
  if (-1 < (int)uVar6) {
    if (local_res20 != 0) {
      uVar5 = SeUtilSecurityInfoFromSecurityDescriptor(local_res20,local_res10,local_res18);
      uVar6 = uVar5 & 0xffffffff;
      if ((int)uVar5 < 0) goto LAB_1400097a3;
      if (local_res10[0] == '\0') {
        *param_2 = *param_2 | 2;
        *(longlong *)(param_2 + 2) = local_res20;
      }
      else {
        ExFreePoolWithTag(local_res20,0);
      }
    }
    uVar4 = CmRegUtilWstrValueGetDword(param_1,0x140006008,0,puVar1);
    uVar6 = (ulonglong)uVar4;
    if ((int)uVar4 < 0) {
      if (uVar4 != 0xc0000034) goto LAB_1400097a3;
    }
    else {
      *param_2 = *param_2 | 1;
    }
    uVar4 = CmRegUtilWstrValueGetDword(param_1,0x140006020,0,puVar2);
    uVar6 = (ulonglong)uVar4;
    if ((int)uVar4 < 0) {
      if (uVar4 != 0xc0000034) goto LAB_1400097a3;
    }
    else {
      *param_2 = *param_2 | 4;
    }
    uVar4 = CmRegUtilWstrValueGetDword(param_1,0x140006050,0,puVar3);
    uVar6 = (ulonglong)uVar4;
    if (-1 < (int)uVar4) {
      *param_2 = *param_2 | 8;
      return uVar6;
    }
    if (uVar4 == 0xc0000034) {
      return 0;
    }
  }
LAB_1400097a3:
  if (*(longlong *)(param_2 + 2) != 0) {
    ExFreePoolWithTag(*(longlong *)(param_2 + 2),0);
  }
  *param_2 = 0;
  *(undefined8 *)(param_2 + 2) = 0;
  *puVar1 = 0;
  *puVar2 = 0;
  *puVar3 = 0;
  return uVar6;
}



void FUN_1400097e8(uint *param_1)

{
  if ((*param_1 & 2) != 0) {
    ExFreePoolWithTag(*(undefined8 *)(param_1 + 2),0);
  }
  return;
}



void FUN_14000980c(undefined8 *param_1)

{
  *param_1 = 0;
  param_1[1] = 0;
  param_1[2] = 0;
  return;
}



void FUN_14000981c(undefined8 param_1,undefined4 *param_2)

{
  *param_2 = 2;
  *(undefined8 *)(param_2 + 2) = param_1;
  return;
}



/* Library Function - Single Match
    PpRegStateReadCreateClassCreationSettings
   
   Library: Visual Studio 2019 Release */

ulonglong PpRegStateReadCreateClassCreationSettings
                    (undefined8 param_1,longlong param_2,uint *param_3)

{
  uint uVar1;
  ulonglong uVar2;
  undefined8 uVar3;
  ushort *puVar4;
  int local_res18 [2];
  undefined8 local_res20;
  undefined8 local_18 [2];
  
  *param_3 = 0;
  *(undefined8 *)(param_3 + 2) = 0;
  param_3[1] = 0;
  param_3[4] = 0;
  param_3[5] = 0;
  local_res18[0] = 0;
  local_res20 = 0;
  local_18[0] = 0;
  uVar2 = PiRegStateOpenClassKey(param_1,0xf003f,1,local_res18,&local_res20);
  uVar3 = local_res20;
  if (-1 < (int)uVar2) {
    if (local_res18[0] == 2) {
      uVar1 = CmRegUtilOpenExistingWstrKey(local_res20,0x140005f90,0x20019,local_18);
      ZwClose(local_res20);
      if ((int)uVar1 < 0) {
        if (uVar1 != 0xc0000034) {
          return (ulonglong)uVar1;
        }
        return 0;
      }
      uVar2 = FUN_140009620(local_18[0],param_3);
      uVar2 = uVar2 & 0xffffffff;
      uVar3 = local_18[0];
    }
    else {
      puVar4 = (ushort *)(*(longlong *)(param_2 + 0x30) + 0x18);
      if (puVar4 == (ushort *)0x0) {
        uVar2 = 0xc000009a;
      }
      else {
        uVar1 = CmRegUtilWstrValueSetUcString(local_res20,0x140005fa8,puVar4);
        uVar2 = (ulonglong)uVar1;
        if (-1 < (int)uVar1) {
          local_res18[0] = 0x31;
          uVar1 = CmRegUtilWstrValueSetWstrString(uVar3,0x140005fb8,(ulonglong)local_res18);
          uVar2 = (ulonglong)uVar1;
          if (-1 < (int)uVar1) {
            uVar1 = CmRegUtilWstrValueSetWstrString(uVar3,0x140005fd8,(ulonglong)local_res18);
            uVar2 = (ulonglong)uVar1;
          }
        }
      }
    }
    ZwClose(uVar3);
  }
  return uVar2;
}



/* Library Function - Single Match
    PpRegStateUpdateStackCreationSettings
   
   Library: Visual Studio 2019 Release */

ulonglong PpRegStateUpdateStackCreationSettings(undefined8 param_1,longlong param_2)

{
  int iVar1;
  uint uVar2;
  undefined4 uVar3;
  ulonglong uVar4;
  undefined *local_res18;
  undefined8 local_res20;
  undefined8 local_18 [2];
  
  local_res18 = (undefined *)0x0;
  local_18[0] = 0;
  local_res20 = 0;
  uVar4 = PiRegStateOpenClassKey(param_1,0xf003f,0,(undefined4 *)0x0,&local_res20);
  if (-1 < (int)uVar4) {
    if (DAT_140007274 == 0) {
      iVar1 = SeCaptureSecurityDescriptor(&DAT_140007020,0,1,1,&local_res18);
      if (iVar1 < 0) {
        DAT_140007274 = 2;
      }
      else {
        DAT_140007274 = 1;
        ExFreePoolWithTag(local_res18,0);
      }
    }
    local_res18 = &DAT_140007020;
    if (DAT_140007274 != 1) {
      local_res18 = (undefined *)0x0;
    }
    uVar2 = CmRegUtilCreateWstrKey
                      (local_res20,0x140005f90,0xf003f,0,local_res18,(undefined4 *)0x0,local_18);
    ZwClose(local_res20);
    if (-1 < (int)uVar2) {
      uVar3 = RtlLengthSecurityDescriptor(*(undefined8 *)(param_2 + 8));
      uVar2 = CmRegUtilWstrValueSetFullBuffer
                        (local_18[0],0x140005ff0,3,*(undefined8 *)(param_2 + 8),uVar3);
      ZwClose(local_18[0]);
    }
    uVar4 = (ulonglong)uVar2;
  }
  return uVar4;
}



ulonglong FUN_140009ab0(ushort *param_1,int param_2,undefined (**param_3) [16])

{
  ulonglong uVar1;
  short *_Dst;
  
  uVar1 = (ulonglong)*param_1 + 2;
  if ((param_1[1] == uVar1) && ((*(short **)(param_1 + 4))[*param_1 >> 1] == 0)) {
    uVar1 = FUN_14000a384(*(short **)(param_1 + 4),param_2,param_3);
    return uVar1;
  }
  _Dst = (short *)ExAllocatePoolWithTag(1,uVar1,0x73546553);
  if (_Dst == (short *)0x0) {
    *param_3 = (undefined (*) [16])0x0;
    uVar1 = 0xc000009a;
  }
  else {
    memset(_Dst,0,(ulonglong)*param_1 + 2);
    memmove(_Dst,*(void **)(param_1 + 4),(ulonglong)*param_1);
    _Dst[*param_1 >> 1] = 0;
    uVar1 = FUN_14000a384(_Dst,param_2,param_3);
    uVar1 = uVar1 & 0xffffffff;
    ExFreePoolWithTag(_Dst,0);
  }
  return uVar1;
}



/* Library Function - Single Match
    SepSddlAddAceToAcl
   
   Library: Visual Studio 2019 Release */

undefined8
SepSddlAddAceToAcl(undefined (**param_1) [16],uint *param_2,undefined8 param_3,undefined8 param_4,
                  undefined4 param_5,int param_6,undefined8 param_7)

{
  uint uVar1;
  undefined (*pauVar2) [16];
  int iVar3;
  undefined (*_Dst) [16];
  undefined8 uVar4;
  uint uVar5;
  uint uVar6;
  
  pauVar2 = *param_1;
  uVar1 = *param_2;
  iVar3 = RtlLengthSid(param_7);
  uVar6 = iVar3 + 8 + uVar1;
  _Dst = pauVar2;
  if (*(ushort *)((longlong)*pauVar2 + 2) < uVar6) {
    uVar5 = (iVar3 + 8) * param_6 + uVar1;
    _Dst = (undefined (*) [16])ExAllocatePoolWithTag(1,uVar5,0x6c416553);
    if (_Dst == (undefined (*) [16])0x0) {
      return 0xc000009a;
    }
    memset(_Dst,0,(ulonglong)uVar5);
    memmove(_Dst,*param_1,(ulonglong)uVar1);
    *(short *)((longlong)*_Dst + 2) = (short)uVar5;
    ExFreePoolWithTag(pauVar2,0);
    *param_1 = _Dst;
  }
  *param_2 = uVar6;
  uVar4 = RtlAddAccessAllowedAce(_Dst,2,param_5,param_7);
  return uVar4;
}



/* Library Function - Single Match
    SepSddlDaclFromSDDLString
   
   Library: Visual Studio 2019 Release */

ulonglong SepSddlDaclFromSDDLString
                    (short *param_1,int param_2,uint *param_3,undefined (**param_4) [16])

{
  ulonglong uVar1;
  wchar_t *pwVar2;
  ulonglong uVar3;
  uint uVar4;
  undefined (*local_res8) [16];
  wchar_t *local_res18;
  
  *param_4 = (undefined (*) [16])0x0;
  *param_3 = 0;
  local_res8 = (undefined (*) [16])0x0;
  local_res18 = (wchar_t *)0x0;
  for (; *param_1 == 0x20; param_1 = param_1 + 1) {
  }
  if ((*param_1 == 0x44) && (param_1[1] == 0x3a)) {
    pwVar2 = param_1 + 2;
    uVar4 = -(uint)(param_2 != 0) & 8;
    if (*pwVar2 == L'P') {
      pwVar2 = param_1 + 3;
      uVar4 = (-(uint)(param_2 != 0) & 8) + 0x1000;
    }
    uVar1 = FUN_140009d74(pwVar2,&local_res8,&local_res18);
    uVar3 = uVar1 & 0xffffffff;
    pwVar2 = local_res18;
    if ((int)uVar1 == 0) {
      for (; *pwVar2 == L' '; pwVar2 = pwVar2 + 1) {
      }
      if (*pwVar2 == L'\0') {
        *param_4 = local_res8;
        *param_3 = uVar4;
        return uVar3;
      }
      uVar3 = 0xc000000d;
    }
    if (local_res8 != (undefined (*) [16])0x0) {
      ExFreePoolWithTag(local_res8,0);
    }
  }
  else {
    uVar3 = 0xc000000d;
  }
  return uVar3;
}



ulonglong FUN_140009d74(wchar_t *param_1,undefined (**param_2) [16],wchar_t **param_3)

{
  wchar_t wVar1;
  int iVar2;
  wchar_t *pwVar3;
  ulonglong uVar4;
  longlong lVar5;
  undefined (*pauVar6) [16];
  undefined8 uVar7;
  ulonglong uVar8;
  wchar_t *pwVar9;
  uint uVar10;
  uint uVar11;
  uint uVar12;
  uint uVar13;
  ulonglong uVar14;
  undefined8 local_res8;
  uint local_res10 [2];
  uint local_res18 [2];
  wchar_t *local_res20;
  ulonglong local_58 [3];
  
  uVar4 = 0;
  *param_2 = (undefined (*) [16])0x0;
  local_res10[0] = 0;
  local_res20 = (wchar_t *)0x0;
  local_58[0] = 0;
  iVar2 = 0;
  pwVar3 = wcschr(param_1,L':');
  *param_3 = pwVar3;
  if (pwVar3 == param_1) {
    uVar4 = 0xc000000d;
  }
  else {
    if (pwVar3 == (wchar_t *)0x0) {
      lVar5 = -1;
      do {
        lVar5 = lVar5 + 1;
      } while (param_1[lVar5] != L'\0');
      pwVar3 = param_1 + lVar5;
    }
    else {
      pwVar3 = pwVar3 + -1;
    }
    *param_3 = pwVar3;
    pwVar9 = param_1;
    uVar8 = uVar4;
    uVar14 = uVar4;
    if (param_1 < pwVar3) {
      do {
        if (*pwVar9 == L';') {
          uVar14 = (ulonglong)((int)uVar14 + 1);
        }
        else if (*pwVar9 != L' ') {
          uVar8 = 1;
        }
        iVar2 = (int)uVar8;
        pwVar9 = pwVar9 + 1;
      } while (pwVar9 < pwVar3);
    }
    uVar12 = (uint)(uVar14 / 5);
    if (((int)uVar14 == uVar12 * 5) && (((int)uVar14 != 0 || (iVar2 == 0)))) {
      if (uVar12 == 0) {
        pauVar6 = (undefined (*) [16])ExAllocatePoolWithTag(1,8);
        *param_2 = pauVar6;
        if (pauVar6 != (undefined (*) [16])0x0) {
          *(undefined8 *)*pauVar6 = 0;
          (**param_2)[0] = 2;
          (**param_2)[1] = 0;
          *(undefined2 *)(**param_2 + 2) = 8;
          *(undefined2 *)(**param_2 + 4) = 0;
          *(undefined2 *)(**param_2 + 6) = 0;
          return 0;
        }
      }
      else {
        uVar13 = uVar12 * 0x30 + 8;
        if (0xffff < uVar13) {
          uVar13 = 0xffff;
        }
        pauVar6 = (undefined (*) [16])ExAllocatePoolWithTag(1,(ulonglong)uVar13,0x6c416553);
        *param_2 = pauVar6;
        if (pauVar6 != (undefined (*) [16])0x0) {
          local_res18[0] = 8;
          memset(pauVar6,0,(ulonglong)uVar13);
          uVar11 = 0;
          (**param_2)[0] = 2;
          (**param_2)[1] = 0;
          *(short *)(**param_2 + 2) = (short)uVar13;
          *(undefined2 *)(**param_2 + 4) = 0;
          *(undefined2 *)(**param_2 + 6) = 0;
          uVar13 = 8;
          if (uVar12 != 0) {
            do {
              uVar10 = 0;
              local_res8 = (wchar_t *)((ulonglong)local_res8 & 0xffffffff00000000);
              for (; *param_1 == L' '; param_1 = param_1 + 1) {
              }
              if (*param_1 == L'(') {
                do {
                  param_1 = param_1 + 1;
                } while (*param_1 == L' ');
              }
              iVar2 = _wcsnicmp(param_1,L"A",1);
              if (iVar2 != 0) {
LAB_14000a102:
                uVar4 = 0xc000000d;
LAB_14000a105:
                pauVar6 = *param_2;
                goto LAB_14000a0ea;
              }
              for (pwVar3 = param_1 + 2; *pwVar3 == L' '; pwVar3 = pwVar3 + 1) {
              }
              if (*pwVar3 != L';') goto LAB_14000a102;
              do {
                pwVar3 = pwVar3 + 1;
                wVar1 = *pwVar3;
              } while (wVar1 == L' ');
              while (wVar1 != L';') {
                while (wVar1 == L' ') {
                  pwVar3 = pwVar3 + 1;
                  wVar1 = *pwVar3;
                }
                uVar7 = SepSddlLookupAccessMaskInTable(pwVar3,local_res10,&local_res20);
                if ((int)uVar7 == 0) {
                  FUN_14000a280((ushort *)pwVar3,(ushort **)&local_res20,(uint *)&local_res8);
                  if (local_res20 == pwVar3) goto LAB_14000a102;
                  uVar10 = (uint)local_res8;
                }
                else {
                  local_res8 = (wchar_t *)CONCAT44(local_res8._4_4_,uVar10 | local_res10[0]);
                  uVar10 = uVar10 | local_res10[0];
                }
                pwVar3 = local_res20;
                wVar1 = *local_res20;
              }
              pwVar3 = pwVar3 + 1;
              if ((int)uVar4 != 0) goto LAB_14000a105;
              lVar5 = 2;
              do {
                for (; *pwVar3 == L' '; pwVar3 = pwVar3 + 1) {
                }
                if (*pwVar3 != L';') {
                  uVar4 = 0xc000000d;
                }
                pwVar3 = pwVar3 + 1;
                lVar5 = lVar5 + -1;
              } while (lVar5 != 0);
              if ((int)uVar4 != 0) goto LAB_14000a105;
              for (; *pwVar3 == L' '; pwVar3 = pwVar3 + 1) {
              }
              local_res8 = (wchar_t *)0x0;
              uVar8 = FUN_14000a12c(pwVar3,local_58,(wchar_t **)&local_res8);
              uVar4 = uVar8 & 0xffffffff;
              uVar13 = local_res18[0];
              if ((int)uVar8 != 0) break;
              pwVar3 = local_res8;
              if (local_res8 == (wchar_t *)0x0) {
LAB_14000a071:
                uVar4 = 0xc0000077;
                goto LAB_14000a105;
              }
              for (; *pwVar3 == L' '; pwVar3 = pwVar3 + 1) {
              }
              if (*pwVar3 != L')') goto LAB_14000a071;
              param_1 = pwVar3 + 1;
              if (local_58[0] != 0) {
                uVar8 = SepSddlAddAceToAcl(param_2,local_res18,0,0,uVar10,uVar12 - uVar11,
                                           local_58[0]);
                uVar4 = uVar8 & 0xffffffff;
                if ((int)uVar8 != 0) goto LAB_14000a105;
              }
              if (*param_1 == L'(') {
                param_1 = pwVar3 + 2;
              }
              uVar11 = uVar11 + 1;
              uVar13 = local_res18[0];
            } while (uVar11 < uVar12);
          }
          pauVar6 = *param_2;
          if ((int)uVar4 == 0) {
            *(short *)(*pauVar6 + 2) = (short)uVar13;
            return uVar4;
          }
LAB_14000a0ea:
          ExFreePoolWithTag(pauVar6,0);
          *param_2 = (undefined (*) [16])0x0;
          return uVar4;
        }
      }
      uVar4 = 0xc000009a;
    }
    else {
      uVar4 = 0xc000000d;
    }
  }
  return uVar4;
}



undefined8 FUN_14000a12c(wchar_t *param_1,ulonglong *param_2,wchar_t **param_3)

{
  char cVar1;
  int iVar2;
  undefined8 uVar3;
  ulonglong uVar4;
  uint *puVar5;
  uint uVar6;
  ulonglong uVar7;
  
  uVar4 = 0;
  puVar5 = &DAT_140007064;
  uVar7 = uVar4;
  do {
    iVar2 = _wcsnicmp(param_1,(wchar_t *)(&DAT_14000705c + uVar7 * 0x18),(ulonglong)*puVar5);
    if (iVar2 == 0) {
      uVar6 = (&DAT_140007064)[uVar7 * 6];
      *param_3 = param_1 + uVar6;
      if ((&DAT_140007058)[uVar7 * 6] == 1) {
        cVar1 = IoIsWdmVersionAvailable(1,CONCAT71((int7)((ulonglong)(param_1 + uVar6) >> 8),0x20));
        if (cVar1 != '\0') goto LAB_14000a1bc;
      }
      else {
LAB_14000a1bc:
        uVar4 = *(ulonglong *)((&DAT_140007050)[uVar7 * 3] + *(longlong *)SeExports_exref);
      }
      uVar3 = 0;
      goto LAB_14000a1d1;
    }
    uVar6 = (int)uVar7 + 1;
    uVar7 = (ulonglong)uVar6;
    puVar5 = puVar5 + 6;
  } while (uVar6 < 0xe);
  uVar3 = 0xc0000073;
LAB_14000a1d1:
  *param_2 = uVar4;
  return uVar3;
}



/* Library Function - Single Match
    SepSddlLookupAccessMaskInTable
   
   Library: Visual Studio 2019 Release */

undefined8 SepSddlLookupAccessMaskInTable(wchar_t *param_1,undefined4 *param_2,wchar_t **param_3)

{
  int iVar1;
  uint uVar2;
  undefined **ppuVar3;
  
  ppuVar3 = &PTR_DAT_1400071a0;
  uVar2 = 0;
  do {
    iVar1 = _wcsnicmp(param_1,(wchar_t *)*ppuVar3,(ulonglong)*(uint *)((wchar_t **)ppuVar3 + 1));
    if (iVar1 == 0) {
      *param_2 = (&DAT_1400071ac)[(ulonglong)uVar2 * 4];
      *param_3 = param_1 + (uint)(&DAT_1400071a8)[(ulonglong)uVar2 * 4];
      return 1;
    }
    uVar2 = uVar2 + 1;
    ppuVar3 = (undefined **)((wchar_t **)ppuVar3 + 2);
  } while (uVar2 < 8);
  *param_2 = 0;
  *param_3 = param_1;
  return 0;
}



undefined8 FUN_14000a280(ushort *param_1,ushort **param_2,uint *param_3)

{
  ushort uVar1;
  int iVar2;
  uint uVar3;
  uint uVar4;
  short sVar5;
  uint uVar6;
  int iVar7;
  ushort *puVar8;
  
  *param_3 = 0;
  *param_2 = param_1;
  uVar1 = *param_1;
  uVar6 = (uint)uVar1;
  if (uVar1 == 0x30) {
    if ((param_1[1] - 0x58 & 0xffdf) != 0) {
LAB_14000a303:
      iVar7 = 10;
      uVar4 = 0;
      puVar8 = param_1;
      goto LAB_14000a2e1;
    }
    param_1 = param_1 + 2;
    uVar6 = (uint)*param_1;
  }
  else if ((ushort)(uVar1 - 0x30) < 10) goto LAB_14000a303;
  iVar7 = 0x10;
  puVar8 = param_1;
  uVar4 = 0;
  if ((short)uVar6 == 0) {
    return 0;
  }
LAB_14000a2e1:
  do {
    sVar5 = (short)uVar6;
    if ((ushort)(sVar5 - 0x30U) < 10) {
      iVar2 = uVar6 - 0x30;
    }
    else {
      if (iVar7 != 0x10) break;
      if ((ushort)(sVar5 - 0x41U) < 6) {
        iVar2 = uVar6 - 0x37;
      }
      else {
        if (5 < (ushort)(sVar5 - 0x61U)) break;
        iVar2 = uVar6 - 0x57;
      }
    }
    uVar3 = iVar7 * uVar4 + iVar2;
    if (uVar3 < uVar4) {
      return 0;
    }
    param_1 = param_1 + 1;
    uVar6 = (uint)*param_1;
    uVar4 = uVar3;
  } while (*param_1 != 0);
  if (param_1 == puVar8) {
    return 0;
  }
  *param_2 = param_1;
  *param_3 = uVar4;
  return 1;
}



ulonglong FUN_14000a384(short *param_1,int param_2,undefined (**param_3) [16])

{
  undefined (*pauVar1) [16];
  uint uVar2;
  code *pcVar3;
  ulonglong uVar4;
  ulonglong uVar5;
  undefined (*pauVar6) [16];
  uint local_res20 [2];
  uint local_68 [2];
  undefined (*local_60) [16];
  undefined local_58 [16];
  undefined local_48 [4];
  undefined auStack_44 [12];
  undefined local_38 [16];
  undefined8 local_28;
  
  local_58 = ZEXT816(0);
  RtlInitUnicodeString(local_58,L"SeConvertStringSecurityDescriptorToSecurityDescriptor");
  pcVar3 = (code *)MmGetSystemRoutineAddress(local_58);
  if (pcVar3 == (code *)0x0) {
    local_68[0] = 0;
    local_res20[0] = 0;
    local_60 = (undefined (*) [16])0x0;
    *param_3 = (undefined (*) [16])0x0;
    pauVar6 = (undefined (*) [16])0x0;
    _local_48 = ZEXT816(0);
    local_28 = 0;
    local_38 = ZEXT816(0);
    uVar5 = SepSddlDaclFromSDDLString(param_1,param_2,local_68,&local_60);
    pauVar1 = local_60;
    uVar4 = uVar5 & 0xffffffff;
    if (-1 < (int)uVar5) {
      RtlCreateSecurityDescriptor(local_48,1);
      RtlSetDaclSecurityDescriptor(local_48,1,pauVar1,0);
      local_48._2_2_ = local_48._2_2_ | (ushort)local_68[0];
      RtlAbsoluteToSelfRelativeSD(local_48,0,local_res20);
      pauVar6 = (undefined (*) [16])ExAllocatePoolWithTag(1,local_res20[0],0x64536553);
      if (pauVar6 == (undefined (*) [16])0x0) {
        uVar4 = 0xc000009a;
      }
      else {
        memset(pauVar6,0,(ulonglong)local_res20[0]);
        uVar2 = RtlAbsoluteToSelfRelativeSD(local_48,pauVar6,local_res20);
        uVar4 = (ulonglong)uVar2;
        if (-1 < (int)uVar2) {
          ExFreePoolWithTag(pauVar1,0);
          *param_3 = pauVar6;
          return uVar4;
        }
      }
    }
    if (pauVar1 != (undefined (*) [16])0x0) {
      ExFreePoolWithTag(pauVar1,0);
    }
    if (pauVar6 != (undefined (*) [16])0x0) {
      ExFreePoolWithTag(pauVar6,0);
    }
  }
  else {
    uVar4 = (*pcVar3)(param_1,1,param_3);
    if ((-1 < (int)uVar4) && (param_2 != 0)) {
      *(ushort *)(**param_3 + 2) = *(ushort *)(**param_3 + 2) | 8;
    }
  }
  return uVar4;
}



int FUN_14000a530(undefined8 param_1,undefined8 param_2,undefined4 param_3,undefined4 param_4,
                 undefined8 param_5,undefined4 *param_6,undefined8 *param_7)

{
  int iVar1;
  undefined4 local_res8 [2];
  undefined8 local_res10 [3];
  undefined8 local_38;
  undefined8 local_30;
  undefined8 local_28;
  undefined4 local_20;
  undefined4 local_1c;
  undefined8 local_18;
  undefined8 local_10;
  
  local_1c = 0;
  local_res8[0] = 0;
  local_res10[0] = 0;
  local_10 = 0;
  local_18 = param_5;
  local_38 = 0x30;
  local_20 = 0x240;
  local_30 = param_1;
  local_28 = param_2;
  iVar1 = ZwCreateKey(local_res10,param_3,&local_38,0,0,param_4,local_res8);
  if (iVar1 < 0) {
    local_res10[0] = 0;
    local_res8[0] = 0;
  }
  *param_7 = local_res10[0];
  if (param_6 != (undefined4 *)0x0) {
    *param_6 = local_res8[0];
  }
  return iVar1;
}



/* Library Function - Single Match
    CmRegUtilCreateWstrKey
   
   Library: Visual Studio 2019 Release */

void CmRegUtilCreateWstrKey
               (undefined8 param_1,ulonglong param_2,undefined4 param_3,undefined4 param_4,
               undefined8 param_5,undefined4 *param_6,undefined8 *param_7)

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    FUN_14000a530(param_1,local_18,param_3,param_4,param_5,param_6,param_7);
  }
  return;
}



void FUN_14000a648(undefined8 param_1,undefined8 param_2,undefined4 param_3,undefined8 *param_4)

{
  int iVar1;
  undefined8 local_res8;
  undefined8 local_38;
  undefined8 local_30;
  undefined8 local_28;
  undefined4 local_20;
  undefined4 local_1c;
  undefined local_18 [16];
  
  local_1c = 0;
  local_res8 = 0;
  *param_4 = 0;
  local_38 = 0x30;
  local_20 = 0x240;
  local_18 = ZEXT816(0);
  local_30 = param_1;
  local_28 = param_2;
  iVar1 = ZwOpenKey(&local_res8,param_3,&local_38);
  if (-1 < iVar1) {
    *param_4 = local_res8;
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilOpenExistingWstrKey
   
   Library: Visual Studio 2019 Release */

void CmRegUtilOpenExistingWstrKey
               (undefined8 param_1,ulonglong param_2,undefined4 param_3,undefined8 *param_4)

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    FUN_14000a648(param_1,local_18,param_3,param_4);
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilUcValueGetDword
   
   Library: Visual Studio 2019 Release */

void CmRegUtilUcValueGetDword
               (undefined8 param_1,undefined8 param_2,undefined4 param_3,undefined4 *param_4)

{
  int iVar1;
  undefined auStack_58 [32];
  undefined4 local_38;
  undefined4 *local_30;
  undefined4 local_28 [2];
  undefined local_20 [4];
  int local_1c;
  undefined4 local_14;
  ulonglong local_10;
  
  local_10 = DAT_140007010 ^ (ulonglong)auStack_58;
  local_28[0] = 0;
  local_30 = local_28;
  local_38 = 0x10;
  iVar1 = ZwQueryValueKey(param_1,param_2,2,local_20);
  if ((-1 < iVar1) && (local_1c == 4)) {
    param_3 = local_14;
  }
  *param_4 = param_3;
  __check_stack_cookie(local_10 ^ (ulonglong)auStack_58);
  return;
}



int FUN_14000a78c(undefined8 param_1,ushort *param_2,int param_3,int param_4,
                 undefined (**param_5) [16])

{
  uint uVar1;
  int iVar2;
  undefined (*pauVar3) [16];
  uint uVar4;
  uint local_res20 [2];
  
  local_res20[0] = 0;
  *param_5 = (undefined (*) [16])0x0;
  uVar1 = *param_2 + 0x1f & 0xfffffff8;
  uVar4 = uVar1 + param_4;
  if (uVar4 < uVar1) {
    return -0x3fffff6b;
  }
  pauVar3 = (undefined (*) [16])ExAllocatePoolWithTag(0x200,uVar4,0x62527050);
  if (pauVar3 == (undefined (*) [16])0x0) {
LAB_14000a8f1:
    iVar2 = -0x3fffff66;
  }
  else {
    memset(pauVar3,0,(ulonglong)uVar4);
    iVar2 = ZwQueryValueKey(param_1,param_2,1,pauVar3,uVar4,local_res20);
    if (iVar2 < 0) {
      ExFreePoolWithTag(pauVar3,0);
      if ((iVar2 != -0x7ffffffb) && (iVar2 != -0x3fffffdd)) {
        return iVar2;
      }
      pauVar3 = (undefined (*) [16])ExAllocatePoolWithTag(0x200,local_res20[0],0x62527050);
      if (pauVar3 == (undefined (*) [16])0x0) goto LAB_14000a8f1;
      memset(pauVar3,0,(ulonglong)local_res20[0]);
      iVar2 = ZwQueryValueKey(param_1,param_2,1,pauVar3,local_res20[0],local_res20);
      if (-1 < iVar2) goto LAB_14000a8c6;
    }
    else {
LAB_14000a8c6:
      if ((param_3 == 0) || (*(int *)((longlong)*pauVar3 + 4) == param_3)) {
        *param_5 = pauVar3;
        return 0;
      }
      iVar2 = -0x3fffffdc;
    }
    ExFreePoolWithTag(pauVar3,0);
  }
  return iVar2;
}



/* Library Function - Single Match
    CmRegUtilUcValueSetUcString
   
   Library: Visual Studio 2019 Release */

ulonglong CmRegUtilUcValueSetUcString(undefined8 param_1,undefined8 param_2,ushort *param_3)

{
  ushort uVar1;
  void *_Dst;
  uint uVar2;
  ulonglong uVar3;
  ulonglong uVar4;
  undefined local_18 [8];
  void *pvStack_10;
  
  uVar1 = *param_3;
  _local_18 = ZEXT816(0);
  if ((ulonglong)param_3[1] - (ulonglong)uVar1 < 2) {
    uVar3 = FUN_14000abe8((undefined2 *)local_18,uVar1);
    uVar4 = uVar3 & 0xffffffff;
    if (-1 < (int)uVar3) {
      _Dst = pvStack_10;
      local_18._0_2_ = *param_3;
      memmove(_Dst,*(void **)(param_3 + 4),(ulonglong)*param_3);
      *(undefined2 *)((longlong)pvStack_10 + (ulonglong)((ushort)local_18._0_2_ >> 1) * 2) = 0;
      uVar2 = ZwSetValueKey(param_1,param_2,0,1,pvStack_10,(ushort)local_18._0_2_ + 2);
      uVar4 = (ulonglong)uVar2;
      RtlFreeUnicodeString(local_18);
    }
  }
  else {
    *(undefined2 *)(*(longlong *)(param_3 + 4) + (ulonglong)(uVar1 >> 1) * 2) = 0;
    uVar2 = ZwSetValueKey(0,param_2,0,1,*(undefined8 *)(param_3 + 4),*param_3 + 2);
    uVar4 = (ulonglong)uVar2;
  }
  return uVar4;
}



/* Library Function - Single Match
    CmRegUtilWstrValueGetDword
   
   Library: Visual Studio 2019 Release */

void CmRegUtilWstrValueGetDword
               (undefined8 param_1,ulonglong param_2,undefined4 param_3,undefined4 *param_4)

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    CmRegUtilUcValueGetDword(param_1,local_18,param_3,param_4);
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilWstrValueGetFullBuffer
   
   Library: Visual Studio 2019 Release */

void CmRegUtilWstrValueGetFullBuffer
               (undefined8 param_1,ulonglong param_2,int param_3,int param_4,
               undefined (**param_5) [16])

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    FUN_14000a78c(param_1,(ushort *)local_18,param_3,param_4,param_5);
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilWstrValueSetFullBuffer
   
   Library: Visual Studio 2019 Release */

void CmRegUtilWstrValueSetFullBuffer
               (undefined8 param_1,ulonglong param_2,undefined4 param_3,undefined8 param_4,
               undefined4 param_5)

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    ZwSetValueKey(param_1,local_18,0,param_3,param_4,param_5);
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilWstrValueSetUcString
   
   Library: Visual Studio 2019 Release */

void CmRegUtilWstrValueSetUcString(undefined8 param_1,ulonglong param_2,ushort *param_3)

{
  undefined8 uVar1;
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
  if (-1 < (int)uVar1) {
    CmRegUtilUcValueSetUcString(param_1,local_18,param_3);
  }
  return;
}



/* Library Function - Single Match
    CmRegUtilWstrValueSetWstrString
   
   Library: Visual Studio 2019 Release */

void CmRegUtilWstrValueSetWstrString(undefined8 param_1,ulonglong param_2,ulonglong param_3)

{
  undefined8 uVar1;
  undefined local_28 [16];
  undefined local_18 [16];
  
  local_18 = ZEXT816(0);
  local_28 = ZEXT816(0);
  uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_28,param_3);
  if (-1 < (int)uVar1) {
    uVar1 = WdmlibRtlInitUnicodeStringEx((ushort *)local_18,param_2);
    if (-1 < (int)uVar1) {
      CmRegUtilUcValueSetUcString(param_1,local_18,(ushort *)local_28);
    }
  }
  return;
}



undefined8 FUN_14000abe8(undefined2 *param_1,ushort param_2)

{
  void *pvVar1;
  undefined8 uVar2;
  
  param_1[1] = param_2 + 2;
  *param_1 = 0;
  pvVar1 = (void *)ExAllocatePoolWithTag(1,(ulonglong)param_2 + 2,0x63557050);
  *(void **)(param_1 + 4) = pvVar1;
  if (pvVar1 == (void *)0x0) {
    uVar2 = 0xc000009a;
  }
  else {
    memset(pvVar1,0,(ulonglong)param_2 + 2);
    uVar2 = 0;
  }
  return uVar2;
}



void entry(longlong param_1)

{
  __security_init_cookie();
  DriverEntry((DRIVER_OBJECT *)param_1);
  return;
}



/* Library Function - Single Match
    __security_init_cookie
   
   Library: Visual Studio 2019 Release */

void __cdecl __security_init_cookie(void)

{
  ulonglong uVar1;
  
  if ((DAT_140007010 == 0) || (DAT_140007010 == 0x2b992ddfa232)) {
    uVar1 = rdtsc();
    DAT_140007010 =
         ((CONCAT44((int)(DAT_140007010 >> 0x20),(int)uVar1) | uVar1 & 0xffffffff00000000) ^
         0x140007010) & 0xffffffffffff;
    if (DAT_140007010 == 0) {
      DAT_140007010 = 0x2b992ddfa232;
    }
  }
  DAT_140007018 = ~DAT_140007010;
  return;
}


